/**
 * @file  bscc2XML.c
 * @brief Impletation of bscc to XML functions
 *
 * @date 2004/02/12
 * @version 1.0.0
 *
 * Copyright (c) 2002-2004, BBR Co.Ltd., All rights reserved.
 */
/*
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include "bscc2sts.h"
#include "sts2XML.h"
#include "bscc2XML.h"


/**
 * ץ󥿥ơ̵֤ͭϤ롣
 *
 * @param pSts ibscc2sts_tbl(ΰϤδؿƤ¦ǳ)
 * @retval 0 Ǥʤ
 * @retval 1 
 */
int isPrinting(bscc2sts_tbl *pSts)
{
    char printing = pSts->base_stat[0];
    if (printing >= '0' && printing <= '7') {
        return 0;
    } else {
        return 1;
    }
}

/**
 * ץ󥿤ץ󥿥ơɤ߽Ф bscc2sts_tbl롣
 *
 * @param printerName ץ̾
 * @param readFd      ɽФѥץ󥿥եǥץ
 * @param writeFd     ѥץ󥿥եǥץ
 * @param pSts        ibscc2sts_tbl(ΰϤδؿƤ¦ǳ)
 * @retval 0               
 * @retval <0              bscc2stsΥ顼:bscc2sts.h
 * @retval ERR_FILE_READ   File read error.
 */
int genSts(char *printerName, int readFd, int writeFd, bscc2sts_tbl *pSts)
{
    int retCode;

    /* ץ */
    char bsccBuf[BSCC_BUF_SIZE];
    retCode = readControl(readFd, writeFd, bsccBuf, BSCC_BUF_SIZE);
    if(retCode != 0){
        return retCode;
    }

    /* bscc2sts()ƤӽФɬפͤ2Ѱ */
    /* 1:prnId */
    char prnId[16];
    strncpy(prnId, printerName, 16);

    /* 2:output_dev */
    output_dev outDev;
    outDev.dev = 1;
    outDev.status = '0';
    outDev.recv[0] = ' ';
    outDev.recv[1] = ' ';
    outDev.recv[2] = ' ';

    retCode = bscc2sts(prnId, bsccBuf, &outDev, pSts);
    if(retCode != 0){
        return retCode;
    }

//    ** for debug **
//    pSts->reserve = '\0';
//    printf("%s\n", (char *)pSts);
    return 0;
}

/**
 * ץ󥿤ɽФ椷ץ󥿥ơɤ߽Ф
 *
 * @param readFd  ɽФѥץ󥿥եǥץ
 * @param writeFd ѥץ󥿥եǥץ
 * @param bscc    ץ󥿥ơǡ(ΰϤδؿƤ¦ǳ)
 * @retval 0               
 * @retval ERR_FILE_READ   File read error.
 */
static int readControl(int readFd, int writeFd, char *bscc, int bsccSize){
    int rSize;

    /* ޤǽˡɤ߹ʼΤƤ */
    memset((void *)bscc, 0, bsccSize);
    rSize = read(readFd, bscc, bsccSize);

    /* ɤ߹ߥξɤ߹ߥ顼 */
    if (rSize < 0) {
        return ERR_FILE_READ;
    } else if (rSize == 0) {
    /* nullread硢ǥץ󥿤Ƥ */
        setupBscc(writeFd);

        /* ץ󥿤ɹ߲ǽˤʤΤǧ */
        while (rSize == 0) {
//            sleep(1);
            rSize = read(readFd, bscc, bsccSize);
        }
    }

    /* ֥ǡɤ߽Ф */
    memset((void *)bscc, 0, bsccSize);
    rSize = read(readFd, bscc, bsccSize);

    /* ɤ߹ߥξɤ߹ߥ顼 */
    if (rSize < 0){
        return ERR_FILE_READ;
    }

    return 0;
}

/*
 * PIXUS 850iץ󥿤饹ơƤӽФν
 * ץŸˡδؿΥޥɥåȤ
 * ץ󥿤뤳Ȥˤꡢץ󥿤ΥơɽФǽȤʤ롣
 * CANONθ饤֥ž¤
 * ΤdefinesetupBscc()ؿľ֤롣
 */
#define BSCC_SSR "SSR=BST,SFA,CHD,CIL,CIR,LVR,DBS,DWS,DOC,DSC,DJS;"
#define COUNT_LO 3
#define COUNT_HI 4
#define LENGTH_HI 7
#define LENGTH_LO 8
#define BSCC_HEADSIZE 9

static int setupBscc(int fd)
{
    int retVal;
    int count;
    char *bscc_com;
    //                  ESC   [    K   C-lo C-hi INIT  ID  L-hi L-lo (total 9byte)
    char bscc_head[] = {0x1b,0x5b,0x4b,0x00,0x00,0x00,0x1e,0x00,0x00};
    int size = BSCC_HEADSIZE; /* bscc_head size */

    bscc_com = (char *)(malloc(512));
    if(bscc_com == NULL)
        return -1; /* no RAM */

    /* build BSCC command */
    memset(bscc_com, 0, 512);
    memcpy((void *)bscc_com, (void *)bscc_head, size);    /* BSCC header copy */
    size = strlen(BSCC_SSR);
    memcpy((void *)(bscc_com+BSCC_HEADSIZE), (void *)BSCC_SSR, size); /* BSCC body copy */
    size = strlen(BSCC_SSR) + 4;             /* size = BSCC body + 4(INIT,ID,L-hi,L-lo) */
    *(bscc_com + COUNT_LO) = (char)(0x0000ff & size);     /* set Low  byte */
    *(bscc_com + COUNT_HI) = (char)((0x00ff00 & size) >> 8);     /* set High byte */
    size = strlen(BSCC_SSR) + 2;             /* size = BSCC body + 2(L-hi,L-lo) */
    *(bscc_com + LENGTH_LO) = (char)(0x0000ff & size);    /* set Low  byte */
    *(bscc_com + LENGTH_HI) = (char)((0x00ff00 & size) >> 8);    /* set High byte */

    size = strlen(BSCC_SSR) + BSCC_HEADSIZE;

    retVal = write(fd, bscc_com, size);

    free(bscc_com);

    return retVal;
}
