/**
 * @file  sts2XML.c
 * @brief Impletation of sts to XML functions
 *
 * @date 2004/02/12
 * @version 1.0.0
 *
 * Copyright (c) 2002-2004, BBR Co.Ltd., All rights reserved.
 */
/*
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <locale.h>
#include <langinfo.h>
#include <iconv.h>
#include <errno.h>
#include <libintl.h>
#include "bidiDbg.h"
#include "bscc2sts.h"
#include "sts2XML.h"

#define PACKAGE "bidiPIXUS850i" /** for multi locale. */

/**
 * ¤(bscc2sts_tbl)ǡꡢXMLѴϤ롣
 *
 * @param sts     bscc2sts_tbl¤Ρ
 * @param xml     ̤֤ʸΥݥ󥿡(mallocϤδؿƤ¦ǹԤ)
 *                ڰʲ͡
 *   إåեå      :200byte
 *   饤:250byte
 *   ޡץ饤    :450byte*n(nϥ󥯥ȥåθĿ:8?)=4k
 *   顼            :300byte*n(nϥ顼˥󥰤θĿ:8+12?)=6k
 *                     :ͤ200+250+450*8+300*20=10,050byte
 * @param xmlSize XMLκ祵
 * @param pLang   ʸɡ
 * @return Number of bytes of the xml string, -1 when error.
 */
int sts2XML(bscc2sts_tbl *sts, char *xml, int xmlSize, char *pLang)
{
    /* set locale and prepare for gettext */
    if (pLang == NULL) {
        pLang = "C";
    }
    setlocale(LC_ALL, pLang);
    bindtextdomain(PACKAGE, LOCALE_DIR);
    textdomain(PACKAGE);

    /* ܸ쥳ѴѤ˥Хåեγ */
//    char *pXmlBuf = (char *) malloc(xmlSize);
//    if (pXmlBuf == NULL) {
//        return -1;
//    }

    /* Хåեޤꥫ */
    int cnt = 0;

    /* ѥХåե˥ԡƤ */
    /* Ϥޤ */
    if (xmlSize <= cnt + strlen(XML_HEADER)) {
//        free(pXmlBuf);
        return -1;
    }
    memcpy(xml + cnt, XML_HEADER, strlen(XML_HEADER));
    cnt += strlen(XML_HEADER);

    if (xmlSize <= cnt + strlen(XML_MIBS)) {
//        free(pXmlBuf);
        return -1;
    }
    memcpy(xml + cnt, XML_MIBS, strlen(XML_MIBS));
    cnt += strlen(XML_MIBS);

    /* 3ĤΥ벼̴ؿƤӽФ˻ѤХåե */
    char txml[6144];
    int txmlSize;

    /* 饤󡦥ơ֥ */
    txmlSize = createPrtLocalizationTable(txml, sizeof(txml));
    if (xmlSize <= cnt + txmlSize || txmlSize == -1) {
//        free(pXmlBuf);
        return -1;
    }
    memcpy(xml + cnt, txml, txmlSize);
    cnt += txmlSize;

    /* ޡץ饤ơ֥ */
    txmlSize = createPrtMarkerSuppliesTable(sts, txml, sizeof(txml));
    if (xmlSize <= cnt + txmlSize || txmlSize == -1) {
//        free(pXmlBuf);
        return -1;
    }
    memcpy(xml + cnt, txml, txmlSize);
    cnt += txmlSize;

    /* 顼ȡơ֥ */
    txmlSize = createPrtAlertTable(sts, txml, sizeof(txml));
    if (xmlSize <= cnt + txmlSize || txmlSize == -1) {
//        free(pXmlBuf);
        return -1;
    }
    memcpy(xml + cnt, txml, txmlSize);
    cnt += txmlSize;

    /* ޤ */
    if (xmlSize <= cnt + strlen(XML_MIBE)) {
//        free(pXmlBuf);
        return -1;
    }
    memcpy(xml + cnt, XML_MIBE, strlen(XML_MIBE));
    cnt += strlen(XML_MIBE);

    /* ǰΤᡢüդƤ */
    memcpy(xml + cnt, "\0", 1);

/*   int ret = xmlConv(pXmlBuf, strlen(pXmlBuf), xml, xmlSize);
    if (ret == -1) {
        free(pXmlBuf);
        return -1;
    }
    free(pXmlBuf);
*/
    return cnt;
}

/**
 * XMLܸ쥳ɤUTF-8 Ѵ
 *
 * @param fromStr    ѴʸǼƤХåեʸƤӸǳݡ
 * @param fromStrLen fromStrʸĹ
 * @param toStr      ѴʸǼ뤿ΥХåեʸƤӸǳݡ
 * @param toStrLen   toStrʸĹ
 * @return Number of bytes of the toStr string, -1 when error.
 * @deprecated 2004/02/10 
 */
static int xmlConv(char *fromStr, int fromStrLen, char *toStr, int toStrLen)
{
    iconv_t cd;
    size_t nconv;
    char *fromPtr, *toPtr;
    int fromLen, toLen;
    int ret;

    /* locale б */
//    setlocale(LC_ALL, "");
//    char *code = nl_langinfo(CODESET);
//    cd = iconv_open("UTF-8", code); /* EUC-JP -> UTF-8 */
    cd = iconv_open("UTF-8", "EUC-JP"); /* EUC-JP -> UTF-8 */
    if (cd == (iconv_t)-1) {
        _dbgPrint(sys_errlist[errno]);
        return -1;
    }
    fromPtr = fromStr;
    fromLen = strlen(fromPtr);
    toPtr = toStr;
    toLen = toStrLen;
    nconv = iconv(cd, &fromPtr, &fromLen, &toPtr, &toLen);
    if (nconv == -1) {
        _dbgPrint(sys_errlist[errno]);
        return -1;
    }
    *toPtr = '\0';
    ret = iconv_close(cd);
    if (ret == -1) {
        _dbgPrint(sys_errlist[errno]);
        return -1;
    }
    return toStr - toPtr;
}

/**
 * XMLΰ(饤󡦥ơ֥)롣
 *
 * @param xml     LocalizationTableXML(mallocϤδؿƤ¦ǹԤ)
 * @param xmlSize LocalizationTableXMLκ祵
 * @return Number of bytes of the xml string, -1 when error.
 */
static int createPrtLocalizationTable(char *xml, int xmlSize)
{
    /* ʸιץХȿϥХåեĶʤå */
    /* (ʬ224byte;͵򸫤250ȤƤ */
    if (xmlSize <= 250) {
        return -1;
    }

    /* Хåեޤꥫ */
    int cnt = 0;

    memcpy(xml + cnt, XML_PLTS, strlen(XML_PLTS));
    cnt += strlen(XML_PLTS);
    memcpy(xml + cnt, XML_PLES, strlen(XML_PLES));
    cnt += strlen(XML_PLES);
    memcpy(xml + cnt, XML_PLLS, strlen(XML_PLLS));
    cnt += strlen(XML_PLLS);
    memcpy(xml + cnt, LANGUAGE, strlen(LANGUAGE));
    cnt += strlen(LANGUAGE);
    memcpy(xml + cnt, XML_PLLE, strlen(XML_PLLE));
    cnt += strlen(XML_PLLE);
    memcpy(xml + cnt, XML_PLCS, strlen(XML_PLCS));
    cnt += strlen(XML_PLCS);
    memcpy(xml + cnt, CHAR_SET, strlen(CHAR_SET));
    cnt += strlen(CHAR_SET);
    memcpy(xml + cnt, XML_PLCE, strlen(XML_PLCE));
    cnt += strlen(XML_PLCE);
    memcpy(xml + cnt, XML_PLEE, strlen(XML_PLEE));
    cnt += strlen(XML_PLEE);
    memcpy(xml + cnt, XML_PLTE, strlen(XML_PLTE));
    cnt += strlen(XML_PLTE);

    /* ǰΤᡢüդƤ */
    memcpy(xml + cnt, "\0", 1);
    return cnt;
}

/**
 * XMLΰ(ޡץ饤ơ֥)롣
 *
 * @param sts     bscc2sts_tbl¤
 * @param xml     MarkerSuppliesTableXML(mallocϤδؿƤ¦ǹԤ)
 * @param xmlSize MarkerSuppliesTableXMLκ祵
 * @return Number of bytes of the xml string, -1 when error.
 */
static int createPrtMarkerSuppliesTable(bscc2sts_tbl *sts, char *xml, int xmlSize)
{
    /* 롼ץ:-) */
    int i;

    /* Хåեޤꥫ */
    int cnt = 0;

    /* Ϥޤ*/
    if (xmlSize <= cnt + strlen(XML_PMTS)) {
        return -1;
    }
    memcpy(xml + cnt, XML_PMTS, strlen(XML_PMTS));
    cnt += strlen(XML_PMTS);

    /* ޡץ饤ȥ */
    char txml[512];
    int txmlSize;
    char pmsIndex[2];
    char pmsDescription[20];
    char pmsLevel[4];

    /* 󥯹¤Τå */
    for (i = 0; i < sizeof(*sts->ink); i++) {
        if (sts->ink[i].color == ' ') {
            continue;
        }

        /* Ѵ */
        switch (sts->ink[i].color) {
            case 'C':
                strcpy(pmsIndex, PMI_C_NUM);
                strcpy(pmsDescription, PMD_C_CHR);
                break;
            case 'M':
                strcpy(pmsIndex, PMI_M_NUM);
                strcpy(pmsDescription, PMD_M_CHR);
                break;
            case 'Y':
                strcpy(pmsIndex, PMI_Y_NUM);
                strcpy(pmsDescription, PMD_Y_CHR);
                break;
            case 'L':
                strcpy(pmsIndex, PMI_L_NUM);
                strcpy(pmsDescription, PMD_L_CHR);
                break;
            default:
                strcpy(pmsIndex, "");
                strcpy(pmsDescription, "");
        }

        /* pmsLevelκ0 */
        if (sts->ink[i].fill[0] == '0' || sts->ink[i].fill[0] == ' ') {
            if (sts->ink[i].fill[1] == '0' || sts->ink[i].fill[1] == ' ') {
                if (sts->ink[i].fill[2] == ' ') {
                    pmsLevel[0] = '\0';
                } else {
                    pmsLevel[0] = sts->ink[i].fill[2];
                    pmsLevel[1] = '\0';
                }
            } else {
                pmsLevel[0] = sts->ink[i].fill[1];
                pmsLevel[1] = sts->ink[i].fill[2];
                pmsLevel[2] = '\0';
            }
        } else {
            pmsLevel[0] = sts->ink[i].fill[0];
            pmsLevel[1] = sts->ink[i].fill[1];
            pmsLevel[2] = sts->ink[i].fill[2];
            pmsLevel[3] = '\0';
        }

        /* ̴ؿƤӽФ */
        txmlSize = createPrtMarkerSuppliesEntry(pmsIndex, PMY,
                                                 pmsDescription, PMS, PMM,
                                                 pmsLevel, txml, sizeof(txml));
        if (xmlSize <= cnt + txmlSize || txmlSize == -1) {
            return -1;
        }
        memcpy(xml + cnt, txml, txmlSize);
        cnt += txmlSize;
    }

    /* ޤ */
    if (xmlSize <= cnt + strlen(XML_PMTE)) {
        return -1;
    }
    memcpy(xml + cnt, XML_PMTE, strlen(XML_PMTE));
    cnt += strlen(XML_PMTE);

    /* ǰΤᡢüդƤ */
    memcpy(xml + cnt, "\0", 1);
    return cnt;
}

/**
 * XMLΰ(ޡץ饤ȥ)롣
 *
 * @param pmsIndex       
 * @param pmsType        
 * @param pmsDescription 
 * @param pmsSupplyUnit  
 * @param pmsMaxCapacity 
 * @param pmsLevel       
 * @param xml            MarkerSuppliesEntryXML(mallocϤδؿƤ¦ǹԤ)
 * @param xmlSize        MarkerSuppliesEntryXMLκ祵
 * @return Number of bytes of the xml string, -1 when error.
 */
static int createPrtMarkerSuppliesEntry(char *pmsIndex, char *pmsType,
                                        char *pmsDescription, char *pmsSupplyUnit,
                                        char *pmsMaxCapacity, char *pmsLevel,
                                        char *xml, int xmlSize)
{
    /* ѿȸʸιץХȿϥХåեĶʤå */
    /* (ʬ406byte;͵򸫤420ȤƤ */
    if (xmlSize <= 420 + strlen(pmsIndex) + strlen(pmsType)
                      + strlen(pmsDescription) + strlen(pmsSupplyUnit)
                      + strlen(pmsMaxCapacity) + strlen(pmsLevel)) {
        return -1;
    }

    /* Хåեޤꥫ */
    int cnt = 0;

    memcpy(xml + cnt, XML_PMES, strlen(XML_PMES));
    cnt += strlen(XML_PMES);
    memcpy(xml + cnt, XML_PMIS, strlen(XML_PMIS));
    cnt += strlen(XML_PMIS);
    memcpy(xml + cnt, pmsIndex, strlen(pmsIndex));
    cnt += strlen(pmsIndex);
    memcpy(xml + cnt, XML_PMIE, strlen(XML_PMIE));
    cnt += strlen(XML_PMIE);
    memcpy(xml + cnt, XML_PMYS, strlen(XML_PMYS));
    cnt += strlen(XML_PMYS);
    memcpy(xml + cnt, pmsType, strlen(pmsType));
    cnt += strlen(pmsType);
    memcpy(xml + cnt, XML_PMYE, strlen(XML_PMYE));
    cnt += strlen(XML_PMYE);
    memcpy(xml + cnt, XML_PMDS, strlen(XML_PMDS));
    cnt += strlen(XML_PMDS);
    memcpy(xml + cnt, pmsDescription, strlen(pmsDescription));
    cnt += strlen(pmsDescription);
    memcpy(xml + cnt, XML_PMDE, strlen(XML_PMDE));
    cnt += strlen(XML_PMDE);
    memcpy(xml + cnt, XML_PMSS, strlen(XML_PMSS));
    cnt += strlen(XML_PMSS);
    memcpy(xml + cnt, pmsSupplyUnit, strlen(pmsSupplyUnit));
    cnt += strlen(pmsSupplyUnit);
    memcpy(xml + cnt, XML_PMSE, strlen(XML_PMSE));
    cnt += strlen(XML_PMSE);
    memcpy(xml + cnt, XML_PMMS, strlen(XML_PMMS));
    cnt += strlen(XML_PMMS);
    memcpy(xml + cnt, pmsMaxCapacity, strlen(pmsMaxCapacity));
    cnt += strlen(pmsMaxCapacity);
    memcpy(xml + cnt, XML_PMME, strlen(XML_PMME));
    cnt += strlen(XML_PMME);
    memcpy(xml + cnt, XML_PMLS, strlen(XML_PMLS));
    cnt += strlen(XML_PMLS);
    memcpy(xml + cnt, pmsLevel, strlen(pmsLevel));
    cnt += strlen(pmsLevel);
    memcpy(xml + cnt, XML_PMLE, strlen(XML_PMLE));
    cnt += strlen(XML_PMLE);
    memcpy(xml + cnt, XML_PMEE, strlen(XML_PMEE));
    cnt += strlen(XML_PMEE);

    /* ǰΤᡢüդƤ */
    memcpy(xml + cnt, "\0", 1);
    return cnt;
}

/**
 * XMLΰ(顼ȡơ֥)롣
 *
 * @param sts     bscc2_sts_tbl¤
 * @param xml     AlertTableXML(mallocϤδؿƤ¦ǹԤ)
 * @param xmlSize AlertTableXMLκ祵
 * @return Number of bytes of the xml string, -1 when error.
 */
static int createPrtAlertTable(bscc2sts_tbl *sts, char *xml, int xmlSize)
{
    /* 롼ץ:-) */
    int i;

    /* Хåեޤꥫ */
    int cnt = 0;

    /* Ϥޤ */
    if (xmlSize <= cnt + strlen(XML_PATS)) {
        return -1;
    }
    memcpy(xml + cnt, XML_PATS, strlen(XML_PATS));
    cnt += strlen(XML_PATS);

    /* 顼ȡȥ */
    char txml[512];
    int txmlSize;
    char paSeverityLevel[2];
    char paGroup[3];
    char paGroupIndex[2];
    char paCode[5];
    char paDescription[100];

    /* ڥ졼å */
    for (i = 0; i < sizeof(sts->operator_call); i++) {
        if (sts->operator_call[i] == ' ') {
            continue;
        }

        /* Ѵ */
        switch (sts->operator_call[i]) {
            case 'O': //Сץ
                strcpy(paGroup, PAG_COVER);
                strcpy(paCode, PAC_O_COD);
                strcpy(paDescription, PAD_O_CHR);
                break;
            case 'P': //ѻʤ
                strcpy(paGroup, PAG_INPUT);
                strcpy(paCode, PAC_P_COD);
                strcpy(paDescription, PAD_P_CHR);
                break;
            case 'J': //楸
                strcpy(paGroup, PAG_INPUT);
                strcpy(paCode, PAC_J_COD);
                strcpy(paDescription, PAD_J_CHR);
                break;
            case 'H': //إå̤
            case 'S': //ݡȳإå
                strcpy(paGroup, PAG_MAKER);
                strcpy(paCode, PAC_H_COD);
                strcpy(paDescription, PAD_H_CHR);
                break;
            case 'G': //CDȥ쥤å
                strcpy(paGroup, PAG_OTHER);
                strcpy(paCode, PAC_G_COD);
                strcpy(paDescription, PAD_G_CHR);
                break;
            case 'K':
            case 'c':
            case 'm':
            case 'C':
            case 'M':
            case 'Y':
            case 'k':
            case 'L':
            case 'B':
            case 'E':
                // 󥯥ȤΥ顼 ink[] ǽ뤿ᤳǤϥå
                continue;
                break;
            default:
                strcpy(paGroup, PAG_OTHER);
                strcpy(paCode, PAC_X_COD);
                strcpy(paDescription, PAD_X_CHR);
        }
        strcpy(paSeverityLevel, PAS_C_NUM);
        strcpy(paGroupIndex, PAI_NOTINK);

        /* ̴ؿƤӽФ */
        txmlSize = createPrtAlertEntry(paSeverityLevel, paGroup, paGroupIndex,
                                        paCode, paDescription,
                                        txml, sizeof(txml));
        if (xmlSize <= cnt + txmlSize || txmlSize == -1) {
            return -1;
        }
        memcpy(xml + cnt, txml, txmlSize);
        cnt += txmlSize;
    }

    /* 󥯥Ȥå */
    for (i = 0; i < sizeof(*sts->ink); i++) {
        if (sts->ink[i].color == ' ') {
            continue;
        }
        if (sts->ink[i].alert != 'O') {
            continue;
        }

        /* Ѵ */
        switch (sts->ink[i].color) {
            case 'C':
                strcpy(paGroupIndex, PMI_C_NUM);
                strcpy(paDescription, PMD_C_CHR);
                strcat(paDescription, PAD_IO_CHR);
                break;
            case 'M':
                strcpy(paGroupIndex, PMI_M_NUM);
                strcpy(paDescription, PMD_M_CHR);
                strcat(paDescription, PAD_IO_CHR);
                break;
            case 'Y':
                strcpy(paGroupIndex, PMI_Y_NUM);
                strcpy(paDescription, PMD_Y_CHR);
                strcat(paDescription, PAD_IO_CHR);
                break;
            case 'L':
                strcpy(paGroupIndex, PMI_L_NUM);
                strcpy(paDescription, PMD_L_CHR);
                strcat(paDescription, PAD_IO_CHR);
                break;
            default:
                strcpy(paGroupIndex, "");
                strcpy(paDescription, "");
        }
        strcpy(paSeverityLevel, PAS_C_NUM);
        strcpy(paGroup, PAG_MKSPL);
        strcpy(paCode, PAC_IO_COD);

        /* ̴ؿƤӽФ */
        txmlSize = createPrtAlertEntry(paSeverityLevel, paGroup, paGroupIndex,
                                        paCode, paDescription,
                                        txml, sizeof(txml));
        if (xmlSize <= cnt + txmlSize || txmlSize == -1) {
            return -1;
        }
        memcpy(xml + cnt, txml, txmlSize);
        cnt += txmlSize;
    }

    /* 󥯥å */
    for (i = 0; i < sizeof(*sts->ink); i++) {
        if (sts->ink[i].color == ' ') {
            continue;
        }
        if (sts->ink[i].alert != 'L') {
            continue;
        }

        /* Ѵ */
        switch (sts->ink[i].color) {
            case 'C':
                strcpy(paGroupIndex, PMI_C_NUM);
                strcpy(paDescription, PMD_C_CHR);
                strcat(paDescription, PAD_IL_CHR);
                break;
            case 'M':
                strcpy(paGroupIndex, PMI_M_NUM);
                strcpy(paDescription, PMD_M_CHR);
                strcat(paDescription, PAD_IL_CHR);
                break;
            case 'Y':
                strcpy(paGroupIndex, PMI_Y_NUM);
                strcpy(paDescription, PMD_Y_CHR);
                strcat(paDescription, PAD_IL_CHR);
                break;
            case 'L':
                strcpy(paGroupIndex, PMI_L_NUM);
                strcpy(paDescription, PMD_L_CHR);
                strcat(paDescription, PAD_IL_CHR);
                break;
            default:
                strcpy(paGroupIndex, "");
                strcpy(paDescription, "");
        }
        strcpy(paSeverityLevel, PAS_W_NUM);
        strcpy(paGroup, PAG_MKSPL);
        strcpy(paCode, PAC_IL_COD);

        /* ̴ؿƤӽФ */
        txmlSize = createPrtAlertEntry(paSeverityLevel, paGroup, paGroupIndex,
                                        paCode, paDescription,
                                        txml, sizeof(txml));
        if (xmlSize <= cnt + txmlSize || txmlSize == -1) {
            return -1;
        }
        memcpy(xml + cnt, txml, txmlSize);
        cnt += txmlSize;
    }

    /* ޤ */
    if (xmlSize <= cnt + strlen(XML_PATE)) {
        return -1;
    }
    memcpy(xml + cnt, XML_PATE, strlen(XML_PATE));
    cnt += strlen(XML_PATE);

    /* ǰΤᡢüդƤ */
    memcpy(xml + cnt, "\0", 1);
    return cnt;
}

/**
 * XMLΰ(顼ȡȥ)롣
 *
 * @param paSeverityLevel 
 * @param paGroup         
 * @param paGroupIndex    
 * @param paCode          
 * @param paDescription   
 * @param xml             AlertEntryXML(mallocϤδؿƤ¦ǹԤ)
 * @param xmlSize         AlertEntryXMLκ祵
 * @retuern Number of bytes of the xml string, -1 when error.
 */
static int createPrtAlertEntry(char *paSeverityLevel, char *paGroup,
                               char *paGroupIndex, char *paCode, char *paDescription,
                               char *xml, int xmlSize)
{
    /* ѿȸʸιץХȿϥХåեĶʤå */
    /* (ʬ251byte;͵򸫤270ȤƤ */
    if (xmlSize <= 270 + strlen(paSeverityLevel) + strlen(paGroup)
                       + strlen(paGroupIndex) + strlen(paCode)
                       + strlen(paDescription)) {
        return -1;
    }

    /* Хåեޤꥫ */
    int cnt = 0;

    memcpy(xml + cnt, XML_PAES, strlen(XML_PAES));
    cnt += strlen(XML_PAES);
    memcpy(xml + cnt, XML_PASS, strlen(XML_PASS));
    cnt += strlen(XML_PASS);
    memcpy(xml + cnt, paSeverityLevel, strlen(paSeverityLevel));
    cnt += strlen(paSeverityLevel);
    memcpy(xml + cnt, XML_PASE, strlen(XML_PASE));
    cnt += strlen(XML_PASE);
    memcpy(xml + cnt, XML_PAGS, strlen(XML_PAGS));
    cnt += strlen(XML_PAGS);
    memcpy(xml + cnt, paGroup, strlen(paGroup));
    cnt += strlen(paGroup);
    memcpy(xml + cnt, XML_PAGE, strlen(XML_PAGE));
    cnt += strlen(XML_PAGE);
    memcpy(xml + cnt, XML_PAIS, strlen(XML_PAIS));
    cnt += strlen(XML_PAIS);
    memcpy(xml + cnt, paGroupIndex, strlen(paGroupIndex));
    cnt += strlen(paGroupIndex);
    memcpy(xml + cnt, XML_PAIE, strlen(XML_PAIE));
    cnt += strlen(XML_PAIE);
    memcpy(xml + cnt, XML_PACS, strlen(XML_PACS));
    cnt += strlen(XML_PACS);
    memcpy(xml + cnt, paCode, strlen(paCode));
    cnt += strlen(paCode);
    memcpy(xml + cnt, XML_PACE, strlen(XML_PACE));
    cnt += strlen(XML_PACE);
    memcpy(xml + cnt, XML_PADS, strlen(XML_PADS));
    cnt += strlen(XML_PADS);
    memcpy(xml + cnt, paDescription, strlen(paDescription));
    cnt += strlen(paDescription);
    memcpy(xml + cnt, XML_PADE, strlen(XML_PADE));
    cnt += strlen(XML_PADE);
    memcpy(xml + cnt, XML_PAEE, strlen(XML_PAEE));
    cnt += strlen(XML_PAEE);

    /* ǰΤᡢüդƤ */
    memcpy(xml + cnt, "\0", 1);
    return cnt;
}
