/**
 * @file  sts2XML.h
 * @brief Header file of sts to XML functions
 *
 * @date 2004/02/12
 * @version 1.0.0
 *
 * Copyright (c) 2002-2004, BBR Co.Ltd., All rights reserved.
 */
/*
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#ifndef _STS2XML_H
#define _STS2XML_H

/* for multi locale */
#define _(String)   gettext(String)
#define N_(String)  gettext_noop(String)
#define gettext_noop(String) (String)

/* definitions of constants */
/* XMLѸʸ */
#define XML_HEADER "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
#define XML_MIBS "<Printer-MIB xmlns:xsi=" \
    "\"http://www.w3.org/2001/XMLSchema-instance\"" \
    " xsi:noNamespaceSchemaLocation=\"printer2.xsd\">\n"
#define XML_MIBE "</Printer-MIB>\n"
#define XML_PLTS "\t<PrtLocalizationTable>\n"
#define XML_PLTE "\t</PrtLocalizationTable>\n"
#define XML_PMTS "\t<PrtMarkerSuppliesTable>\n"
#define XML_PMTE "\t</PrtMarkerSuppliesTable>\n"
#define XML_PATS "\t<PrtAlertTable>\n"
#define XML_PATE "\t</PrtAlertTable>\n"
#define XML_PLES "\t\t<PrtLocalizationEntry>\n"
#define XML_PLEE "\t\t</PrtLocalizationEntry>\n"
#define XML_PMES "\t\t<PrtMarkerSuppliesEntry>\n"
#define XML_PMEE "\t\t</PrtMarkerSuppliesEntry>\n"
#define XML_PAES "\t\t<PrtAlertEntry>\n"
#define XML_PAEE "\t\t</PrtAlertEntry>\n"
#define XML_PLLS "\t\t\t<PrtLocalizationLanguage>"
#define XML_PLLE "</PrtLocalizationLanguage>\n"
#define XML_PLCS "\t\t\t<PrtLocalizationCharacterSet>"
#define XML_PLCE "</PrtLocalizationCharacterSet>\n"
#define XML_PMIS "\t\t\t<PrtMarkerSuppliesIndex>"
#define XML_PMIE "</PrtMarkerSuppliesIndex>\n"
#define XML_PMYS "\t\t\t<PrtMarkerSuppliesType>"
#define XML_PMYE "</PrtMarkerSuppliesType>\n"
#define XML_PMDS "\t\t\t<PrtMarkerSuppliesDescription>"
#define XML_PMDE "</PrtMarkerSuppliesDescription>\n"
#define XML_PMSS "\t\t\t<PrtMarkerSuppliesSupplyUnit>"
#define XML_PMSE "</PrtMarkerSuppliesSupplyUnit>\n"
#define XML_PMMS "\t\t\t<PrtMarkerSuppliesMaxCapacity>"
#define XML_PMME "</PrtMarkerSuppliesMaxCapacity>\n"
#define XML_PMLS "\t\t\t<PrtMarkerSuppliesLevel>"
#define XML_PMLE "</PrtMarkerSuppliesLevel>\n"
#define XML_PASS "\t\t\t<PrtAlertSeverityLevel>"
#define XML_PASE "</PrtAlertSeverityLevel>\n"
#define XML_PAGS "\t\t\t<PrtAlertGroup>"
#define XML_PAGE "</PrtAlertGroup>\n"
#define XML_PAIS "\t\t\t<PrtAlertGroupIndex>"
#define XML_PAIE "</PrtAlertGroupIndex>\n"
#define XML_PACS "\t\t\t<PrtAlertCode>"
#define XML_PACE "</PrtAlertCode>\n"
#define XML_PADS "\t\t\t<PrtAlertDescription>"
#define XML_PADE "</PrtAlertDescription>\n"

/* XMLѴʸ */
#define LANGUAGE "JA"
#define CHAR_SET "106"
#define PMY "6"
#define PMS "19"
#define PMM "100"

#define PMI_C_NUM "1"
#define PMI_M_NUM "2"
#define PMI_Y_NUM "3"
#define PMI_L_NUM "4"

/* 2004/02/10 for multi locale
#define PMD_C_CHR ""
#define PMD_M_CHR "ޥ"
#define PMD_Y_CHR ""
#define PMD_L_CHR "֥å"
*/
#define PMD_C_CHR _("Cyan")
#define PMD_M_CHR _("Magenta")
#define PMD_Y_CHR _("Yellow")
#define PMD_L_CHR _("Black")

#define PAS_C_NUM "3"
#define PAS_W_NUM "4"

#define PAC_O_COD "3"
#define PAC_H_COD "1"
#define PAC_P_COD "808"
#define PAC_J_COD "8"
#define PAC_G_COD "1"
#define PAC_X_COD "1"

/* 2004/02/10 for multi locale
#define PAD_O_CHR "СƤޤ"
#define PAD_H_CHR "BJȥå˰۾郎ȯޤ"
#define PAD_P_CHR "ѻ椬ޤ"
#define PAD_J_CHR "ѻ椬ͤޤޤ"
#define PAD_G_CHR "ץ󥿤CD-Rȥ쥤դƤޤ"
#define PAD_X_CHR "¾Υ顼ȯƤޤ"
*/
#define PAD_O_CHR _("Printer cover open.")
#define PAD_H_CHR _("BJ cartridge.")
#define PAD_P_CHR _("Paper not loaded correctly.")
#define PAD_J_CHR _("Paper jam.")
#define PAD_G_CHR _("CD-R tray.")
#define PAD_X_CHR _("Unknown.")

#define PAC_IO_COD "1102"
#define PAC_IL_COD "1105"

/* 2004/02/10 for multi locale
#define PAD_IO_CHR "Υ󥯤ʤʤޤ"
#define PAD_IL_CHR "Υ󥯤ʤʤäƤޤ"
*/
#define PAD_IO_CHR _(" ink out.")
#define PAD_IL_CHR _(" ink low.")

#define PAG_MKSPL "11"
#define PAG_COVER "6"
#define PAG_MAKER "10"
#define PAG_INPUT "8"
#define PAG_OTHER "1"

#define PAI_NOTINK ""


/* declarations of ... */
/* bscc2sts_tblXMLѴ */
int sts2XML(bscc2sts_tbl *sts, char *xml, int xml_size, char *pLang);
/* XMLUTF-8Ѵ */
static int xmlConv(char *fromStr, int fromStrLen, char *toStr, int toStrLen);
/* 饤󡦥ơ֥ */
static int createPrtLocalizationTable(char *sxml, int sxml_size);
/* ޡץ饤ơ֥ */
static int createPrtMarkerSuppliesTable(bscc2sts_tbl *sts, char *sxml, int sxml_size);
/* 顼ȡơ֥ */
static int createPrtAlertTable(bscc2sts_tbl *sts, char *sxml, int sxml_size);
/* ޡץ饤ȥ */
static int createPrtMarkerSuppliesEntry(char *pmsIndex, char *pmsType,
                                        char *pmsDescription, char *pmsSupplyUnit,
                                        char *pmsMaxCapacity, char *pmsLevel,
                                        char *xml, int xml_size);
/* 顼ȡȥ */
static int createPrtAlertEntry(char *paSeverityLevel, char *paGroup,
                               char *paGroupIndex, char *paCode, char *paDescription,
                               char *xml, int xml_size);

#endif // _STS2XML_H
