/*

Copyright (c) 2003-2004, AXE, Inc.  All rights reserved.

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
/*
 * oppd_common.c
 * open-printing printer daemon common
 */
#include <stdio.h>
#include <fcntl.h>
#include <memory.h>
#include <malloc.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netdb.h>
#include <signal.h>
#include <sys/select.h>
#include <sys/wait.h>
#include <ctype.h>
#include <string.h>

#include "oppd_common.h"

/*
 * trim text
 */
char		*oppd_strtrim(
	char		*dst,
	char		*src
	)
{
	char		*p;

	if (!src) return NULL;
	if (!dst) return NULL;

	for (p = src; (isspace(*p) && (*p)); p++);
	strcpy(dst, p);
	for (p = dst + strlen(dst) - 1;
	     (isspace(*p) && (p > dst));
	     *p = '\0', p--);
	if (isspace(*p)) *p = '\0';

	return dst;
}

/*
 * strip comment
 */
char		*oppd_strstrip(
	char		*dst,
	char		*src
	)
{
	char		*p;
	char		*buf;
	char		*result;

	if (!src) return NULL;
	if (!dst) return NULL;

	if (!(buf = malloc(strlen(src) + 1))) return NULL;
	strcpy(buf, src);
	if ((p = index(buf, '#'))) *p = '\0';
	result = oppd_strtrim(dst, buf);
	free(buf);

	return result;
}

/*
 * get token
 */
char		*oppd_strtoken(
	char		*tok,
	char		*src,
	char		*del
	)
{
	char		*p;
	char		*buf;

	if (!src) return NULL;
	if (!tok) return NULL;
	if (!del) return NULL;

	if (!(buf = malloc(strlen(src) + 1))) return NULL;
	strcpy(buf, src);
	if ((p = strpbrk(buf, del))) *p = '\0';
	strcpy(tok, buf);
	free(buf);

	return buf;
}

/*
 * get next token
 */
char		*oppd_strnext(
	char		*src,
	char		*del
	)
{
	char		*p;

	if (!src) return NULL;
	if (!del) return NULL;

	if ((p = strpbrk(src, del))) p++;

	return p;
}

/*
 * get service port number
 */
int		oppd_get_port(
	char		*name,
	char		*proto,
	int		def_port
	)
{
struct	servent		*ent;
	int		port = def_port;

	ent = getservbyname(name, proto);
	if (ent) {
		port = ent->s_port;
	}

	return port;
}

/*
 * get spool directory
 */
char		*oppd_get_spool_dir(
	char		*name
	)
{
static	char		buf[OPPD_BUF_LEN];
	struct	stat	st;

	buf[0] = '\0';
	strcpy(buf, OPPD_SPL_DIR);
	strcat(buf, name);
	if (lstat(buf, &st) < 0) {
		if (mkdir(buf, 0777)) buf[0] = '\0';
	} else {
		if (!S_ISDIR(st.st_mode)) buf[0] = '\0';
	}
	if (!strlen(buf)) strcpy(buf, "/var/spool");

	return buf;
}

/*
 * open spool file
 */
int		oppd_open_spool(
	char		*name,
	char		**file
	)
{
static	char		buf[OPPD_BUF_LEN];
	int		fd;

	strcpy(buf, oppd_get_spool_dir(name));
	strcat(buf, "/XXXXXX");

	if ((fd = mkstemp(buf)) < 0) {
		*file = NULL;
	} else {
		*file = buf;
	}

	return fd;
}

/*
 * add job
 */
int		oppd_add_job(
	char		*name,
	oppd_job	*job
	)
{
static	char		buf[OPPD_BUF_LEN];
	FILE		*fp = NULL;
	int		ecode = 0;

	strcpy(buf, oppd_get_spool_dir(name));
	strcat(buf, "/");
	strcat(buf, OPPD_JOB_FILE);

	if (!(fp = fopen(buf, "a"))) ecode = -1;

	if (!ecode) {
		fprintf(fp, "%s\n", job->spool);
	}

	if (fp) fclose(fp);

	return ecode;
}

/*
 * get job
 */
oppd_job	*oppd_get_job(
	char		*name
	)
{
static	oppd_job	job;
static	char		buf[OPPD_BUF_LEN];
	FILE		*fp = NULL;
	int		ecode = 0;
	int		pos;

	job.id = 0;
	job.copies = 1;
	job.title = NULL;
	job.owner = NULL;
	job.spool = NULL;

	strcpy(buf, oppd_get_spool_dir(name));
	strcat(buf, "/");
	strcat(buf, OPPD_JOB_FILE);

	if (!(fp = fopen(buf, "r"))) ecode = -1;

	if (!ecode) {
		if (fgets(buf, OPPD_BUF_LEN - 1, fp)) {
			pos = strlen(buf) - 1;
			while (iscntrl(buf[pos])) buf[pos--] = '\0';
			job.spool = buf;
		}
	}

	if (fp) fclose(fp);

	if (job.spool) return &job;
	return NULL;
}

/*
 * delete job
 */
int		oppd_del_job(
	char		*name,
	oppd_job	*job
	)
{
static	char		buf[OPPD_BUF_LEN];
static	char		sf[OPPD_BUF_LEN];
static	char		df[OPPD_BUF_LEN];
	FILE		*src = NULL;
	FILE		*dst = NULL;
	int		ecode = 0;

	strcpy(sf, oppd_get_spool_dir(name));
	strcat(sf, "/");
	strcat(sf, OPPD_JOB_FILE);

	if (!(src = fopen(sf, "r"))) ecode = -1;

	if (!ecode) {
		strcpy(df, sf);
		strcat(df, ".$$$");
		if (!(dst = fopen(df, "w+"))) ecode = -1;
	}

	if (!ecode) {
		while (fgets(buf, OPPD_BUF_LEN - 1, src)) {
			if (strncmp(job->spool, buf, strlen(job->spool))) {
				fprintf(dst, "%s", buf);
			}
		}
	}

	if (src) fclose(src);
	if (dst) fclose(dst);

	if (ecode) {
		unlink(df);
	} else {
		unlink(sf);
		rename(df, sf);
		unlink(job->spool);
	}

	return ecode;
}

