/*
 *  OpenPrinting Vector Printer Driver for Canon Pixus850i
 *  Copyright (c) 2004 E&D Inc.
 *  All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * NOTE:
 *  - As a special exception, this program is permissible to link with the
 *    libraries released as the binary modules.
 *  - If you write modifications of your own for these programs, it is your
 *    choice whether to permit this exception to apply to your modifications.
 *    If you do not wish that, delete this exception.
*/

/* $Revision: 1.5 $ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <dlfcn.h>

#include "bjp_context.h"
#include "bjp_draw.h"


#define BJFILTERCONFPATH	"/usr/lib/bjlib/bjfilter"
#define BJFILTERCONFEXT		".conf"
#define BJLIBPATH		"/usr/lib/bjlib"
#define BJPCONFERROR		-256
#define OPTSTRMODEL		"model"
#define OPTSTRPSIZE		"papersize"
#define CNCL_PAPERSIZE		8

#define BJP_DEFAULT_MEDIASIZE_NAME	BJP_PAPERSIZE_A4
#define BJP_DEFAULT_RESOLUTION_X	600
#define BJP_DEFAULT_RESOLUTION_Y	600

#define PSID	0
#define	YOKO	1
#define	TATE	2

const long PaperTbl[][3] = 
{
	{0, 14800, 21000},		/* A5 (148mm x 210mm) */
	{1, 21000, 14800},		/* A5-L (210mm x 148mm) */
	{2, 21000, 29700},		/* A4 (210mm x 297mm) */
	{3, 29700, 21000},		/* A4-L (297mm x 210mm) */
	{4, 29700, 42000},		/* A3 (297mm x 420mm) */
	{5, 42000, 29700},		/* A3-L (420mm x 297mm) */
	{6, 42000, 59400},		/* A2 (420mm x 594mm) */
	{7, 18200, 25700},		/* B5 (182mm x 257mm) */
	{8, 25700, 18200},		/* B5-L (257mm x 182mm) */
	{9, 25700, 36400},		/* B4 (257mm x 364mm) */
	{10, 36400, 25700},		/* B4-L (364mm x 257mm) */
	{11, 36400, 51500},		/* B3 (364mm x 515mm) */
	{12, 21590, 27940},		/* Letter (8.5x 11inch) */
	{13, 27940, 21590},		/* Letter-L (11inch x 8.5inch) */
	{14, 21590, 35560},		/* Legal (8.5inch x 14inch) */
	{15, 35560, 21590},		/* Legal-L (14inch x 8.5inch) */
	{16, 27940, 43180},		/* Ledger  (11inch x 17inch) */
	{17, 43180, 27940},		/* Ledger-L (17inch x 11inch) */
	{18, 43180, 55800},		/* C size (17 x 22inch) */
	{19, 10000, 14800},		/* Postcard (10mm 0x 148mm) */
	{20, 14800, 10000},		/* Postcard-L (14mm 8x 100mm) */
	{21, 24130, 10478},		/* #10 Envelope (9.5inch x 4.125inch) */
	{22, 22000, 11000},		/* DL Envelope (220mm x 110mm) */
	{23, 34290, 43180},		/* Foolscap (13.5inch x 17inch) */
	{24, 21590, 27940},		/* Fanfold (8.5inch x 11inch) */
	{25, 21590, 30480},		/* Fanfold (9.5inch x 11inch) */
	{26, 24130, 27940},		/* Fanfold (10inch x 11inch) */
	{27, 25400, 27940},		/* Fanfold (10inch x 12inch) */
	{28, 25400, 30480},		/* Fanfold (12inch x 8.5inch) */
	{29, 30480, 21590},		/* Fanfold (14.5inch x 11inch) */
	{30, 36830, 27940},		/* Fanfold (15inch x 11inch) */
	{31, 38100, 27940},		/* Fanfold (15inch x 12inch) */
	{32, 38100, 30480},		/* Fanfold (14.875inch x 11inch) */
	{33, 23500, 10500},		/* Envelope-J4 */
	{34, 19000,  9800},		/* Envelope-J6 */
	{35, 22860, 33782},		/* Letter full-bleed */
	{36, 22350, 35560},		/* A4 full-bleed */
	{37, 32900, 48300},		/* A3 plus */
	{38, 11860, 21390},		/* Photocard (FM-101) */
	{39, 0,0},				/* User define */
	{40, 0,0},				/* Not Applicable */
	{1046, 10478, 24130},	/* #10 Envelope Tate (4.125inch x 9.5inch) */
	{1047, 11000, 22000},	/* DL Envelope Tate ( 110mm x 220mm ) */
	{1048, 10500, 23500},	/* Envelope-J4 Tate ( 105mm x 235mm ) */
	{1049,  9800, 19000},	/* Envelope-J6 Tate (  98mm x 190mm ) */
	{1050,  8900, 12700},	/* Actual Size L ( 89mm x 127mm ) */
	{1051, 12700, 17800},	/* Actual Size 2L ( 127mm x 178mm ) */
	{1052, 10160, 15240},	/* Actual Size 4x6 ( 101.6mm x 152.4mm ) */
	{1056,  8900, 25400},	/* Panorama or L x 2 ( 89mm x 254mm ) */
	{-1, -1, -1}			/* Max indicator */	
};

/*-------------------------------------------------------------*/

static short
bjp_GetPaperSize(short id,
		 long *width,
		 long *height )
{
	short paper = 0;
	short result = -1;

	while((id != PaperTbl[paper][PSID]) && (PaperTbl[paper][PSID] != -1))
		paper++;

	if( PaperTbl[paper][PSID] == -1 ) goto onErr;

	*width  = PaperTbl[paper][YOKO];
	*height = PaperTbl[paper][TATE];

	result = 0;
onErr:
	return result;	
}

static char *
bjp_GetModelname(char *model,
		 char *modelname,
		 int len)
{
	char *rv = NULL;
	if(model && strlen(model) < len){
		char *p = model, *q = modelname;
		while((*q++ = toupper(*p++)) != '\0');
		rv = modelname;
	}
	return(rv);
}

static char *
bjp_GetConfname(char *model,
		char *confName,
		int len)
{
	char *rv = NULL;
	char smallname[32];
	int  slen = strlen(BJFILTERCONFPATH) + strlen(BJFILTERCONFEXT) + 1;

	if(model && strlen(model)+slen < len){
		char *p = model, *q = smallname;

		while((*q++ = tolower(*p++)) != '\0');
		sprintf(confName, "%s%s%s", BJFILTERCONFPATH, smallname, BJFILTERCONFEXT);
		rv = confName;
	}
	return(rv);
}

static BJPContext *
bjp_InitValueByCNCLDB(char	*modelname,
		      char	*confname,
		      BJPControlInfo *ctrlInfo)
{
	int	id;
	short	gamma = 0;
	short	ret;
	short	dbsize;
	short	model_id;
	char	bjlibdir[] = BJLIBPATH;
	LPCNCLDB	db = NULL;
	CNCLNAMEINFO	name_info;
	BJPContext	*context = NULL;
	BJFLTDEVICE	*dev;
	CNCLPAPERSIZE	*psize;
	BJFLTCOLORSYSTEM	*color;
	BJFLTCOMSYSTEM		*comsys;
	CNCLErr	err;
	void *hndl = NULL;
	int (*func)();
	char *errStr;

	if((id = bjp_GetConfID( confname, OPTSTRMODEL, modelname )) == BJPCONFERROR ){
		fprintf(stderr, "Error: invalid printer model name\n");
		goto onError;
	}

	/*
	 * DYNAMIC LINK
	 */
	hndl = dlopen("libcnbpcnclui156.so", RTLD_NOW);
	if(!hndl) goto onError;

	/* Retrieve Name Information */
	func = dlsym(hndl, "CNCL_GetDBName");
	if((errStr = dlerror()) != NULL){
		goto onError;
	}
	if( func( &name_info, (void *) &bjlibdir, id ) < 0){
		fprintf(stderr, "Error: Internal Database Name Error\n");
		goto onError;
	}

	/* Allocate Internal DB Area */
	func = dlsym(hndl, "CNCL_GetDBSize");
	if((errStr = dlerror()) != NULL){
		goto onError;
	}
	dbsize = func( &name_info, (void *)bjlibdir);
	db = (LPCNCLDB)malloc(sizeof(CNCLDB) * dbsize);
	model_id = id;

	/* Check errors. */
	func = dlsym(hndl, "CNCL_GetEntries");
	if((errStr = dlerror()) != NULL){
		goto onError;
	}
	if( func( &name_info, (void *)bjlibdir, db) < 0){
		fprintf(stderr, "Error: CNCL_GetEntries\n");
		goto onError;
	}

	func = dlsym(hndl, "CNCL_GetDefaults");
	if((errStr = dlerror()) != NULL){
		goto onError;
	}
	if( func( &name_info, (void *)bjlibdir, db, dbsize) < 0){
		fprintf(stderr, "Error: CNCL_GetDefaults\n");
		goto onError;
	}

	func = dlsym(hndl, "CNCL_InitMenulink");
	if((errStr = dlerror()) != NULL){
		goto onError;
	}
	if( func( &name_info, (void *)bjlibdir, db, dbsize) < 0){
		fprintf(stderr, "Error: CNCL_InitMenulink\n");
		goto onError;
	}

	/*
	 * make BJPContext and set values
	 */
	context = (BJPContext*)calloc(1, sizeof(BJPContext));
	dev    = &(context->device);
	psize  = &(context->papersize);
	color  = &(context->color);
	comsys = &(context->comsystem);

	context->NumCopies = 1;
	strcpy(context->MediaSizeName,BJP_DEFAULT_MEDIASIZE_NAME);
	context->resolutionX = BJP_DEFAULT_RESOLUTION_X;
	context->resolutionY = BJP_DEFAULT_RESOLUTION_Y;

	if(ctrlInfo){
		if(ctrlInfo->NumCopies > 0)
			context->NumCopies   = ctrlInfo->NumCopies;
		if(ctrlInfo->MediaSizeName[0] != '\0')
			strcpy(context->MediaSizeName,ctrlInfo->MediaSizeName);
		if(ctrlInfo->resolutionX > 0)
			context->resolutionX = ctrlInfo->resolutionX;
		if(ctrlInfo->resolutionY > 0)
			context->resolutionY = ctrlInfo->resolutionY;
	}

	/*
	 * Initialize
	 */
	color->bjfltIntent = 0;
	color->bjfltGamma = BJP_GAMMA_14;
	if((gamma = bjp_GetCurrentnValue(db, dbsize, BJPID_INPUT_GAMMA)) > 0){
		color->bjfltGamma = gamma;
	}
	
	color->bjfltBalanceC = 0;
	color->bjfltBalanceM = 0;
	color->bjfltBalanceY = 0;
	color->bjfltBalanceK = 0;
	color->bjfltDensity = 0;

	/* PaperSize */
	if(context->MediaSizeName){
		if((id = bjp_GetConfID( confname, OPTSTRPSIZE, context->MediaSizeName )) == BJPCONFERROR ){
			fprintf(stderr, "Error: invalid papersize\n");
			goto onError;
		}
		if(bjp_QueryValue(db, dbsize, CNCL_PAPERSIZE, id) >= 0){
			bjp_SetTemporaryFlag(db, dbsize, CNCL_PAPERSIZE, id, 1);

			func = dlsym(hndl, "CNCL_GetMenulink");
			if((errStr = dlerror()) != NULL){
				goto onError;
			}
			func( &name_info, (void *)bjlibdir, db, dbsize);
		}else{
			fprintf(stderr, "Error: inappropriate papersize selection\n");
			goto onError;
		}
	}

	dev->bjfltPaperSize = bjp_GetCurrentnValue(db, dbsize, BJPID_PAPERSIZE);

	if((ret = bjp_GetPaperSize( dev->bjfltPaperSize, &dev->bjfltPaperWidth, &dev->bjfltPaperHeight)) < 0)
		goto onError;

	func = dlsym(hndl, "CNCL_GetMargin");
	if((errStr = dlerror()) != NULL){
		goto onError;
	}
	if((ret = func( &name_info, (void *)bjlibdir, db, psize, dbsize)) < 0){
		if(ret < -2){/* height out of range. */
			fprintf(stderr, "Error: invalid paperheight value. (integer: %5ld - %5ld [1/100mm])\n",
					psize->nMinCustomLength, psize->nMaxCustomLength);
			ret += 2;
		}
		if(ret == -2){/* width out of range. */
			fprintf(stderr, "Error: invalid paperwidth value.  (integer: %5ld - %5ld [1/100mm])\n",
					psize->nMinCustomWidth, psize->nMaxCustomWidth);
		}
		goto onErrorInappropriate;
	}

#if 0
	fprintf(stderr,"--- Paper size ---\n");
	fprintf(stderr,"nMinCustomWidth=%ld\n", psize->nMinCustomWidth);
	fprintf(stderr,"nMaxCustomWidth=%ld\n", psize->nMaxCustomWidth);
	fprintf(stderr,"nMinCustomLength=%ld\n", psize->nMinCustomLength);
	fprintf(stderr,"nMaxCustomLength=%ld\n", psize->nMaxCustomLength);
	fprintf(stderr,"nSelPaperWidth=%ld\n", psize->nSelPaperWidth);
	fprintf(stderr,"nSelPaperLength=%ld\n", psize->nSelPaperLength);
	fprintf(stderr,"nTopMargin=%ld\n", psize->nTopMargin);
	fprintf(stderr,"nLeftMargin=%ld\n", psize->nLeftMargin);
	fprintf(stderr,"nWidthPrintArea=%ld\n", psize->nWidthPrintArea);
	fprintf(stderr,"nLengthPrintArea=%ld\n", psize->nLengthPrintArea);
#endif

	/* Prepare bjdevice structure, to handover main routine */

	dev->bjfltModelID = model_id;
	dev->bjfltMediaType = bjp_GetCurrentnValue(db, dbsize, BJPID_MEDIATYPE);
	dev->bjfltPrintQuality = bjp_GetCurrentnValue(db, dbsize, BJPID_PRINTQUALITY);
	dev->bjfltMediaSupply = bjp_GetCurrentnValue(db, dbsize, BJPID_MEDIASUPPLY);
	dev->bjfltPaperGap = bjp_GetCurrentnValue(db, dbsize, BJPID_PAPERGAP_COMMAND);
	if( dev->bjfltPaperGap == -1 )
		dev->bjfltPaperGap = BJP_PGAP_CMD_NA;

	dev->bjfltBinMethod = bjp_GetCurrentnValue(db, dbsize, BJPID_DITHER_PAT);
	if( dev->bjfltBinMethod == BJP_UIBIN_PATTERN_HS  )
		dev->bjfltBinMethod = BJP_UIBIN_PATTERN;
		
	dev->bjfltBanner = bjp_GetCurrentnValue(db, dbsize, BJPID_BANNER);
	dev->bjfltInkType = bjp_GetCurrentnValue(db, dbsize, BJPID_CARTRIDGE);
	dev->bjfltGrayScale = bjp_GetCurrentnValue(db, dbsize, BJPID_GRAYSCALE);

	dev->bjfltMarginType = bjp_GetCurrentnValue(db, dbsize, BJPID_MARGINTYPE);
	if( dev->bjfltMarginType == -1 )
		dev->bjfltMarginType = BJP_MARGIN_NORMAL;

	/*
	 * get ComSystem
	 */
	func = dlsym(hndl, "CNCL_GetComSystem");
	if((errStr = dlerror()) != NULL){
		goto onError;
	}
	err = func( &name_info, (void *)bjlibdir, dev, comsys);
	if(err < 0) goto onError;

Exit:
	if(db) free(db);
	if(hndl) dlclose(hndl);

	return context;

onError:
	ret = -1;
	goto Exit;

onErrorInappropriate:
	ret = -2;
	goto Exit;
}

/*----------------------------------------------------------------------
 * public functions
 *----------------------------------------------------------------------*/

BJPContext *
bjp_NewContext(char	*model,
	       int	outfd,
	       BJPControlInfo *ctrlInfo)
{
	char modelname[40];
	char confname[256];
	char *p;
	BJPContext *context;
	CNCLData   *cnclData;
	CNCLErr	   cnclerr = -1;
	BJFLTDEVICE	*dev;
	BJFLTCOLORSYSTEM *color;
	BJFLTCOMSYSTEM	*comsys;
	CNCLPAPERSIZE	*psize;
	void *hndl = NULL;
	int (*func)();
	char *errStr;

	p = bjp_GetModelname(model, modelname, 40);
	if(!p) goto err1;
	p = bjp_GetConfname(model, confname, 256);
	if(!p) goto err1;

	context = bjp_InitValueByCNCLDB(modelname, confname, ctrlInfo);
	if(!context) goto err1;

	context->outfd = outfd;

	/*
	 * DYNAMIC LINK
	 */
	hndl = dlopen("libcnbpcnclapi156.so", RTLD_NOW);
	if(!hndl) goto err1;

	context->dlHndl = hndl;

	cnclData = (CNCLData *)calloc(1, sizeof(CNCLData));
	if(!cnclData) goto err2;

	func = dlsym(hndl, "CNCL_Init");
	if((errStr = dlerror()) != NULL){
		goto err2;
	}
	if((cnclerr = func(cnclData)) != BJPID_OK) goto err2;
	cnclData->Work1Buf = (CPKVoid CPKPTR)malloc(cnclData->Work1Size);
	if(!cnclData->Work1Buf) goto err2;
	memset((CPKVoid CPKPTR)cnclData->Work1Buf, 0, cnclData->Work1Size);
	context->cnclData = cnclData;

	dev = &(context->device);
	comsys = &(context->comsystem);
	psize = &(context->papersize);
	color = &(context->color);

	cnclData->InkType      = comsys->bjfltInkCartridge;
	cnclData->MediaType    = dev->bjfltMediaType;
	cnclData->PaperSize    = dev->bjfltPaperSize;
	cnclData->MediaSupply  = dev->bjfltMediaSupply;
	cnclData->PaperGap     = dev->bjfltPaperGap;

	switch( dev->bjfltBinMethod ){
	case BJP_UIBIN_PATTERN:
	case BJP_UIBIN_PATTERN_HS:
	case BJP_UIBIN_V_PATTERN:
		cnclData->BinMethod = BJP_BIN_DITHER;
		break;
	case BJP_UIBIN_ED:
		cnclData->BinMethod = BJP_BIN_ED;
		break;
	case BJP_UIBIN_NA:
		cnclData->BinMethod = BJP_BIN_NA;
		break;
	case BJP_UIBIN_MAX:
		cnclData->BinMethod = BJP_BIN_MAX;
		break;
	}
	
	cnclData->Grayscale	= dev->bjfltGrayScale;
	cnclData->PrintQuality	= dev->bjfltPrintQuality;
	cnclData->Banner	= dev->bjfltBanner;
	cnclData->PaperWidth	= psize->nSelPaperWidth;
	cnclData->PaperHeight	= psize->nSelPaperLength;
	cnclData->Smoothing	= BJP_SMOOTH_NA ;
	cnclData->MarginType	= dev->bjfltMarginType;
	
	cnclData->Intent	= color->bjfltIntent;
	cnclData->Gamma		= color->bjfltGamma;
	cnclData->BalanceC	= color->bjfltBalanceC;
	cnclData->BalanceM	= color->bjfltBalanceM;
	cnclData->BalanceY 	= color->bjfltBalanceY;
	cnclData->BalanceK	= color->bjfltBalanceK;
	cnclData->Density	= color->bjfltDensity;
	
	return(context);

 err2:
	if(cnclData)
		free(cnclData);
 err1:
	return(NULL);
}

void
bjp_FreeContext(BJPContext *context)
{
	CNCLData *cnclData = context->cnclData;

	if(!context)
		return;
	if(cnclData){
		if(cnclData->Work1Buf)
			free(cnclData->Work1Buf);
		free(cnclData);
	}
	free(context);
}

int
bjp_StartJob(BJPContext *context)
{
	time_t	_time;
	struct tm  _tm;
	int rv = -1;
	CNCLData *cnclData = context->cnclData;
	CNCLErr  cnclerr = -1;
	int (*func)();
	char *errStr;

	cnclData->ModelID = context->device.bjfltModelID;

	time(&_time);
	localtime_r(&_time, &_tm);
	cnclData->CurrentTime.TimeInfo	= BJP_TIME_CORRECT;
	cnclData->CurrentTime.Year	= _tm.tm_year + 1900;
	cnclData->CurrentTime.Month	= _tm.tm_mon + 1;
	cnclData->CurrentTime.Day	= _tm.tm_mday;
	cnclData->CurrentTime.Hour	= _tm.tm_hour;
	cnclData->CurrentTime.Min	= _tm.tm_min;
	cnclData->CurrentTime.Sec	= _tm.tm_sec;

	func = dlsym(context->dlHndl, "CNCL_StartJob");
	if((errStr = dlerror()) != NULL){
		goto finish1;
	}
	if((cnclerr = func(cnclData)) != BJPID_OK)
		goto finish1;

	bjp_OutCmd(cnclData->outputBuffer, cnclData->outputSize, context->outfd);
	rv = 0;

 finish1:
	return(rv);
}

int
bjp_EndJob(BJPContext *context)
{
	CNCLData *cnclData = context->cnclData;
	CNCLErr  cnclerr = -1;
	int (*func)();
	char *errStr;

	func = dlsym(context->dlHndl, "CNCL_EndJob");
	if((errStr = dlerror()) != NULL){
		goto finish1;
	}
	if((cnclerr = func(cnclData)) == BJPID_OK){
		bjp_OutCmd(cnclData->outputBuffer, cnclData->outputSize, context->outfd);
	}
 finish1:
	return(cnclerr);
}

void
bjp_SetControlInfo(BJPContext *context,
		   BJPControlInfo *ctrlInfo)
{
	if(ctrlInfo){
		if(ctrlInfo->NumCopies > 0)
			context->NumCopies   = ctrlInfo->NumCopies;
		if(ctrlInfo->MediaSizeName[0] != '\0')
			strcpy(context->MediaSizeName,ctrlInfo->MediaSizeName);
		if(ctrlInfo->resolutionX > 0)
			context->resolutionX = ctrlInfo->resolutionX;
		if(ctrlInfo->resolutionY > 0)
			context->resolutionY = ctrlInfo->resolutionY;
	}
}
