/*
 *  OpenPrinting Vector Printer Driver for Canon Pixus850i
 *  Copyright (c) 2004 E&D Inc.
 *  All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * NOTE:
 *  - As a special exception, this program is permissible to link with the
 *    libraries released as the binary modules.
 *  - If you write modifications of your own for these programs, it is your
 *    choice whether to permit this exception to apply to your modifications.
 *    If you do not wish that, delete this exception.
*/

/* $Revision: 1.4 $ */

#ifndef _BJPCONTEXT_H
#define _BJPCONTEXT_H

#include "cncl.h"
#include "cnclucom.h"
#include "bjp_defs.h"
#include "bjp_conf.h"
#include "bjp_cncldb.h"

#define BJP_PAPERSIZE_A5	"a5"
#define BJP_PAPERSIZE_A4	"a4"
#define BJP_PAPERSIZE_A3	"a3"
#define BJP_PAPERSIZE_B5	"b5"
#define BJP_PAPERSIZE_B4	"b4"
#define BJP_PAPERSIZE_B3	"b3"
#define BJP_PAPERSIZE_LETTER	"letter"
#define BJP_PAPERSIZE_LEGAL	"legal"
#define BJP_PAPERSIZE_POSTCARD	"postcard"
#define BJP_PAPERSIZE_PHOTOCARD	"photocard"

typedef struct _BJPPos
{
	long	img_width;	/* original image width */
	long	img_height;	/* original image height */
	double	xresolution;	/* original iamge x resolution */
	double	yresolution;	/* original image y resolution */

	long	topmargin;	/* printer top margin */
	long	leftmargin;	/* printer left margin */
	double  prn_xresolution;/* printer x resolution */
	double	prn_yresolution;/* printer y resolution */
	long	prn_width;	/* printer available width */
	long	prn_height;	/* printer available height */

	double	mask_width;	/* actual image width */
	double	mask_height;	/* actual image height */
	double	width_offset;	/* x offset when output pixel is over
				   printer available width */
	double	height_offset;	/* y offset when output pixel is over
				   printer available heigth */
	double	out_width;	/* output width */
	double	out_height;	/* otuput height */
	double	leftskip;	/* left skip ( pixel ) */
	double	topskip;	/* top skip ( pixel ) */
	double	page_width;	/* page width ( pixel ) */
	double	scaling;	/* scaling parameter */

	double	curtopmargin;	/* printer top margin [1/100mm] */
	double	curleftmargin;	/* printer left margin [1/100mm] */
	double	currightmargin;	/* printer top margin [1/100mm] */
	double	curbottommargin;/* printer left margin [1/100mm] */
	double	curprn_width;	/* printer available width [1/100mm] */
	double	curprn_height;	/* printer available height [1/100mm] */
	long	PaperWidth;
	long	PaperLength;
	long	bbox[4];
	short	extension;
} BJPPos;

typedef struct _BJImageContext {
	int	width;
	int	height;
	int	depth;
	int	rassize;	/* raster size */
	unsigned char *rasbuf;	/* raster buffer */
	int	xpos;		/* raster position */
	int	ypos;
} BJImageContext;

typedef struct {
	int	NumCopies;
	char	MediaSizeName[32];
	int	resolutionX;
	int	resolutionY;
} BJPControlInfo;

typedef struct _BJPContext
{
	int	outfd;
	int	NumCopies;
	char	MediaSizeName[32];
	int	resolutionX;
	int	resolutionY;
	void	*dlHndl;
	BJFLTDEVICE		device;
	CNCLPAPERSIZE   	papersize;
	BJFLTCOLORSYSTEM	color;
	CNCLData		*cnclData;
	BJFLTCOMSYSTEM		comsystem;
	long	width;
	long	height;
	long	width_offset;
	long	height_offset;
	long	ImageWidth;
	long	ImageHeight;
	long	leftskip;
	long	topskip;
	long	Rest;
	long	TotalRest;
	long	PrevPos;
	long	CurrPos;
	long	Quotient;
	long	page_width;
	short	bpp;
	CPKByte CPKPTR	srcBuf;
	CPKByte CPKPTR	rgbBuf;
	BJPPos	bjppos;
	BJImageContext	img_context;
	int	rasterdrawing;
} BJPContext;

BJPContext *bjp_NewContext(char *, int, BJPControlInfo *);
void bjp_FreeContext(BJPContext *);
int bjp_StartJob(BJPContext *);
int bjp_EndJob(BJPContext *);
void bjp_SetControlInfo(BJPContext *, BJPControlInfo *);

#endif /* _BJPCONTEXT_H */

