/* Copyright (c) 2004 E-D Inc.  All rights reserved.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
/*******************************************************************

  OpenPrinting Vector Printer Driver API : for Canon

 ==================================================================

  API LIPS Functions : Attribute Set / Get

********************************************************************/
/* system include files */
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>

/* local include files */
#include "pdapi.h"
#include "pd_print.h"
#include "pd_lips.h"

/*==================================================================
  API Functions
===================================================================*/
/* ----------------------------------------------------------------
 *
 *	LIPS_ResetCTM
 *
 * ---------------------------------------------------------------- */
int LIPS_ResetCTM( int printerContext )
{
	pd_GraphicsState	*gs;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	gs = pd_GetGraphicsState( printerContext );
	if( !gs )	return( pdERROR );

	pd_InitCTM( &gs->ctm );
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_SetCTM
 *
 * ---------------------------------------------------------------- */
int LIPS_SetCTM( int printerContext, pdCTM *pCTM )
{
	pd_GraphicsState	*gs;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	gs = pd_GetGraphicsState( printerContext );
	if( !gs )	return( pdERROR );

	gs->ctm.a = pCTM->a;	gs->ctm.b = pCTM->b;
	gs->ctm.c = pCTM->c;	gs->ctm.d = pCTM->d;
	gs->ctm.e = pCTM->e;	gs->ctm.f = pCTM->f;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_GetCTM
 *
 * ---------------------------------------------------------------- */
int LIPS_GetCTM( int printerContext, pdCTM *pCTM )
{
	pd_GraphicsState	*gs;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	gs = pd_GetGraphicsState( printerContext );
	if( !gs )	return( pdERROR );

	pCTM->a = gs->ctm.a;	pCTM->b = gs->ctm.b;
	pCTM->c = gs->ctm.c;	pCTM->d = gs->ctm.d;
	pCTM->e = gs->ctm.e;	pCTM->f = gs->ctm.f;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_InitGS
 *
 * ---------------------------------------------------------------- */
int LIPS_InitGS( int printerContext )
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	pd_ClearGraphicsState( gs );
	pd_InitGraphicsState( gs );
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_SaveGS
 *
 * ---------------------------------------------------------------- */
int LIPS_SaveGS( int printerContext )
{
	pd_PrinterInfo		*pri;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	if( !pd_PushGraphicsState( pri ) )	return( pdERROR );
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_RestoreGS
 *
 * ---------------------------------------------------------------- */
int LIPS_RestoreGS( int printerContext )
{
	pd_PrinterInfo		*pri;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	if( !pd_PopGraphicsState( pri ) )	return( pdERROR );
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_QueryColorSpace
 *
 * ---------------------------------------------------------------- */
int LIPS_QueryColorSpace( int printerContext, pdColorSpace *pcspace, int *pnum )
{
	pd_PrinterInfo		*pri;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	if( pcspace ) {
		if( *pnum < 1 ) {
			errorno = pdPARAMERROR;
			return( pdERROR );
		}
		*pcspace = pdcspaceStandardRGB;
	}
	*pnum = 1;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_SetColorSpace
 *
 * ---------------------------------------------------------------- */
int LIPS_SetColorSpace( int printerContext, pdColorSpace cspace )
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	if( cspace != pdcspaceStandardRGB ) {
		pd_DebugPrint( "pdapi:cspace = %d \n", cspace );	// For Debug!
		errorno = pdPARAMERROR;
		return( pdERROR );
	}
	gs->colorSpace = cspace;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_GetColorSpace
 *
 * ---------------------------------------------------------------- */
int LIPS_GetColorSpace( int printerContext, pdColorSpace *pcspace )
{
	pd_GraphicsState	*gs;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	gs = pd_GetGraphicsState( printerContext );
	if( !gs )	return( pdERROR );

	*pcspace = gs->colorSpace;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_QueryROP
 *
 * ---------------------------------------------------------------- */
int LIPS_QueryROP( int printerContext, int *pnum, int *prop )
{
	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	*pnum = 0;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_SetROP
 *
 * ---------------------------------------------------------------- */
int LIPS_SetROP( int printerContext, int rop )
{
	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_GetROP
 *
 * ---------------------------------------------------------------- */
int LIPS_GetROP( int printerContext, int *prop )
{
	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_SetFillMode
 *
 * ---------------------------------------------------------------- */
int LIPS_SetFillMode( int printerContext, pdFillMode fillmode )
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	gs->fillMode = fillmode;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_GetFillMode
 *
 * ---------------------------------------------------------------- */
int LIPS_GetFillMode( int printerContext, pdFillMode *pfillmode )
{
	pd_GraphicsState	*gs;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	gs = pd_GetGraphicsState( printerContext );
	if( !gs )	return( pdERROR );

	*pfillmode = gs->fillMode;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_SetAlphaConstant
 *
 * ---------------------------------------------------------------- */
int LIPS_SetAlphaConstant( int printerContext, float alpha )
{
	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_GetAlphaConstant
 *
 * ---------------------------------------------------------------- */
int LIPS_GetAlphaConstant( int printerContext, float *palpha )
{
	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_SetLineWidth
 *
 * ---------------------------------------------------------------- */
int LIPS_SetLineWidth( int printerContext, pdFix width )
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	gs->lineWidth = width;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_GetLineWidth
 *
 * ---------------------------------------------------------------- */
int LIPS_GetLineWidth( int printerContext, pdFix *pwidth )
{
	pd_GraphicsState	*gs;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	gs = pd_GetGraphicsState( printerContext );
	if( !gs )	return( pdERROR );

	*pwidth = gs->lineWidth;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_SetLineDash
 *
 * ---------------------------------------------------------------- */
int LIPS_SetLineDash( int printerContext, pdFix pdash[], int num )
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;
	int		dn, n;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	if( num < 0 ) {
		errorno = pdPARAMERROR;
		return( pdERROR );
	}
	dn = num;
	if( num % 2 )	dn *= 2;
	if ( dn > pd_LINEDASH_ARYMAX ) {
		errorno = pdNOTSUPPORTED;
		return( pdERROR );
	}
	gs->dashNum = num;
	memset( gs->lineDash, 0, sizeof( pdFix ) * pd_LINEDASH_ARYMAX );
	for( n = 0; n < num; n++ ) {
		gs->lineDash[n] = pdash[n];
		if( num % 2 ) {
			gs->lineDash[ num + n ] = pdash[n];
		}
	}
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_GetLineDash
 *
 * ---------------------------------------------------------------- */
int LIPS_GetLineDash( int printerContext, pdFix pdash[], int *pnum )
{
	pd_GraphicsState	*gs;
	int		n;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	gs = pd_GetGraphicsState( printerContext );
	if( !gs )	return( pdERROR );

	if( pdash ) {
		if( *pnum < gs->dashNum ) {
			errorno = pdPARAMERROR;
			return( pdERROR );
		}
		for( n = 0; n < gs->dashNum; n++ ) {
			pdash[n] = gs->lineDash[n];
		}
	}
	*pnum = gs->dashNum;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_SetLineDashOffset
 *
 * ---------------------------------------------------------------- */
int LIPS_SetLineDashOffset( int printerContext, pdFix offset )
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	gs->dashOffset = offset;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_GetLineDashOffset
 *
 * ---------------------------------------------------------------- */
int LIPS_GetLineDashOffset( int printerContext, pdFix *poffset )
{
	pd_GraphicsState	*gs;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	gs = pd_GetGraphicsState( printerContext );
	if( !gs )	return( pdERROR );

	*poffset = gs->dashOffset;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_SetLineStyle
 *
 * ---------------------------------------------------------------- */
int LIPS_SetLineStyle( int printerContext, pdLineStyle linestyle )
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	gs->lineStyle = linestyle;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_GetLineStyle
 *
 * ---------------------------------------------------------------- */
int LIPS_GetLineStyle( int printerContext, pdLineStyle *plinestyle )
{
	pd_GraphicsState	*gs;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	gs = pd_GetGraphicsState( printerContext );
	if( !gs )	return( pdERROR );

	*plinestyle = gs->lineStyle;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_SetLineCap
 *
 * ---------------------------------------------------------------- */
int LIPS_SetLineCap( int printerContext, pdLineCap linecap )
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	gs->lineCap = linecap;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_GetLineCap
 *
 * ---------------------------------------------------------------- */
int LIPS_GetLineCap( int printerContext, pdLineCap *plinecap )
{
	pd_GraphicsState	*gs;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	gs = pd_GetGraphicsState( printerContext );
	if( !gs )	return( pdERROR );

	*plinecap = gs->lineCap;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_SetLineJoin
 *
 * ---------------------------------------------------------------- */
int LIPS_SetLineJoin( int printerContext, pdLineJoin linejoin )
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	gs->lineJoin = linejoin;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_GetLineJoin
 *
 * ---------------------------------------------------------------- */
int LIPS_GetLineJoin( int printerContext, pdLineJoin *plinejoin )
{
	pd_GraphicsState	*gs;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	gs = pd_GetGraphicsState( printerContext );
	if( !gs )	return( pdERROR );

	*plinejoin = gs->lineJoin;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_SetMiterLimit
 *
 * ---------------------------------------------------------------- */
int LIPS_SetMiterLimit( int printerContext, pdFix miterlimit )
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	gs->miterLimit = miterlimit;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_GetMiterLimit
 *
 * ---------------------------------------------------------------- */
int LIPS_GetMiterLimit( int printerContext, pdFix *pmiterlimit )
{
	pd_GraphicsState	*gs;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	gs = pd_GetGraphicsState( printerContext );
	if( !gs )	return( pdERROR );

	*pmiterlimit = gs->miterLimit;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_SetPaintMode
 *
 * ---------------------------------------------------------------- */
int LIPS_SetPaintMode( int printerContext, pdPaintMode paintmode )
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	gs->paintMode = paintmode;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_GetPaintMode
 *
 * ---------------------------------------------------------------- */
int LIPS_GetPaintMode( int printerContext, pdPaintMode *ppaintmode )
{
	pd_GraphicsState	*gs;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	gs = pd_GetGraphicsState( printerContext );
	if( !gs )	return( pdERROR );

	*ppaintmode = gs->paintMode;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_SetStrokeColor
 *
 * ---------------------------------------------------------------- */
int LIPS_SetStrokeColor( int printerContext, pdBrush *brush )
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	if( brush->colorSpace != pdcspaceStandardRGB ) {
		pd_DebugPrint( "pdapi:colorSpace = %d \n", brush->colorSpace );	// For Debug!
		errorno = pdPARAMERROR;
		return( pdERROR );
	}
	if( !pd_CopyBrush( &gs->strokeBrush, brush ) )	return( pdERROR );
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_SetFillColor
 *
 * ---------------------------------------------------------------- */
int LIPS_SetFillColor( int printerContext, pdBrush *brush )
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	if( brush->colorSpace != pdcspaceStandardRGB ) {
		pd_DebugPrint( "pdapi:colorSpace = %d \n", brush->colorSpace );	// For Debug!
		errorno = pdPARAMERROR;
		return( pdERROR );
	}
	if( !pd_CopyBrush( &gs->fillBrush, brush ) )	return( pdERROR );
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_SetBgColor
 *
 * ---------------------------------------------------------------- */
int LIPS_SetBgColor( int printerContext, pdBrush *brush )
{
	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );
	return( pdOK );
}
