/* Copyright (c) 2004 E-D Inc.  All rights reserved.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
/*******************************************************************

  OpenPrinting Vector Printer Driver API : for Canon

 ==================================================================

  API LIPS Functions : Except Vector Drawing

********************************************************************/
/* system include files */
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>

/* local include files */
#include "pdapi.h"
#include "pd_print.h"
#include "pd_lips.h"

/*==================================================================
  API Functions
===================================================================*/
/* ----------------------------------------------------------------
 *
 *	LIPS_DrawBitmapText
 *
 * ---------------------------------------------------------------- */
int LIPS_DrawBitmapText( int printerContext, int width, int height, int pitch, void *fontdata )
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;
	char	obuf[128];
	char	prm1[16], prm2[16], prm3[16], prm4[16];
	int		pt, n;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );
	pd_DebugPrint( "pdapi:width = %d height = %d pitch = %d \n", width, height, pitch );	// For Debug!

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	/* State Flag Check */
	if( !pri->pageStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* Data Out */
//	if( !pdLips_Out_PaintMode( pri ) )	return( FALSE );
	sprintf( obuf, "}H1%c", C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );

	pdLips_Prm_Color( gs->fillBrush.color, prm1 );
	sprintf( obuf, "}T%s%c", prm1, C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );

	pdLips_Param(    pri->curPntX, prm1 );
	pdLips_ParamAdd( pri->curPntY - height, prm1 );
	pdLips_Param(    pri->curPntX + width, prm2 );
	pdLips_ParamAdd( pri->curPntY, prm2 );
	pdLips_Param(    pri->curPntX, prm3 );
	pdLips_ParamAdd( pri->curPntY, prm3 );
	pdLips_Param(    height, prm4 );
	pdLips_ParamAdd( width, prm4 );
	sprintf( obuf, "}U%s%s%s%s1001%c", prm1, prm2, prm3, prm4, C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );

	pt = 0;
	if( width > 0 ) {
		pt = ( ( width - 1 ) / 8 ) + 1;
	}
	pdLips_Param( pt * height, prm1 );
	sprintf( obuf, "}Q110%s%c", prm1, C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );
	for( n = 0; n < height; n++ ) {
		if( !pd_DvWrite( pri, (char *)fontdata + ( pitch * n ), pt ) )	return( FALSE );
	}
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_DrawImage
 *
 * ---------------------------------------------------------------- */
int LIPS_DrawImage( int printerContext, int sourceWidth, int sourceHeight, int colorDepth, pdImageFormat imageFormat, pdRectangle destinationSize, int count, void *imageData )
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;
	char	obuf[128];
	char	prm1[16], prm2[16], prm3[16], prm4[16], prm5[16];
	int		wx, wy;
	int		wsz, dpt, pt, ln, n;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );
	pd_DebugPrint( "pdapi:width = %d height = %d depth = %d count = %d \n", sourceWidth, sourceHeight, colorDepth, count );	// For Debug!

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	/* State Flag Check */
	if( !pri->pageStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	if( pri->drwImgStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* Data Out */
	if( imageFormat != pdiformatRaw ) {
		errorno = pdNOTSUPPORTED;
		return( pdERROR );
	}
	if( colorDepth == 1 ) {
//		if( !pdLips_Out_PaintMode( pri ) )	return( FALSE );
		sprintf( obuf, "}H1%c", C_IS2 );
		if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );

		pdLips_Prm_Color( gs->fillBrush.color, prm1 );
		sprintf( obuf, "}T%s%c", prm1, C_IS2 );
		if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );
	}
	wx = pd_DtoI( Fixtof( destinationSize.p1.x ) - Fixtof( destinationSize.p0.x ) );
	if( wx < 0 )	wx *= -1;
	wy = pd_DtoI( Fixtof( destinationSize.p1.y ) - Fixtof( destinationSize.p0.y ) );
	if( wy < 0 )	wy *= -1;
	pdLips_Param(    pri->curPntX, prm1 );
	pdLips_ParamAdd( pri->curPntY, prm1 );
	pdLips_Param(    pri->curPntX + wx, prm2 );
	pdLips_ParamAdd( pri->curPntY + wy, prm2 );
	pdLips_Param(    pri->curPntX, prm3 );
	pdLips_ParamAdd( pri->curPntY + wy, prm3 );
	pdLips_Param(    sourceHeight, prm4 );
	pdLips_ParamAdd( sourceWidth, prm4 );
	if( colorDepth == 1 ) {
		sprintf( obuf, "}U%s%s%s%s1001%c", prm1, prm2, prm3, prm4, C_IS2 );
	} else if( colorDepth == 24 ) {
		pdLips_Param(     8, prm5 );
		pdLips_ParamAdd( 10, prm5 );
		sprintf( obuf, "}U%s%s%s%s%s11%c", prm1, prm2, prm3, prm4, prm5, C_IS2 );
	} else {
		errorno = pdNOTSUPPORTED;
		return( pdERROR );
	}
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );

	wsz = sourceWidth * colorDepth;
	dpt = pt = ln = 0;
	if( wsz > 0 ) {
		dpt = ( ( ( wsz - 1 ) / 32 ) + 1 ) * 4;
		pt = ( ( wsz - 1 ) / 8 ) + 1;
		ln = count / dpt;
	}
	pdLips_Param( pt * ln, prm1 );
	sprintf( obuf, "}Q110%s%c", prm1, C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );
	for( n = 0; n < ln; n++ ) {
		if( !pd_DvWrite( pri, (char *)imageData + ( dpt * n ), pt ) )	return( FALSE );
	}
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_StartDrawImage
 *
 * ---------------------------------------------------------------- */
int LIPS_StartDrawImage( int printerContext, int sourceWidth, int sourceHeight, int colorDepth, pdImageFormat imageFormat, pdRectangle destinationSize )
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;
	char	obuf[128];
	char	prm1[16], prm2[16], prm3[16], prm4[16], prm5[16];
	int		wx, wy;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );
	pd_DebugPrint( "pdapi:width = %d height = %d depth = %d \n", sourceWidth, sourceHeight, colorDepth );	// For Debug!

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	/* State Flag Check */
	if( !pri->pageStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	if( pri->drwImgStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* Data Out */
	if( imageFormat != pdiformatRaw ) {
		errorno = pdNOTSUPPORTED;
		return( pdERROR );
	}
	if( colorDepth == 1 ) {
//		if( !pdLips_Out_PaintMode( pri ) )	return( FALSE );
		sprintf( obuf, "}H1%c", C_IS2 );
		if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );

		pdLips_Prm_Color( gs->fillBrush.color, prm1 );
		sprintf( obuf, "}T%s%c", prm1, C_IS2 );
		if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );
	}
	wx = pd_DtoI( Fixtof( destinationSize.p1.x ) - Fixtof( destinationSize.p0.x ) );
	if( wx < 0 )	wx *= -1;
	wy = pd_DtoI( Fixtof( destinationSize.p1.y ) - Fixtof( destinationSize.p0.y ) );
	if( wy < 0 )	wy *= -1;
	pdLips_Param(    pri->curPntX, prm1 );
	pdLips_ParamAdd( pri->curPntY, prm1 );
	pdLips_Param(    pri->curPntX + wx, prm2 );
	pdLips_ParamAdd( pri->curPntY + wy, prm2 );
	pdLips_Param(    pri->curPntX, prm3 );
	pdLips_ParamAdd( pri->curPntY + wy, prm3 );
	pdLips_Param(    sourceHeight, prm4 );
	pdLips_ParamAdd( sourceWidth, prm4 );
	if( colorDepth == 1 ) {
		sprintf( obuf, "}U%s%s%s%s1000%c", prm1, prm2, prm3, prm4, C_IS2 );
	} else if( colorDepth == 24 ) {
		pdLips_Param(     8, prm5 );
		pdLips_ParamAdd( 10, prm5 );
		sprintf( obuf, "}U%s%s%s%s%s10%c", prm1, prm2, prm3, prm4, prm5, C_IS2 );
	} else {
		errorno = pdNOTSUPPORTED;
		return( pdERROR );
	}
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );

	pri->drwImgStarted = TRUE;
	pri->drwImgWidth = sourceWidth;
	pri->drwImgDepth = colorDepth;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_TransferDrawImage
 *
 * ---------------------------------------------------------------- */
int LIPS_TransferDrawImage( int printerContext, int count, void *imageData )
{
	pd_PrinterInfo		*pri;
	char	obuf[128];
	char	prm1[16];
	int		wsz, dpt, pt, ln, n;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );
	pd_DebugPrint( "pdapi:count = %d \n", count );	// For Debug!

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	/* State Flag Check */
	if( !pri->drwImgStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* Data Out */
	wsz = pri->drwImgWidth * pri->drwImgDepth;
	dpt = pt = ln = 0;
	if( wsz > 0 ) {
//		dpt = ( ( ( wsz - 1 ) / 32 ) + 1 ) * 4;
		pt = dpt = ( ( wsz - 1 ) / 8 ) + 1;
		ln = count / dpt;
	}
	pdLips_Param( pt * ln, prm1 );
	sprintf( obuf, "}Q100%s%c", prm1, C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );
	for( n = 0; n < ln; n++ ) {
		if( !pd_DvWrite( pri, (char *)imageData + ( dpt * n ), pt ) )	return( FALSE );
	}
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_EndDrawImage
 *
 * ---------------------------------------------------------------- */
int LIPS_EndDrawImage( int printerContext )
{
	pd_PrinterInfo		*pri;
	char	obuf[128];

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	/* State Flag Check */
	if( !pri->drwImgStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* Data Out */
	sprintf( obuf, "}Q1100%c", C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );

	pri->drwImgStarted = FALSE;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_StartScanline
 *
 * ---------------------------------------------------------------- */
int LIPS_StartScanline( int printerContext, int yposition )
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;
	char	obuf[128];
	char	prm1[16];

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	/* State Flag Check */
	if( !pri->pageStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	if( pri->scnLinStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* Data Out */
	sprintf( obuf, "F11%c", C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );
	sprintf( obuf, "E101%c", C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );
	sprintf( obuf, "}E0%c", C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );

	pdLips_Prm_Color( gs->strokeBrush.color, prm1 );
	sprintf( obuf, "G%s%c", prm1, C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );
	sprintf( obuf, "}G0%c", C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );

	pri->scnLinStarted = TRUE;
	pri->scnLinPntY = yposition;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_Scanline
 *
 * ---------------------------------------------------------------- */
int LIPS_Scanline( int printerContext, int nscanpairs, int *scanpairs )
{
	pd_PrinterInfo		*pri;
	char	obuf[128];
	char	prm1[16], prm2[16];
	int		n;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	/* State Flag Check */
	if( !pri->scnLinStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* Data Out */
	for( n = 0; n < nscanpairs; n++ ) {
		pdLips_Param( scanpairs[ n * 2 + 0 ], prm1 );
		pdLips_ParamAdd( pri->scnLinPntY, prm1 );
		pdLips_Param( scanpairs[ n * 2 + 1 ] - scanpairs[ n * 2 + 0 ], prm2 );
		pdLips_ParamAdd( 0, prm2 );
		sprintf( obuf, "1%s%s%c", prm1, prm2, C_IS2 );
		if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );
	}
	pri->scnLinPntY++;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_EndScanline
 *
 * ---------------------------------------------------------------- */
int LIPS_EndScanline( int printerContext )
{
	pd_PrinterInfo		*pri;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	/* State Flag Check */
	if( !pri->scnLinStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	pri->scnLinStarted = FALSE;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_StartRaster
 *
 * ---------------------------------------------------------------- */
int LIPS_StartRaster( int printerContext, int rasterWidth )
{
	pd_PrinterInfo		*pri;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	/* State Flag Check */
	if( !pri->pageStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	if( pri->rasterStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	pri->rasterStarted = TRUE;
	pri->rasterWidth = rasterWidth;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_TransferRasterData
 *
 * ---------------------------------------------------------------- */
int LIPS_TransferRasterData( int printerContext, int count, unsigned char *data )
{
	pd_PrinterInfo		*pri;
	char	obuf[128];
	char	prm1[16], prm2[16], prm3[16], prm4[16], prm5[16];

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	/* State Flag Check */
	if( !pri->rasterStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* Data Out */
	pdLips_Param(    pri->curPntX, prm1 );
	pdLips_ParamAdd( pri->curPntY, prm1 );
	pdLips_Param(    pri->curPntX + pri->rasterWidth, prm2 );
	pdLips_ParamAdd( pri->curPntY + 1, prm2 );
	pdLips_Param(    pri->curPntX, prm3 );
	pdLips_ParamAdd( pri->curPntY + 1, prm3 );
	pdLips_Param(    1, prm4 );
	pdLips_ParamAdd( pri->rasterWidth, prm4 );
	pdLips_Param(     8, prm5 );
	pdLips_ParamAdd( 10, prm5 );
	sprintf( obuf, "}U%s%s%s%s%s01%c", prm1, prm2, prm3, prm4, prm5, C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );

	pdLips_Param( count, prm1 );
	sprintf( obuf, "}Q110%s%c", prm1, C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );
	if( !pd_DvWrite( pri, data, count ) )	return( FALSE );

	pri->curPntY++;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_SkipRaster
 *
 * ---------------------------------------------------------------- */
int LIPS_SkipRaster( int printerContext, int count )
{
	pd_PrinterInfo		*pri;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	/* State Flag Check */
	if( !pri->rasterStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	pri->curPntY += count;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_EndRaster
 *
 * ---------------------------------------------------------------- */
int LIPS_EndRaster( int printerContext )
{
	pd_PrinterInfo		*pri;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	/* State Flag Check */
	if( !pri->rasterStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	pri->rasterStarted = FALSE;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_StartStream
 *
 * ---------------------------------------------------------------- */
int LIPS_StartStream( int printerContext )
{
	pd_PrinterInfo		*pri;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	/* State Flag Check */
	if( pri->streamStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	pri->streamStarted = TRUE;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_TransferStreamData
 *
 * ---------------------------------------------------------------- */
int LIPS_TransferStreamData( int printerContext, int count, void *data )
{
	pd_PrinterInfo		*pri;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	/* State Flag Check */
	if( !pri->streamStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* Data Out */
	if( !pd_DvWrite( pri, data, count ) )	return( FALSE );
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_EndStream
 *
 * ---------------------------------------------------------------- */
int LIPS_EndStream( int printerContext )
{
	pd_PrinterInfo		*pri;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	/* State Flag Check */
	if( !pri->streamStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	pri->streamStarted = FALSE;
	return( pdOK );
}
