/* Copyright (c) 2004 E-D Inc.  All rights reserved.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
/*******************************************************************

  OpenPrinting Vector Printer Driver API : for Canon

 ==================================================================

  API LIPS Functions : Job Control

********************************************************************/
/* system include files */
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>

/* local include files */
#include "pdapi.h"
#include "pd_print.h"
#include "pd_lips.h"

/*==================================================================
  API Entry
===================================================================*/
int	(*LIPS_Procs[])() =
{
	/* Creating and Managing Print Contexts */
	OpenPrinter,
	ClosePrinter,

	/* Job Control Operations */
	LIPS_StartJob,
	LIPS_EndJob,
	LIPS_StartDoc,
	LIPS_EndDoc,
	LIPS_StartPage,
	LIPS_EndPage,
	QueryDeviceCapability,
	QueryDeviceInfo,

	/* Graphics State Object Operations */
	LIPS_ResetCTM,
	LIPS_SetCTM,
	LIPS_GetCTM,
	LIPS_InitGS,
	LIPS_SaveGS,
	LIPS_RestoreGS,
	LIPS_QueryColorSpace,
	LIPS_SetColorSpace,
	LIPS_GetColorSpace,
	LIPS_QueryROP,
	NULL,	// LIPS_SetROP,
	NULL,	// LIPS_GetROP,
	LIPS_SetFillMode,
	LIPS_GetFillMode,
	NULL,	// LIPS_SetAlphaConstant,
	NULL,	// LIPS_GetAlphaConstant,
	LIPS_SetLineWidth,
	LIPS_GetLineWidth,
	LIPS_SetLineDash,
	LIPS_GetLineDash,
	LIPS_SetLineDashOffset,
	LIPS_GetLineDashOffset,
	LIPS_SetLineStyle,
	LIPS_GetLineStyle,
	LIPS_SetLineCap,
	LIPS_GetLineCap,
	LIPS_SetLineJoin,
	LIPS_GetLineJoin,
	LIPS_SetMiterLimit,
	LIPS_GetMiterLimit,
	LIPS_SetPaintMode,
	LIPS_GetPaintMode,
	LIPS_SetStrokeColor,
	LIPS_SetFillColor,
	NULL,	// LIPS_SetBgColor,

	/* Path Operations */
	LIPS_NewPath,
	LIPS_EndPath,
	LIPS_StrokePath,
	LIPS_FillPath,
	LIPS_StrokeFillPath,
	LIPS_SetClipPath,
	LIPS_ResetClipPath,
	LIPS_SetCurrentPoint,
	LIPS_LinePath,
	LIPS_PolygonPath,
	LIPS_RectanglePath,
	LIPS_RoundRectanglePath,
	LIPS_BezierPath,
	LIPS_ArchPath,

	/* Text Operations */
	LIPS_DrawBitmapText,

	/* Bitmap Image Operations */
	LIPS_DrawImage,
	LIPS_StartDrawImage,
	LIPS_TransferDrawImage,
	LIPS_EndDrawImage,

	/* Scan Line Operations */
	LIPS_StartScanline,
	LIPS_Scanline,
	LIPS_EndScanline,

	/* Raster Image Operations */
	LIPS_StartRaster,
	LIPS_TransferRasterData,
	LIPS_SkipRaster,
	LIPS_EndRaster,

	/* Stream Data Operations */
	LIPS_StartStream,
	LIPS_TransferStreamData,
	LIPS_EndStream,
};

/* support DeviceInfo */
static int	sdv_mediaCopy[]   = { 1, -1, 255, 0 };
static int	sdv_orientation[] = { pdDev_orient_portrait, pdDev_orient_landscape, 0 };
static int	sdv_pageSizeID[]  = {
	pdDev_pgSize_a4, pdDev_pgSize_a5, pdDev_pgSize_b4, pdDev_pgSize_b5,
	pdDev_pgSize_letter, pdDev_pgSize_legal,
0 };
static int	sdv_resolution[]  = {
	300, 300, 600, 600, 1200, 1200,
0 };
pd_supDevInfo	LIPS_supDev = {
	sdv_mediaCopy,
	sdv_orientation,
	sdv_pageSizeID,	TRUE,
	sdv_resolution
};

/*==================================================================
  Local Functions
===================================================================*/
/* ----------------------------------------------------------------
 *
 *	prtDevInit
 *
 * ---------------------------------------------------------------- */
static bool	prtDevInit( pd_PrinterInfo *pri )
{
	char	obuf[128];

	/* LIPS Start */
	sprintf( obuf, "%c%%@", C_ESC );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );

	return( TRUE );
}

/* ----------------------------------------------------------------
 *
 *	prtDevStart
 *
 * ---------------------------------------------------------------- */
static bool	prtDevStart( pd_PrinterInfo *pri, pd_DeviceInfo *dev )
{
	char	obuf[128];
	char	prm1[16], prm2[16];
	int		dpi, fid, ofs;

	dpi = dev->resolutionX;

	/* JOB Start : LIPS4 / dpi / JIS-sjis */
	sprintf( obuf, "%cP41;%d;1J%c\\", C_ESC, dpi, C_ESC );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );

	/* Color Mode Set : full color */
	sprintf( obuf, "%c[1\"p", C_ESC );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );

	/* Soft Reset */
	sprintf( obuf, "%c<", C_ESC );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );

	/* Page Format */
	if( dev->pageSizeID ) {
		fid = 0;
		switch( dev->pageSizeID ) {
			case pdDev_pgSize_a3 :		fid = 12;	break;
			case pdDev_pgSize_a4 :		fid = 14;	break;
			case pdDev_pgSize_a5 :		fid = 16;	break;
			case pdDev_pgSize_b4 :		fid = 24;	break;
			case pdDev_pgSize_b5 :		fid = 26;	break;
			case pdDev_pgSize_b6 :		fid = 28;	break;
			case pdDev_pgSize_hagaki :	fid = 18;	break;
			case pdDev_pgSize_letter :	fid = 30;	break;
			case pdDev_pgSize_legal :	fid = 32;	break;
		}
		if( dev->orientation == pdDev_orient_landscape )	fid++;
		sprintf( obuf, "%c[%dp", C_ESC, fid );
	} else {
		fid = 80;
		if( dev->orientation == pdDev_orient_landscape )	fid++;
		if( dev->pageSizeY >= dev->pageSizeX ) {
			sprintf( obuf, "%c[%d;%d;%dp", C_ESC, fid, dev->pageSizeY, dev->pageSizeX );
		} else {
			sprintf( obuf, "%c[%d;%d;%dp", C_ESC, fid, dev->pageSizeX, dev->pageSizeY );
		}
	}
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );

	/* Number of Copy */
	sprintf( obuf, "%c[%dv", C_ESC, dev->mediaCopy );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );

	/* Vector Mode Change */
	sprintf( obuf, "%c[0&}", C_ESC );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );

	/* Page Orientation Set */
	if( dev->orientation == pdDev_orient_landscape ) {
		pdLips_Param( -1, prm1 );
	} else {
		pdLips_Param( 0, prm1 );
	}
	sprintf( obuf, "%c@%s%c", C_SP, prm1, C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );

	/* Picture Start */
	sprintf( obuf, "#%c", C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );

	/* Coordinate Unit Set : dpi */
	pdLips_Param( dpi, prm1 );
	pdLips_Param( 1, prm2 );
	sprintf( obuf, "!0%s%s%c", prm1, prm2, C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );

	/* Color Value Mode Set : RGB value */
	pdLips_Param( 1, prm1 );
	sprintf( obuf, "!1%s%c", prm1, C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );

	/* Picture Body Start */
	sprintf( obuf, "$%c", C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );

	/* Origin Point Move : 5mm */
	ofs = pd_DtoI( (double)dpi * 5.0 / MMETER_PER_INCH );
	pdLips_Param(    -ofs, prm1 );
	pdLips_ParamAdd( -ofs, prm1 );
	sprintf( obuf, "}\"%s%c", prm1, C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );

	return( TRUE );
}
/* ----------------------------------------------------------------
 *
 *	prtDevEnd
 *
 * ---------------------------------------------------------------- */
static bool	prtDevEnd( pd_PrinterInfo *pri )
{
	char	obuf[128];

	/* Picture End */
	sprintf( obuf, "%%%c", C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );

	/* Text Mode Change */
	sprintf( obuf, "}p%c", C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );

	/* JOB End */
	sprintf( obuf, "%cP0J%c\\", C_ESC, C_ESC );
	if( !pd_DvWriteStr( pri, obuf ) )	return( FALSE );

	return( TRUE );
}

/*==================================================================
  API Functions
===================================================================*/
/* ----------------------------------------------------------------
 *
 *	LIPS_StartJob
 *
 * ---------------------------------------------------------------- */
int LIPS_StartJob( int printerContext, char *jobInfo )
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );
	if( jobInfo ) {
		pd_DebugPrint( "jobInfo = %s\n", jobInfo );
	}

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	/* State Flag Check */
	if( pri->jobStarted ) {
		errorno = pdNOTSUPPORTED;
		return( pdERROR );
	}
	/* Attribute Set */
	if( jobInfo && *jobInfo ) {
		if( !pd_SetJobInfo( pri, jobInfo ) )	return( pdERROR );
		memcpy( &pri->docDev, &pri->jobDev, sizeof( pd_DeviceInfo ) );
		memcpy( &pri->pageDev, &pri->jobDev, sizeof( pd_DeviceInfo ) );
	}
	/* Printer Device Initialize & Start */ 
	if( !prtDevInit( pri ) )	return( pdERROR );
	if( !prtDevStart( pri, &pri->jobDev ) )	return( pdERROR );
	memcpy( &pri->curDev, &pri->jobDev, sizeof( pd_DeviceInfo ) );

	/* GraphicsState Initialize */
	pd_ClearGraphicsState( gs );
	pd_InitGraphicsState( gs );

	/* State Flag Set */
	pri->jobStarted = TRUE;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_EndJob
 *
 * ---------------------------------------------------------------- */
int LIPS_EndJob( int printerContext )
{
	pd_PrinterInfo		*pri;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	/* State Flag Check */
	if( !pri->jobStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	if( pri->pathActive ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	if( pri->drwImgStarted || pri->scnLinStarted || pri->rasterStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* Printer Device End */ 
	if( !prtDevEnd( pri ) )	return( pdERROR );

	/* Attribute Set */
	memcpy( &pri->jobDev, &pri->defDev, sizeof( pd_DeviceInfo ) );
	memcpy( &pri->docDev, &pri->defDev, sizeof( pd_DeviceInfo ) );
	memcpy( &pri->pageDev, &pri->defDev, sizeof( pd_DeviceInfo ) );

	/* State Flag Set */
	pri->jobStarted = FALSE;
	pri->docStarted = FALSE;
	pri->pageStarted = FALSE;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_StartDoc
 *
 * ---------------------------------------------------------------- */
int LIPS_StartDoc( int printerContext, char *docInfo )
{
	pd_PrinterInfo		*pri;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );
	if( docInfo ) {
		pd_DebugPrint( "docInfo = %s\n", docInfo );
	}

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	/* State Flag Check */
	if( !pri->jobStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	if( pri->pageStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	if( pri->docStarted ) {
		errorno = pdNOTSUPPORTED;
		return( pdERROR );
	}
	/* Attribute Set */
	if( docInfo && *docInfo ) {
		if( !pd_SetDocInfo( pri, docInfo ) )	return( pdERROR );
		memcpy( &pri->pageDev, &pri->docDev, sizeof( pd_DeviceInfo ) );
	}
	/* Printer Device Restart */ 
	if( memcmp( &pri->curDev, &pri->docDev, sizeof( pd_DeviceInfo ) ) ) {
		if( !prtDevEnd( pri ) )		return( pdERROR );
		if( !prtDevStart( pri, &pri->docDev ) )	return( pdERROR );
		memcpy( &pri->curDev, &pri->docDev, sizeof( pd_DeviceInfo ) );
	}
	/* State Flag Set */
	pri->docStarted = TRUE;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_EndDoc
 *
 * ---------------------------------------------------------------- */
int LIPS_EndDoc( int printerContext )
{
	pd_PrinterInfo		*pri;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	/* State Flag Check */
	if( !pri->docStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	if( pri->drwImgStarted || pri->scnLinStarted || pri->rasterStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* Attribute Set */
	memcpy( &pri->docDev, &pri->jobDev, sizeof( pd_DeviceInfo ) );
	memcpy( &pri->pageDev, &pri->jobDev, sizeof( pd_DeviceInfo ) );

	/* State Flag Set */
	pri->docStarted = FALSE;
	pri->pageStarted = FALSE;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_StartPage
 *
 * ---------------------------------------------------------------- */
int LIPS_StartPage( int printerContext, char *pageInfo )
{
	pd_PrinterInfo		*pri;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );
	if( pageInfo ) {
		pd_DebugPrint( "pageInfo = %s\n", pageInfo );
	}

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	/* State Flag Check */
	if( !pri->jobStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	if( pri->pageStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* Attribute Set */
	if( pageInfo && *pageInfo ) {
		if( !pd_SetPageInfo( pri, pageInfo ) )	return( pdERROR );
	}
	/* Printer Device Restart */ 
	if( memcmp( &pri->curDev, &pri->pageDev, sizeof( pd_DeviceInfo ) ) ) {
		if( !prtDevEnd( pri ) )		return( pdERROR );
		if( !prtDevStart( pri, &pri->pageDev ) )	return( pdERROR );
		memcpy( &pri->curDev, &pri->pageDev, sizeof( pd_DeviceInfo ) );
	}
	/* State Flag Set */
	pri->pageStarted = TRUE;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_EndPage
 *
 * ---------------------------------------------------------------- */
int LIPS_EndPage( int printerContext )
{
	pd_PrinterInfo		*pri;
	char	obuf[128];

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	/* State Flag Check */
	if( !pri->pageStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	if( pri->drwImgStarted || pri->scnLinStarted || pri->rasterStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* Form Feed */
	sprintf( obuf, "%c", C_FF );
	if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );

	/* Attribute Set */
	memcpy( &pri->pageDev, &pri->docDev, sizeof( pd_DeviceInfo ) );

	/* State Flag Set */
	pri->pageStarted = FALSE;
	return( pdOK );
}
