/* Copyright (c) 2004 E-D Inc.  All rights reserved.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
/*******************************************************************

  OpenPrinting Vector Printer Driver API : for Canon

 ==================================================================

  API LIPS Functions : Vector Drawing

********************************************************************/
/* system include files */
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>

/* local include files */
#include "pdapi.h"
#include "pd_print.h"
#include "pd_lips.h"

/*==================================================================
  API Functions
===================================================================*/
/* ----------------------------------------------------------------
 *
 *	LIPS_NewPath
 *
 * ---------------------------------------------------------------- */
int LIPS_NewPath( int printerContext )
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;
	char	obuf[128];
	char	prm1[16];
	int		pid;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	/* State Flag Check */
	if( !pri->jobStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	if( pri->pathActive ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* PathID Get */
	pid = pd_GetNewPathID( pri );
	if( pid < 0 ) {
		errorno = pdNOTSUPPORTED;
		return( pdERROR );
	}
	/* Data Out */
	pdLips_Param( pid, prm1 );
	sprintf( obuf, "P(%s0%c", prm1, C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );

	pri->strPntX = pri->curPntX;
	pri->strPntY = pri->curPntY;
	pdLips_Param(    pri->curPntX, prm1 );
	pdLips_ParamAdd( pri->curPntY, prm1 );
	sprintf( obuf, "p10%s%c", prm1, C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );

	/* State Flag Set */
	pri->pathActive = TRUE;
	pri->pathID = pid;
	gs->pathEntry = FALSE;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_EndPath
 *
 * ---------------------------------------------------------------- */
int LIPS_EndPath( int printerContext )
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;
	char	obuf[128];

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	/* State Flag Check */
	if( !pri->pathActive ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* Data Out */
	sprintf( obuf, "P)%c", C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );

	/* State Flag Set */
	gs->pathEntry = TRUE;
	gs->pathID = pri->pathID;
	pri->pathActive = FALSE;
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_StrokePath
 *
 * ---------------------------------------------------------------- */
int LIPS_StrokePath( int printerContext )
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;
	char	obuf[128];
	char	prm1[16];
	int		pid;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	/* State Flag Check */
	if( !pri->pageStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	if( !gs->pathEntry && !gs->clipPathEntry ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* PathID Get */
	if( gs->pathEntry ) {
		pid = gs->pathID;
	} else {
		pid = gs->clipPathID;
	}
	/* Data Out */
	if( !pdLips_AttrOutGS( pri ) )	return( pdERROR );
	pdLips_Param( pid, prm1 );
	sprintf( obuf, "PS%s0%c", prm1, C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_FillPath
 *
 * ---------------------------------------------------------------- */
int LIPS_FillPath( int printerContext )
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;
	char	obuf[128];
	char	prm1[16];
	int		pid;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	/* State Flag Check */
	if( !pri->pageStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	if( !gs->pathEntry && !gs->clipPathEntry ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* PathID Get */
	if( gs->pathEntry ) {
		pid = gs->pathID;
	} else {
		pid = gs->clipPathID;
	}
	/* Data Out */
	if( !pdLips_AttrOutGS( pri ) )	return( pdERROR );
	pdLips_Param( pid, prm1 );
	sprintf( obuf, "PF%s0%c", prm1, C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_StrokeFillPath
 *
 * ---------------------------------------------------------------- */
int LIPS_StrokeFillPath( int printerContext )
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;
	char	obuf[128];
	char	prm1[16];
	int		pid;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	/* State Flag Check */
	if( !pri->pageStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	if( !gs->pathEntry && !gs->clipPathEntry ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* PathID Get */
	if( gs->pathEntry ) {
		pid = gs->pathID;
	} else {
		pid = gs->clipPathID;
	}
	/* Data Out */
	if( !pdLips_AttrOutGS( pri ) )	return( pdERROR );
	pdLips_Param( pid, prm1 );
	sprintf( obuf, "P&%s0%c", prm1, C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_SetClipPath
 *
 * ---------------------------------------------------------------- */
int LIPS_SetClipPath( int printerContext, pdClipRule clipRule )
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;
	char	obuf[128];
	char	prm1[16];
	int		pid;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	/* State Flag Check */
	if( !pri->jobStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	if( !gs->pathEntry && !gs->clipPathEntry ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* PathID Get */
	if( gs->pathEntry ) {
		pid = gs->pathID;
	} else {
		pid = gs->clipPathID;
	}
	/* Attribute Change */
	if( clipRule == pdclipRuleWinding ) {
		sprintf( obuf, "}I1%c", C_IS2 );
	} else {
		sprintf( obuf, "}I0%c", C_IS2 );
	}
	if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );
	sprintf( obuf, "}H1%c", C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );

	/* Data Out */
	pdLips_Param( pid, prm1 );
	sprintf( obuf, "PC%s0%c", prm1, C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );

	sprintf( obuf, "*0%c", C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );

	/* State Flag Set */
	if( gs->pathEntry ) {
		gs->clipPathEntry = TRUE;
		gs->clipPathID = gs->pathID;
		gs->pathEntry = FALSE;
	}
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_ResetClipPath
 *
 * ---------------------------------------------------------------- */
int LIPS_ResetClipPath( int printerContext )
{
	pd_PrinterInfo		*pri;
	char	obuf[128];

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	/* State Flag Check */
	if( !pri->jobStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* Data Out */
	sprintf( obuf, "*1%c", C_IS2 );
	if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_SetCurrentPoint
 *
 * ---------------------------------------------------------------- */
int LIPS_SetCurrentPoint( int printerContext, pdFix x, pdFix y )
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;
	char	obuf[128];
	char	prm1[16];

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	/* State Flag Check */
	if( !pri->jobStarted ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* Data Out */
	pri->curPntX = pd_Dx_Lcd( x, y, &gs->ctm );
	pri->curPntY = pd_Dy_Lcd( x, y, &gs->ctm );
	pri->strPntX = pri->curPntX;
	pri->strPntY = pri->curPntY;
	if( pri->pathActive ) {
		pdLips_Param(    pri->curPntX, prm1 );
		pdLips_ParamAdd( pri->curPntY, prm1 );
		sprintf( obuf, "p10%s%c", prm1, C_IS2 );
		if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );
	}
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_LinePath
 *
 * ---------------------------------------------------------------- */
int LIPS_LinePath( int printerContext, int flag, int npoints, pdPoint *points )
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;
	char	obuf[128];
	char	prm1[16];
	int		n;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	/* State Flag Check */
	if( !pri->pathActive ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* Data Out */
	if( npoints > 0 ) {
		if( !pd_DvWriteStr( pri, "p402" ) )	return( pdERROR );
		for( n = 0; n < npoints; n++ ) {
			pri->curPntX = pd_Dx_Lcd( points[n].x, points[n].y, &gs->ctm );
			pri->curPntY = pd_Dy_Lcd( points[n].x, points[n].y, &gs->ctm );
			pdLips_Param(    pri->curPntX, prm1 );
			pdLips_ParamAdd( pri->curPntY, prm1 );
			if( !pd_DvWriteStr( pri, prm1 ) )	return( pdERROR );
		}
		sprintf( obuf, "%c", C_IS2 );
		if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );
	}
	if( flag == pdPathClose ) {
		pri->curPntX = pri->strPntX;
		pri->curPntY = pri->strPntY;
		sprintf( obuf, "p0%c", C_IS2 );
		if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );
	}
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_PolygonPath
 *
 * ---------------------------------------------------------------- */
int LIPS_PolygonPath( int printerContext, int npolygons, int *nvertexes, pdPoint *points )
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;
	char	obuf[128];
	char	prm1[16];
	int		n0, n1, n2;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	/* State Flag Check */
	if( !pri->pathActive ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* Data Out */
	for( n0 = n1 = 0; n1 < npolygons; n1++, nvertexes++ ) {
		if( *nvertexes <= 0 )	continue;
		pri->curPntX = pd_Dx_Lcd( points[n0].x, points[n0].y, &gs->ctm );
		pri->curPntY = pd_Dy_Lcd( points[n0].x, points[n0].y, &gs->ctm );
		pri->strPntX = pri->curPntX;
		pri->strPntY = pri->curPntY;
		pdLips_Param(    pri->curPntX, prm1 );
		pdLips_ParamAdd( pri->curPntY, prm1 );
		sprintf( obuf, "p10%s%c", prm1, C_IS2 );
		if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );
		n0++;
		if( *nvertexes > 1 ) {
			if( !pd_DvWriteStr( pri, "p402" ) )	return( pdERROR );
			for( n2 = 1; n2 < *nvertexes; n2++, n0++ ) {
				pdLips_Param(    pd_Dx_Lcd( points[n0].x, points[n0].y, &gs->ctm ), prm1 );
				pdLips_ParamAdd( pd_Dy_Lcd( points[n0].x, points[n0].y, &gs->ctm ), prm1 );
				if( !pd_DvWriteStr( pri, prm1 ) )	return( pdERROR );
			}
			sprintf( obuf, "%c", C_IS2 );
			if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );
		}
		sprintf( obuf, "p0%c", C_IS2 );
		if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );
	}
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_RectanglePath
 *
 * ---------------------------------------------------------------- */
int LIPS_RectanglePath( int printerContext, int nrectangles, pdRectangle *rectangles )
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;
	char	obuf[128];
	char	prm1[16], prm2[16], prm3[16];
	int		n;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	/* State Flag Check */
	if( !pri->pathActive ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* Data Out */
	for( n = 0; n < nrectangles; n++ ) {
		pri->curPntX = pd_Dx_Lcd( rectangles[n].p0.x, rectangles[n].p0.y, &gs->ctm );
		pri->curPntY = pd_Dy_Lcd( rectangles[n].p0.x, rectangles[n].p0.y, &gs->ctm );
		pri->strPntX = pri->curPntX;
		pri->strPntY = pri->curPntY;
		pdLips_Param(    pri->curPntX, prm1 );
		pdLips_ParamAdd( pri->curPntY, prm1 );
		sprintf( obuf, "p10%s%c", prm1, C_IS2 );
		if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );

		pdLips_Param(    pd_Dx_Lcd( rectangles[n].p1.x, rectangles[n].p0.y, &gs->ctm ), prm1 );
		pdLips_ParamAdd( pd_Dy_Lcd( rectangles[n].p1.x, rectangles[n].p0.y, &gs->ctm ), prm1 );
		pdLips_Param(    pd_Dx_Lcd( rectangles[n].p1.x, rectangles[n].p1.y, &gs->ctm ), prm2 );
		pdLips_ParamAdd( pd_Dy_Lcd( rectangles[n].p1.x, rectangles[n].p1.y, &gs->ctm ), prm2 );
		pdLips_Param(    pd_Dx_Lcd( rectangles[n].p0.x, rectangles[n].p1.y, &gs->ctm ), prm3 );
		pdLips_ParamAdd( pd_Dy_Lcd( rectangles[n].p0.x, rectangles[n].p1.y, &gs->ctm ), prm3 );
		sprintf( obuf, "p402%s%s%s%c", prm1, prm2, prm3, C_IS2 );
		if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );

		sprintf( obuf, "p0%c", C_IS2 );
		if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );
	}
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_RoundRectanglePath
 *
 * ---------------------------------------------------------------- */
int LIPS_RoundRectanglePath( int printerContext, int nrectangles, pdRoundRectangle *rectangles )
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;
	char	obuf[128];
	char	prm1[16], prm2[16], prm3[16], prm4[16];
	int		ex, ey, lx, ly, x[2], y[2];
	int		n;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	/* State Flag Check */
	if( !pri->pathActive ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	if( gs->ctm.a <= 0 || gs->ctm.b != 0 ||
		gs->ctm.c != 0 || gs->ctm.d <= 0 ) {
		errorno = pdNOTSUPPORTED;
		return( pdERROR );
	}
	/* Data Out */
	for( n = 0; n < nrectangles; n++ ) {
		ex = pd_DtoI( Fixtof( rectangles[n].xellipse ) * gs->ctm.a );
		ey = pd_DtoI( Fixtof( rectangles[n].yellipse ) * gs->ctm.d );
		lx = ex * 2;
		ly = ey * 2;
		x[0] = pd_Dx_Lcd( rectangles[n].p0.x, rectangles[n].p0.y, &gs->ctm );
		y[0] = pd_Dy_Lcd( rectangles[n].p0.x, rectangles[n].p0.y, &gs->ctm );
		x[1] = pd_Dx_Lcd( rectangles[n].p1.x, rectangles[n].p1.y, &gs->ctm );
		y[1] = pd_Dy_Lcd( rectangles[n].p1.x, rectangles[n].p1.y, &gs->ctm );
		pri->curPntX = x[0];
		pri->curPntY = y[0];
		pri->strPntX = pri->curPntX;
		pri->strPntY = pri->curPntY;
		if( x[1] > x[0] ) {
			if( y[1] > y[0] ) {
				pdLips_Param( x[0],      prm1 );	pdLips_ParamAdd( y[0] + ey, prm1 );
				sprintf( obuf, "p10%s%c", prm1, C_IS2 );
				if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );

				pdLips_Param( x[0],      prm1 );	pdLips_ParamAdd( y[0],      prm1 );
				pdLips_Param( x[0] + lx, prm2 );	pdLips_ParamAdd( y[0] + ly, prm2 );
				pdLips_Param( -1, prm3 );	pdLips_ParamAdd(  0, prm3 );
				pdLips_Param(  0, prm4 );	pdLips_ParamAdd( -1, prm4 );
				sprintf( obuf, "p300%s%s%s%s001%c", prm1, prm2, prm3, prm4, C_IS2 );
				if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );

				pdLips_Param( x[1] - lx, prm1 );	pdLips_ParamAdd( y[0],      prm1 );
				pdLips_Param( x[1],      prm2 );	pdLips_ParamAdd( y[0] + ly, prm2 );
				pdLips_Param(  0, prm3 );	pdLips_ParamAdd( -1, prm3 );
				pdLips_Param(  1, prm4 );	pdLips_ParamAdd(  0, prm4 );
				sprintf( obuf, "p300%s%s%s%s001%c", prm1, prm2, prm3, prm4, C_IS2 );
				if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );

				pdLips_Param( x[1] - lx, prm1 );	pdLips_ParamAdd( y[1] - ly, prm1 );
				pdLips_Param( x[1],      prm2 );	pdLips_ParamAdd( y[1],      prm2 );
				pdLips_Param(  1, prm3 );	pdLips_ParamAdd(  0, prm3 );
				pdLips_Param(  0, prm4 );	pdLips_ParamAdd(  1, prm4 );
				sprintf( obuf, "p300%s%s%s%s001%c", prm1, prm2, prm3, prm4, C_IS2 );
				if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );

				pdLips_Param( x[0],      prm1 );	pdLips_ParamAdd( y[1] - ly, prm1 );
				pdLips_Param( x[0] + lx, prm2 );	pdLips_ParamAdd( y[1],      prm2 );
				pdLips_Param(  0, prm3 );	pdLips_ParamAdd(  1, prm3 );
				pdLips_Param( -1, prm4 );	pdLips_ParamAdd(  0, prm4 );
				sprintf( obuf, "p300%s%s%s%s001%c", prm1, prm2, prm3, prm4, C_IS2 );
				if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );
			} else {
				pdLips_Param( x[0],      prm1 );	pdLips_ParamAdd( y[0] - ey, prm1 );
				sprintf( obuf, "p10%s%c", prm1, C_IS2 );
				if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );

				pdLips_Param( x[0],      prm1 );	pdLips_ParamAdd( y[0] - ly, prm1 );
				pdLips_Param( x[0] + lx, prm2 );	pdLips_ParamAdd( y[0],      prm2 );
				pdLips_Param( -1, prm3 );	pdLips_ParamAdd(  0, prm3 );
				pdLips_Param(  0, prm4 );	pdLips_ParamAdd(  1, prm4 );
				sprintf( obuf, "p300%s%s%s%s011%c", prm1, prm2, prm3, prm4, C_IS2 );
				if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );

				pdLips_Param( x[1] - lx, prm1 );	pdLips_ParamAdd( y[0] - ly, prm1 );
				pdLips_Param( x[1],      prm2 );	pdLips_ParamAdd( y[0],      prm2 );
				pdLips_Param(  0, prm3 );	pdLips_ParamAdd(  1, prm3 );
				pdLips_Param(  1, prm4 );	pdLips_ParamAdd(  0, prm4 );
				sprintf( obuf, "p300%s%s%s%s011%c", prm1, prm2, prm3, prm4, C_IS2 );
				if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );

				pdLips_Param( x[1] - lx, prm1 );	pdLips_ParamAdd( y[1],      prm1 );
				pdLips_Param( x[1],      prm2 );	pdLips_ParamAdd( y[1] + ly, prm2 );
				pdLips_Param(  1, prm3 );	pdLips_ParamAdd(  0, prm3 );
				pdLips_Param(  0, prm4 );	pdLips_ParamAdd( -1, prm4 );
				sprintf( obuf, "p300%s%s%s%s011%c", prm1, prm2, prm3, prm4, C_IS2 );
				if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );

				pdLips_Param( x[0],      prm1 );	pdLips_ParamAdd( y[1],      prm1 );
				pdLips_Param( x[0] + lx, prm2 );	pdLips_ParamAdd( y[1] + ly, prm2 );
				pdLips_Param(  0, prm3 );	pdLips_ParamAdd( -1, prm3 );
				pdLips_Param( -1, prm4 );	pdLips_ParamAdd(  0, prm4 );
				sprintf( obuf, "p300%s%s%s%s011%c", prm1, prm2, prm3, prm4, C_IS2 );
				if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );
			}
		} else {
			if( y[1] > y[0] ) {
				pdLips_Param( x[0],      prm1 );	pdLips_ParamAdd( y[0] + ey, prm1 );
				sprintf( obuf, "p10%s%c", prm1, C_IS2 );
				if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );

				pdLips_Param( x[0] - lx, prm1 );	pdLips_ParamAdd( y[0],      prm1 );
				pdLips_Param( x[0],      prm2 );	pdLips_ParamAdd( y[0] + ly, prm2 );
				pdLips_Param(  1, prm3 );	pdLips_ParamAdd(  0, prm3 );
				pdLips_Param(  0, prm4 );	pdLips_ParamAdd( -1, prm4 );
				sprintf( obuf, "p300%s%s%s%s011%c", prm1, prm2, prm3, prm4, C_IS2 );
				if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );

				pdLips_Param( x[1],      prm1 );	pdLips_ParamAdd( y[0],      prm1 );
				pdLips_Param( x[1] + lx, prm2 );	pdLips_ParamAdd( y[0] + ly, prm2 );
				pdLips_Param(  0, prm3 );	pdLips_ParamAdd( -1, prm3 );
				pdLips_Param( -1, prm4 );	pdLips_ParamAdd(  0, prm4 );
				sprintf( obuf, "p300%s%s%s%s011%c", prm1, prm2, prm3, prm4, C_IS2 );
				if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );

				pdLips_Param( x[1],      prm1 );	pdLips_ParamAdd( y[1] - ly, prm1 );
				pdLips_Param( x[1] + lx, prm2 );	pdLips_ParamAdd( y[1],      prm2 );
				pdLips_Param( -1, prm3 );	pdLips_ParamAdd(  0, prm3 );
				pdLips_Param(  0, prm4 );	pdLips_ParamAdd(  1, prm4 );
				sprintf( obuf, "p300%s%s%s%s011%c", prm1, prm2, prm3, prm4, C_IS2 );
				if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );

				pdLips_Param( x[0] - lx, prm1 );	pdLips_ParamAdd( y[1] - ly, prm1 );
				pdLips_Param( x[0],      prm2 );	pdLips_ParamAdd( y[1],      prm2 );
				pdLips_Param(  0, prm3 );	pdLips_ParamAdd(  1, prm3 );
				pdLips_Param(  1, prm4 );	pdLips_ParamAdd(  0, prm4 );
				sprintf( obuf, "p300%s%s%s%s011%c", prm1, prm2, prm3, prm4, C_IS2 );
				if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );
			} else {
				pdLips_Param( x[0],      prm1 );	pdLips_ParamAdd( y[0] - ey, prm1 );
				sprintf( obuf, "p10%s%c", prm1, C_IS2 );
				if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );

				pdLips_Param( x[0] - lx, prm1 );	pdLips_ParamAdd( y[0] - ly, prm1 );
				pdLips_Param( x[0],      prm2 );	pdLips_ParamAdd( y[0],      prm2 );
				pdLips_Param(  1, prm3 );	pdLips_ParamAdd(  0, prm3 );
				pdLips_Param(  0, prm4 );	pdLips_ParamAdd(  1, prm4 );
				sprintf( obuf, "p300%s%s%s%s001%c", prm1, prm2, prm3, prm4, C_IS2 );
				if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );

				pdLips_Param( x[1],      prm1 );	pdLips_ParamAdd( y[0] - ly, prm1 );
				pdLips_Param( x[1] + lx, prm2 );	pdLips_ParamAdd( y[0],      prm2 );
				pdLips_Param(  0, prm3 );	pdLips_ParamAdd(  1, prm3 );
				pdLips_Param( -1, prm4 );	pdLips_ParamAdd(  0, prm4 );
				sprintf( obuf, "p300%s%s%s%s001%c", prm1, prm2, prm3, prm4, C_IS2 );
				if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );

				pdLips_Param( x[1],      prm1 );	pdLips_ParamAdd( y[1],      prm1 );
				pdLips_Param( x[1] + lx, prm2 );	pdLips_ParamAdd( y[1] + ly, prm2 );
				pdLips_Param( -1, prm3 );	pdLips_ParamAdd(  0, prm3 );
				pdLips_Param(  0, prm4 );	pdLips_ParamAdd( -1, prm4 );
				sprintf( obuf, "p300%s%s%s%s001%c", prm1, prm2, prm3, prm4, C_IS2 );
				if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );

				pdLips_Param( x[0] - lx, prm1 );	pdLips_ParamAdd( y[1],      prm1 );
				pdLips_Param( x[0],      prm2 );	pdLips_ParamAdd( y[1] + ly, prm2 );
				pdLips_Param(  0, prm3 );	pdLips_ParamAdd( -1, prm3 );
				pdLips_Param(  1, prm4 );	pdLips_ParamAdd(  0, prm4 );
				sprintf( obuf, "p300%s%s%s%s001%c", prm1, prm2, prm3, prm4, C_IS2 );
				if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );
			}
		}
		sprintf( obuf, "p0%c", C_IS2 );
		if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );

		pdLips_Param(    pri->curPntX, prm1 );
		pdLips_ParamAdd( pri->curPntY, prm1 );
		sprintf( obuf, "p10%s%c", prm1, C_IS2 );
		if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );
	}
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_BezierPath
 *
 * ---------------------------------------------------------------- */
int LIPS_BezierPath( int printerContext, int npoints, pdPoint *points )
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;
	char	obuf[128];
	char	prm1[16];
	int		n;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	/* State Flag Check */
	if( !pri->pathActive ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	/* Data Out */
	if( npoints > 0 ) {
		if( !pd_DvWriteStr( pri, "p404" ) )	return( pdERROR );
		for( n = 0; n < npoints; n++ ) {
			pri->curPntX = pd_Dx_Lcd( points[n].x, points[n].y, &gs->ctm );
			pri->curPntY = pd_Dy_Lcd( points[n].x, points[n].y, &gs->ctm );
			pdLips_Param(    pri->curPntX, prm1 );
			pdLips_ParamAdd( pri->curPntY, prm1 );
			if( !pd_DvWriteStr( pri, prm1 ) )	return( pdERROR );
		}
		sprintf( obuf, "%c", C_IS2 );
		if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );
	}
	return( pdOK );
}

/* ----------------------------------------------------------------
 *
 *	LIPS_ArchPath
 *
 * ---------------------------------------------------------------- */
int LIPS_ArchPath( int printerContext, int kind, int dir, pdFix bbx0, pdFix bby0, pdFix bbx1, pdFix bby1, pdFix x0, pdFix y0, pdFix x1, pdFix y1 )
{
	pd_PrinterInfo		*pri;
	pd_GraphicsState	*gs;
	char	obuf[128];
	char	prm1[16], prm2[16], prm3[16], prm4[16], dr;
	double	dx, dy;
	int		fg, x[4], y[4], cx, cy, i;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );
	gs = &pri->gstateList.gstate;

	/* State Flag Check */
	if( !pri->pathActive ) {
		errorno = pdBADREQUEST;
		return( pdERROR );
	}
	if( gs->ctm.a <= 0 || gs->ctm.b != 0 ||
		gs->ctm.c != 0 || gs->ctm.d <= 0 ) {
		errorno = pdNOTSUPPORTED;
		return( pdERROR );
	}
	/* Data Out */
	fg = 0;
	if( dir == pdCounterclockwise ) {
		dr = '1';
	} else {
		dr = '0';
	}
	x[0] = pd_Dx_Lcd( bbx0, bby0, &gs->ctm );
	y[0] = pd_Dy_Lcd( bbx0, bby0, &gs->ctm );
	x[1] = pd_Dx_Lcd( bbx1, bby1, &gs->ctm );
	y[1] = pd_Dy_Lcd( bbx1, bby1, &gs->ctm );
	x[2] = pd_Dx_Lcd( x0, y0, &gs->ctm );
	y[2] = pd_Dy_Lcd( x0, y0, &gs->ctm );
	x[3] = pd_Dx_Lcd( x1, y1, &gs->ctm );
	y[3] = pd_Dy_Lcd( x1, y1, &gs->ctm );
	if( x[1] < x[0] ) {
		i = x[1];	x[1] = x[0];	x[0] = i;
	}
	if( y[1] < y[0] ) {
		i = y[1];	y[1] = y[0];	y[0] = i;
	}
	pdLips_Param( x[0], prm1 );	pdLips_ParamAdd( y[0], prm1 );
	pdLips_Param( x[1], prm2 );	pdLips_ParamAdd( y[1], prm2 );

	dx = ( Fixtof( bbx0 ) + Fixtof( bbx1 ) ) / 2.0;
	dy = ( Fixtof( bby0 ) + Fixtof( bby1 ) ) / 2.0;
	cx = pd_Dx_dLcd( dx, dy, &gs->ctm );
	cy = pd_Dy_dLcd( dx, dy, &gs->ctm );
	pdLips_Param( x[2] - cx, prm3 );	pdLips_ParamAdd( y[2] - cy, prm3 );
	pdLips_Param( x[3] - cx, prm4 );	pdLips_ParamAdd( y[3] - cy, prm4 );
	if( kind == pdPie ) {
		sprintf( obuf, "p301%s%s%s%s0%c%c", prm1, prm2, prm3, prm4, dr, C_IS2 );
		if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );
		pri->curPntX = x[0];
		pri->curPntY = y[0];
		fg = 1;
	} else if( kind == pdChord ) {
		sprintf( obuf, "p301%s%s%s%s1%c%c", prm1, prm2, prm3, prm4, dr, C_IS2 );
		if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );
		pri->curPntX = x[0];
		pri->curPntY = y[0];
		fg = 1;
	} else if( Fixtof( x0 ) == Fixtof( x1 ) && Fixtof( y0 ) == Fixtof( y1 ) ) {
		sprintf( obuf, "p301%s%s%s%s1%c%c", prm1, prm2, prm3, prm4, dr, C_IS2 );
		if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );
		pri->curPntX = x[3];
		pri->curPntY = y[3];
		fg = 1;
	} else {
		sprintf( obuf, "p300%s%s%s%s0%c%c", prm1, prm2, prm3, prm4, dr, C_IS2 );
		if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );
		pri->curPntX = x[3];
		pri->curPntY = y[3];
	}
	if( fg ) {
		pri->strPntX = pri->curPntX;
		pri->strPntY = pri->curPntY;
		pdLips_Param(    pri->curPntX, prm1 );
		pdLips_ParamAdd( pri->curPntY, prm1 );
		sprintf( obuf, "p10%s%c", prm1, C_IS2 );
		if( !pd_DvWriteStr( pri, obuf ) )	return( pdERROR );
	}
	return( pdOK );
}
