/* Copyright (c) 2004 E-D Inc.  All rights reserved.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
/*******************************************************************

  OpenPrinting Vector Printer Driver API : for Canon

 ==================================================================

  API Common Functions

********************************************************************/
/* system include files */
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>

/* local include files */
#include "pdapi.h"
#include "pd_print.h"

/*==================================================================
  Entry Data
===================================================================*/
int		errorno;

/*==================================================================
  Entry Functions
===================================================================*/
/* ----------------------------------------------------------------
 *
 *	OpenPrinter
 *
 * ---------------------------------------------------------------- */
int OpenPrinter( int outputFD, char *printerModel, int *nApiEntry, int (**apiEntry[])() )
{
	pd_PrinterInfo		*pri;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );
	if( printerModel ) {
		pd_DebugPrint( "printerModel = %s\n", printerModel );
	}

	/* Addition PrinterInfo */
	pri = pd_NewPrinterInfo( outputFD );
	if( !pri ) {
		return( pdERROR );
	}
	/* Set PrinterInfo */
	*apiEntry = NULL;
	*nApiEntry = 0;
	memset( &pri->supDev, 0, sizeof( pd_supDevInfo ) );
//	if( printerModel && !strcmp( printerModel, "Canon_PIXUS850i" ) ) {
//		*apiEntry = Inkjet_Procs;
//		*nApiEntry = pdAPI_COUNT;
//		memcpy( &pri->supDev, &Inkjet_supDev, sizeof( pd_supDevInfo ) );
//	} else {
		*apiEntry = LIPS_Procs;
		*nApiEntry = pdAPI_COUNT;
		memcpy( &pri->supDev, &LIPS_supDev, sizeof( pd_supDevInfo ) );
//	}
	pd_SetDefDevInfo( pri );
	memcpy( &pri->jobDev, &pri->defDev, sizeof( pd_DeviceInfo ) );
	memcpy( &pri->docDev, &pri->defDev, sizeof( pd_DeviceInfo ) );
	memcpy( &pri->pageDev, &pri->defDev, sizeof( pd_DeviceInfo ) );
	pri->jobStarted = FALSE;
	pri->docStarted = FALSE;
	pri->pageStarted = FALSE;
	pri->curPntX = 0;
	pri->curPntY = 0;
	pri->pathActive = FALSE;
	pri->strPntX = pri->curPntX;
	pri->strPntY = pri->curPntY;
	pri->drwImgStarted = FALSE;
	pri->scnLinStarted = FALSE;
	pri->rasterStarted = FALSE;
	pri->streamStarted = FALSE;
	pd_InitGraphicsState( &pri->gstateList.gstate );
	return( pri->printerContext );
}

/* ----------------------------------------------------------------
 *
 *	ClosePrinter
 *
 * ---------------------------------------------------------------- */
int ClosePrinter( int printerContext )
{
	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	/* Delete PrinterInfo */
	if( !pd_DeletePrinterInfo( printerContext ) ) {
		return( pdERROR );
	}
	return( pdOK );
}

/*==================================================================*/
#define	SCHEME_STR			"updf"
#define	ATTR_KEY_COPY		"MediaCopy"
#define	ATTR_KEY_ORIENT		"MediaPageRotation"
#define	ATTR_KEY_PGSIZE		"MediaSize"
#define	ATTR_KEY_RESOLUT	"DeviceResolution"

#define	ATTR_VALFRM_PGSIZE	"%dx%dmm"
#define	ATTR_VALFRM_RESOLUT	"deviceResolution_%dx%d"

/*------------------------------------------------------------------*/
static char	*attr_value_orient( int id )
{
	switch( id ) {
		case pdDev_orient_portrait :	return( "portrait" );
		case pdDev_orient_landscape :	return( "landscape" );
	}
	return( "" );
}
static char	*attr_value_pgsize( int id )
{
	switch( id ) {
		case pdDev_pgSize_a3 :		return( "a3" );
		case pdDev_pgSize_a4 :		return( "a4" );
		case pdDev_pgSize_a5 :		return( "a5" );
		case pdDev_pgSize_b4 :		return( "b4" );
		case pdDev_pgSize_b5 :		return( "b5" );
		case pdDev_pgSize_b6 :		return( "b6" );
		case pdDev_pgSize_hagaki :	return( "hagaki" );
		case pdDev_pgSize_letter :	return( "letter" );
		case pdDev_pgSize_legal :	return( "legal" );
	}
	return( "" );
}
/*------------------------------------------------------------------*/
static void	set_attr_infobuf( int bln, char *buf, int *fg, int *ln, char *str )
{
	int		n;
	n = strlen( str );
	if( buf && !(*fg) ) {
		if( bln > *ln + n ) {
			strcpy( buf + *ln, str );
		} else {
			*fg = 1;
		}
	}
	*ln += n;
}

/* ----------------------------------------------------------------
 *
 *	QueryDeviceCapability
 *
 * ---------------------------------------------------------------- */
int QueryDeviceCapability( int printerContext, int queryflag, int buflen, char *infoBuf )
{
	pd_PrinterInfo		*pri;
	char	str[128];
	int		sw, fg, ln, n, i, x, y;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	sw = fg = ln = 0;
	sprintf( str, "%s:", SCHEME_STR );
	set_attr_infobuf( buflen, infoBuf, &fg, &ln, str );
	if( queryflag & pdQFMediaCopy ) {
		if( !sw )	sw = 1;
		else		set_attr_infobuf( buflen, infoBuf, &fg, &ln, ";" );
		sprintf( str, "%s=", ATTR_KEY_COPY );
		set_attr_infobuf( buflen, infoBuf, &fg, &ln, str );
		for( n = 0; ( i = *( pri->supDev.mediaCopy + n ) ); n++ ) {
			if( n )	set_attr_infobuf( buflen, infoBuf, &fg, &ln, "," );
			if( i < 0 ) {
				sprintf( str, ".." );	// A,..,B : A <= num <= B
			} else {
				sprintf( str, "%d", i );
			}
			set_attr_infobuf( buflen, infoBuf, &fg, &ln, str );
		}
	}
	if( queryflag & pdQFPageRotation ) {
		if( !sw )	sw = 1;
		else		set_attr_infobuf( buflen, infoBuf, &fg, &ln, ";" );
		sprintf( str, "%s=", ATTR_KEY_ORIENT );
		set_attr_infobuf( buflen, infoBuf, &fg, &ln, str );
		for( n = 0; ( i = *( pri->supDev.orientation + n ) ); n++ ) {
			if( n )	set_attr_infobuf( buflen, infoBuf, &fg, &ln, "," );
			set_attr_infobuf( buflen, infoBuf, &fg, &ln, attr_value_orient( i ) );
		}
	}
	if( queryflag & pdQFMediaSize ) {
		if( !sw )	sw = 1;
		else		set_attr_infobuf( buflen, infoBuf, &fg, &ln, ";" );
		sprintf( str, "%s=", ATTR_KEY_PGSIZE );
		set_attr_infobuf( buflen, infoBuf, &fg, &ln, str );
		for( n = 0; ( i = *( pri->supDev.pageSizeID + n ) ); n++ ) {
			if( n )	set_attr_infobuf( buflen, infoBuf, &fg, &ln, "," );
			set_attr_infobuf( buflen, infoBuf, &fg, &ln, attr_value_pgsize( i ) );
		}
	}
	if( queryflag & pdQFDeviceResolution ) {
		if( !sw )	sw = 1;
		else		set_attr_infobuf( buflen, infoBuf, &fg, &ln, ";" );
		sprintf( str, "%s=", ATTR_KEY_RESOLUT );
		set_attr_infobuf( buflen, infoBuf, &fg, &ln, str );
		for( n = 0; ( x = *( pri->supDev.resolution + n ) ) &&
					( y = *( pri->supDev.resolution + n + 1 ) ); n += 2 ) {
			if( n )	set_attr_infobuf( buflen, infoBuf, &fg, &ln, "," );
			sprintf( str, ATTR_VALFRM_RESOLUT, x, y );
			set_attr_infobuf( buflen, infoBuf, &fg, &ln, str );
		}
	}
	return( ln + 1 );
}

/* ----------------------------------------------------------------
 *
 *	QueryDeviceInfo
 *
 * ---------------------------------------------------------------- */
int QueryDeviceInfo( int printerContext, int queryflag, int buflen, char *infoBuf )
{
	pd_PrinterInfo		*pri;
	char	str[128];
	int		sw, fg, ln;

	pd_DebugPrint( "pdapi: %s\n", __FUNCTION__ );

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( pdERROR );

	sw = fg = ln = 0;
	sprintf( str, "%s:", SCHEME_STR );
	set_attr_infobuf( buflen, infoBuf, &fg, &ln, str );
	if( queryflag & pdQFMediaCopy ) {
		if( !sw )	sw = 1;
		else		set_attr_infobuf( buflen, infoBuf, &fg, &ln, ";" );
		sprintf( str, "%s=", ATTR_KEY_COPY );
		set_attr_infobuf( buflen, infoBuf, &fg, &ln, str );
		sprintf( str, "%d", pri->curDev.mediaCopy );
		set_attr_infobuf( buflen, infoBuf, &fg, &ln, str );
	}
	if( queryflag & pdQFPageRotation ) {
		if( !sw )	sw = 1;
		else		set_attr_infobuf( buflen, infoBuf, &fg, &ln, ";" );
		sprintf( str, "%s=", ATTR_KEY_ORIENT );
		set_attr_infobuf( buflen, infoBuf, &fg, &ln, str );
		set_attr_infobuf( buflen, infoBuf, &fg, &ln, attr_value_orient( pri->curDev.orientation ) );
	}
	if( queryflag & pdQFMediaSize ) {
		if( !sw )	sw = 1;
		else		set_attr_infobuf( buflen, infoBuf, &fg, &ln, ";" );
		sprintf( str, "%s=", ATTR_KEY_PGSIZE );
		set_attr_infobuf( buflen, infoBuf, &fg, &ln, str );
		if( pri->curDev.pageSizeID ) {
			set_attr_infobuf( buflen, infoBuf, &fg, &ln, attr_value_pgsize( pri->curDev.pageSizeID ) );
		} else {
			sprintf( str, ATTR_VALFRM_PGSIZE, pri->curDev.pageSizeX, pri->curDev.pageSizeY );
			set_attr_infobuf( buflen, infoBuf, &fg, &ln, str );
		}
	}
	if( queryflag & pdQFDeviceResolution ) {
		if( !sw )	sw = 1;
		else		set_attr_infobuf( buflen, infoBuf, &fg, &ln, ";" );
		sprintf( str, "%s=", ATTR_KEY_RESOLUT );
		set_attr_infobuf( buflen, infoBuf, &fg, &ln, str );
		sprintf( str, ATTR_VALFRM_RESOLUT, pri->curDev.resolutionX, pri->curDev.resolutionY );
		set_attr_infobuf( buflen, infoBuf, &fg, &ln, str );
	}
	return( ln + 1 );
}
