/* Copyright (c) 2004 E-D Inc.  All rights reserved.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
/*******************************************************************

  OpenPrinting Vector Printer Driver API : for Canon

 ==================================================================

  Common Functions : Printer Control

********************************************************************/
/* system include files */
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>

/* local include files */
#include "pdapi.h"
#include "pd_print.h"

/*******************************************************************
  for Data Out
********************************************************************/
/* ----------------------------------------------------------------
 *
 *	pd_DvWrite
 *
 * ---------------------------------------------------------------- */
bool	pd_DvWrite( pd_PrinterInfo *pri, char *buf, size_t siz )
{
	int			fd;
	ssize_t		wsz;

	fd = pri->outputFD;
	while( siz > 0 ) {
		wsz = write( fd, buf, siz );
		if( wsz < 0 ) {
			if( ( errno != EINTR ) && ( errno != EAGAIN ) && ( errno != EIO ) ) {
				pd_DebugPrint("pdapi: %s,%d\n", __FUNCTION__, __LINE__);
				errorno = pdFATALERROR;
				return( FALSE );
			}
		} else {
			buf += wsz;
			siz -= wsz;
		}
	}
	return( TRUE );
}
/* ----------------------------------------------------------------
 *
 *	pd_DvWriteStr
 *
 * ---------------------------------------------------------------- */
bool	pd_DvWriteStr( pd_PrinterInfo *pri, char *buf )
{
	return( pd_DvWrite( pri, buf, strlen( buf ) ) );
}

/*******************************************************************
  for Printer Control
********************************************************************/
/* ----------------------------------------------------------------
 *
 *	pd_SetDefDevInfo
 *
 * ---------------------------------------------------------------- */
void	pd_SetDefDevInfo( pd_PrinterInfo *pri )
{
	pri->defDev.mediaCopy = 1;
	pri->defDev.orientation = pdDev_orient_portrait;
	pri->defDev.pageSizeID = pdDev_pgSize_a4;
	pri->defDev.pageSizeX = 0;
	pri->defDev.pageSizeY = 0;
	pri->defDev.resolutionX = 300;
	pri->defDev.resolutionY = 300;
	if( pri->supDev.orientation ) {
		pri->defDev.orientation = *pri->supDev.orientation;
	}
	if( pri->supDev.pageSizeID ) {
		pri->defDev.pageSizeID = *pri->supDev.pageSizeID;
	}
	if( pri->supDev.resolution ) {
		pri->defDev.resolutionX = *( pri->supDev.resolution + 0 );
		pri->defDev.resolutionY = *( pri->supDev.resolution + 1 );
	}
}

/*==================================================================*/
#define	SCHEME_STR			"updf"
#define	ATTR_KEY_COPY		"NumCopies" // "MediaCopy"	XXXXXXXXXXXXXXX
#define	ATTR_KEY_ORIENT		"MediaPageRotation"
#define	ATTR_KEY_PGSIZE		"MediaSize"
#define	ATTR_KEY_RESOLUT	"DeviceResolution"

#define	ATTR_VALFRM_PGSIZE	"%fx%f" // "%dx%dmm"	XXXXXXXXXXXXXXX
#define	ATTR_VALFRM_RESOLUT	"deviceResolution_%dx%d"

/*------------------------------------------------------------------*/
static bool	srch_attr_key( char *inf, char *key, char *val )
{
	char	str[128];
	int		cp, sn, n;
	cp = 0;
	sprintf( str, "%s:", SCHEME_STR );
	sn = strlen( str );
	// if( memcmp( inf + cp, str, sn ) )	return( FALSE );
	// cp += sn;
	if( !memcmp( inf + cp, str, sn ) )	cp += sn;	// XXXXXXXXXXXXX
	sprintf( str, "%s=", key );
	sn = strlen( str );
	for( ; ; ) {
		for( ; *( inf + cp ) == ' '; cp++ );
		if( !memcmp( inf + cp, str, sn ) )	break;
		for( ; *( inf + cp ) != ';'; cp++ ) {
			if( !( *( inf + cp ) ) )	return( FALSE );
		}
		cp++;
	}
	for( cp += sn; *( inf + cp ) == ' '; cp++ );
	for( n = 0; *( inf + cp + n ); n++ ) {
		if( *( inf + cp + n ) == ';' ||
			*( inf + cp + n ) == ',' || *( inf + cp + n ) == ' ' )	break;
		if( n >= 64 )	return( FALSE );
		*( val + n ) = *( inf + cp + n );
	}
	*( val + n ) = '\0';
	return( TRUE );
}

/*------------------------------------------------------------------*/
static bool	get_attr_copy( char *inf, int *num )
{
	char	val[128];
	int		n;
	if( !srch_attr_key( inf, ATTR_KEY_COPY, val ) )	return( FALSE );
	if( sscanf( val, "%d", &n ) == 1 ) {
		*num = n;	return( TRUE );
	}
	return( FALSE );
}
static bool	get_attr_orient( char *inf, int *id )
{
	char	val[128];
	if( !srch_attr_key( inf, ATTR_KEY_ORIENT, val ) )	return( FALSE );
	if( !strcmp( val, "portrait" ) ) {	*id = pdDev_orient_portrait;	return( TRUE );	}
	if( !strcmp( val, "landscape" ) ) {	*id = pdDev_orient_landscape;	return( TRUE );	}
	return( FALSE );
}

static bool	chk_pgs_nm( char *inf, char *nm )
{
	char	str[64];
	int		aln, ln, cp;
	aln = strlen( inf );
	sprintf( str, "_%s_", nm );
	ln = strlen( str );
	for( cp = 0; cp <= aln - ln; cp++ ) {
		if( !memcmp( inf + cp, str, ln ) )	return( TRUE );
	}
	return( FALSE );
}
static bool	get_attr_pgsize( char *inf, int *id, int *wx, int *wy )
{
	char	val[128];
	if( !srch_attr_key( inf, ATTR_KEY_PGSIZE, val ) )	return( FALSE );
	if( chk_pgs_nm( val, "a3" ) ) {		*id = pdDev_pgSize_a3;		*wx = *wy = 0;	return( TRUE );	}
	if( chk_pgs_nm( val, "a4" ) ) {		*id = pdDev_pgSize_a4;		*wx = *wy = 0;	return( TRUE );	}
	if( chk_pgs_nm( val, "a5" ) ) {		*id = pdDev_pgSize_a5;		*wx = *wy = 0;	return( TRUE );	}
	if( chk_pgs_nm( val, "b4" ) ) {		*id = pdDev_pgSize_b4;		*wx = *wy = 0;	return( TRUE );	}
	if( chk_pgs_nm( val, "b5" ) ) {		*id = pdDev_pgSize_b5;		*wx = *wy = 0;	return( TRUE );	}
	if( chk_pgs_nm( val, "b6" ) ) {		*id = pdDev_pgSize_b6;		*wx = *wy = 0;	return( TRUE );	}
	if( chk_pgs_nm( val, "hagaki" ) ) {	*id = pdDev_pgSize_hagaki;	*wx = *wy = 0;	return( TRUE );	}
	if( chk_pgs_nm( val, "letter" ) ) {	*id = pdDev_pgSize_letter;	*wx = *wy = 0;	return( TRUE );	}
	if( chk_pgs_nm( val, "legal"  ) ) {	*id = pdDev_pgSize_legal;	*wx = *wy = 0;	return( TRUE );	}
	return( FALSE );
}
static bool	get_attr_resolut( char *inf, int *wx, int *wy )
{
	char	val[128];
	int		x, y;
	if( !srch_attr_key( inf, ATTR_KEY_RESOLUT, val ) )	return( FALSE );
	if( sscanf( val, ATTR_VALFRM_RESOLUT, &x, &y ) == 2 ) {
		*wx = x;	*wy = y;	return( TRUE );
	}
	return( FALSE );
}
/*------------------------------------------------------------------*/
static bool	chk_attr_copy( pd_PrinterInfo *pri, int num )
{
	int		n, m, i, ia, ib;
	for( n = 0; ( i = *( pri->supDev.mediaCopy + n ) ); n++ ) {
		if( i < 0 ) {
			ia = ib = 0;
			for( m = n - 1; m >= 0 && *( pri->supDev.mediaCopy + m ) < 0; m-- );
			if( m >= 0 )	ia = *( pri->supDev.mediaCopy + m );
			for( m = n + 1; *( pri->supDev.mediaCopy + m ) < 0; m++ );
			ib = *( pri->supDev.mediaCopy + m );
			if( ia && ib ) {
				if( num > ia && num < ib )	return( TRUE );
			} else if( ia ) {
				if( num > ia )	return( TRUE );
			} else if( ib ) {
				if( num < ib )	return( TRUE );
			}
		} else {
			if( num == i )	return( TRUE );
		}
	}
	return( FALSE );
}
static bool	chk_attr_orient( pd_PrinterInfo *pri, int id )
{
	int		n, i;
	for( n = 0; ( i = *( pri->supDev.orientation + n ) ); n++ ) {
		if( id == i )	return( TRUE );
	}
	return( FALSE );
}
static bool	chk_attr_pgsize( pd_PrinterInfo *pri, int id )
{
	int		n, i;
	if( id ) {
		for( n = 0; ( i = *( pri->supDev.pageSizeID + n ) ); n++ ) {
			if( id == i )	return( TRUE );
		}
	} else {
		if( pri->supDev.pageSizeXY )	return( TRUE );
	}
	return( FALSE );
}
static bool	chk_attr_resolut( pd_PrinterInfo *pri, int wx, int wy )
{
	int		n, x, y;
	for( n = 0; ( x = *( pri->supDev.resolution + n ) ) &&
				( y = *( pri->supDev.resolution + n + 1 ) ); n += 2 ) {
		if( wx == x && wy == y )	return( TRUE );
	}
	return( FALSE );
}

/*==================================================================*/
/* ----------------------------------------------------------------
 *
 *	pd_SetJobInfo
 *
 * ---------------------------------------------------------------- */
bool	pd_SetJobInfo( pd_PrinterInfo *pri, char *inf )
{
	int		num, id, wx, wy;
	if( get_attr_copy( inf, &num ) ) {
		if( !chk_attr_copy( pri, num ) )	return( FALSE );
		pri->jobDev.mediaCopy = num;
	}
	if( get_attr_orient( inf, &id ) ) {
		if( !chk_attr_orient( pri, id ) )	return( FALSE );
		pri->jobDev.orientation = id;
	}
	if( get_attr_pgsize( inf, &id, &wx, &wy ) ) {
		if( !chk_attr_pgsize( pri, id ) )	return( FALSE );
		pri->jobDev.pageSizeID = id;
		pri->jobDev.pageSizeX = wx;
		pri->jobDev.pageSizeY = wy;
	}
	if( get_attr_resolut( inf, &wx, &wy ) ) {
		if( !chk_attr_resolut( pri, wx, wy ) )	return( FALSE );
		pri->jobDev.resolutionX = wx;
		pri->jobDev.resolutionY = wy;
	}
	return( TRUE );
}
/* ----------------------------------------------------------------
 *
 *	pd_SetDocInfo
 *
 * ---------------------------------------------------------------- */
bool	pd_SetDocInfo( pd_PrinterInfo *pri, char *inf )
{
	int		num, id, wx, wy;
	if( get_attr_copy( inf, &num ) ) {
		if( !chk_attr_copy( pri, num ) )	return( FALSE );
		pri->docDev.mediaCopy = num;
	}
	if( get_attr_orient( inf, &id ) ) {
		if( !chk_attr_orient( pri, id ) )	return( FALSE );
		pri->docDev.orientation = id;
	}
	if( get_attr_pgsize( inf, &id, &wx, &wy ) ) {
		if( !chk_attr_pgsize( pri, id ) )	return( FALSE );
		pri->docDev.pageSizeID = id;
		pri->docDev.pageSizeX = wx;
		pri->docDev.pageSizeY = wy;
	}
	if( get_attr_resolut( inf, &wx, &wy ) ) {
		if( !chk_attr_resolut( pri, wx, wy ) )	return( FALSE );
		pri->docDev.resolutionX = wx;
		pri->docDev.resolutionY = wy;
	}
	return( TRUE );
}
/* ----------------------------------------------------------------
 *
 *	pd_SetPageInfo
 *
 * ---------------------------------------------------------------- */
bool	pd_SetPageInfo( pd_PrinterInfo *pri, char *inf )
{
	int		num, id, wx, wy;
	if( get_attr_copy( inf, &num ) ) {
		if( !chk_attr_copy( pri, num ) )	return( FALSE );
		pri->pageDev.mediaCopy = num;
	}
	if( get_attr_orient( inf, &id ) ) {
		if( !chk_attr_orient( pri, id ) )	return( FALSE );
		pri->pageDev.orientation = id;
	}
	if( get_attr_pgsize( inf, &id, &wx, &wy ) ) {
		if( !chk_attr_pgsize( pri, id ) )	return( FALSE );
		pri->pageDev.pageSizeID = id;
		pri->pageDev.pageSizeX = wx;
		pri->pageDev.pageSizeY = wy;
	}
	if( get_attr_resolut( inf, &wx, &wy ) ) {
		if( !chk_attr_resolut( pri, wx, wy ) )	return( FALSE );
		pri->pageDev.resolutionX = wx;
		pri->pageDev.resolutionY = wy;
	}
	return( TRUE );
}
