/* Copyright (c) 2004 E-D Inc.  All rights reserved.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
/*******************************************************************

  OpenPrinting Vector Printer Driver API : for Canon

 ==================================================================

  Define : Printer Control

********************************************************************/
#ifndef	_PD_PRINTER_H
#define	_PD_PRINTER_H

typedef int bool;

#ifndef FALSE
#define FALSE 0
#endif

#ifndef TRUE
#define TRUE (!FALSE)
#endif

#ifndef NULL
#ifdef __cplusplus
#define NULL 0
#else
#define NULL (void *)0
#endif
#endif

#define	MMETER_PER_INCH	25.4

/* GraphicsState */
#define	pd_LINEDASH_ARYMAX	16
typedef struct _pd_GraphicsState {
	pdCTM			ctm;
	pdColorSpace	colorSpace;
	int				rop;			// < Not Support
	pdFillMode		fillMode;
	float			alpha;			// < Not Support
	pdFix 			lineWidth;
	int				dashNum;
	pdFix			lineDash[pd_LINEDASH_ARYMAX];
	pdFix			dashOffset;
	pdLineStyle		lineStyle;
	pdLineCap		lineCap;
	pdLineJoin		lineJoin;
	pdFix			miterLimit;
	pdPaintMode		paintMode;
	pdBrush 		strokeBrush;
	pdBrush 		fillBrush;
	pdBrush 		bgBrush;		// < Not Support
	bool			pathEntry;		// at EndPath : clear pathActive
	int				pathID;			// Path ID
	bool			clipPathEntry;	// at SetClipPath : move from pathEntry / pathID
	int				clipPathID;		// ClipPath ID
} pd_GraphicsState;

/* GraphicsStateList */
typedef struct _pd_GraphicsStateList {
	struct _pd_GraphicsStateList	*next;
	pd_GraphicsState	gstate;
} pd_GraphicsStateList;

/* DeviceInfo */
enum {
	pdDev_orient_NULL,
	pdDev_orient_portrait,
	pdDev_orient_landscape
};
enum {
	pdDev_pgSize_NULL,
	pdDev_pgSize_a3,
	pdDev_pgSize_a4,
	pdDev_pgSize_a5,
	pdDev_pgSize_b4,
	pdDev_pgSize_b5,
	pdDev_pgSize_b6,
	pdDev_pgSize_hagaki,
	pdDev_pgSize_letter,
	pdDev_pgSize_legal
};
typedef struct _pd_DeviceInfo {
	int		mediaCopy;
	int		orientation;	// pdDev_orient
	int		pageSizeID;		// pdDev_pgSize
	int		pageSizeX;		// mm : at !pageSizeID
	int		pageSizeY;
	int		resolutionX;	// dpi
	int		resolutionY;
} pd_DeviceInfo;

typedef struct _pd_supDevInfo {
	int		*mediaCopy;		// a,-1,b : a <= Num <= b
	int		*orientation;
	int		*pageSizeID;
	bool	pageSizeXY;		// support free size
	int		*resolution;	// X,Y, ...
} pd_supDevInfo;

/* PrinterInfo */
typedef struct _pd_PrinterInfo {
	int				printerContext;
	int				outputFD;
	pd_supDevInfo	supDev;			// support DeviceInfo
	pd_DeviceInfo	defDev;			// default DeviceInfo
	pd_DeviceInfo	curDev;			// current DeviceInfo
	bool			jobStarted;		// StartJob <-> EndJob
	pd_DeviceInfo	jobDev;
	bool			docStarted;		// StartDoc <-> EndDoc
	pd_DeviceInfo	docDev;
	bool			pageStarted;	// StartPage <-> EndPage
	pd_DeviceInfo	pageDev;
	int				curPntX;		// current point
	int				curPntY;
	bool			pathActive;		// NewPath <-> EndPath
	int				pathID;			// Path ID
	int				strPntX;		// Path start point
	int				strPntY;
	bool			drwImgStarted;	// StartDrawImage <-> EndDrawImage
	int				drwImgWidth;
	int				drwImgDepth;
	bool			scnLinStarted;	// StartScanline <-> EndScanline
	int				scnLinPntY;
	bool			rasterStarted;	// StartRaster <-> EndRaster
	int				rasterWidth;
	bool			streamStarted;	// StartStream <-> EndStream
	pd_GraphicsStateList	gstateList;
} pd_PrinterInfo;

/* PrinterInfoList */
typedef struct _pd_PrinterInfoList {
	struct _pd_PrinterInfoList	*next;
	pd_PrinterInfo		printer;
} pd_PrinterInfoList;

/*******************************************************************
  API entry
********************************************************************/
extern int	(*LIPS_Procs[])();
extern pd_supDevInfo	LIPS_supDev;
// extern int	(*Inkjet_Procs[])();
// extern pd_supDevInfo	Inkjet_supDev;

/*******************************************************************
  inside common functions
********************************************************************/
/* for Debug */
void	pd_DebugPrint( const char *, ... );

/* for Data Convert */
int		pd_DtoI( double );
int		pd_Dx_dLcd( double, double, pdCTM * );
int		pd_Dy_dLcd( double, double, pdCTM * );
int		pd_Dx_Lcd( pdFix, pdFix, pdCTM * );
int		pd_Dy_Lcd( pdFix, pdFix, pdCTM * );

/* for Data Management */
pdBrushData	*pd_CreateBrushData( pdBrushData * );
void	pd_EraseBrushData( pdBrushData ** );
bool	pd_CopyBrush( pdBrush *, pdBrush * );
void	pd_ClearGraphicsState( pd_GraphicsState * );
bool	pd_CopyGraphicsState( pd_GraphicsState *, pd_GraphicsState * );

pd_PrinterInfo	*pd_GetPrinterInfo( int );
pd_PrinterInfo	*pd_NewPrinterInfo( int );
bool	pd_DeletePrinterInfo( int );
pd_GraphicsState	*pd_GetGraphicsState( int );
void	pd_InitCTM( pdCTM * );
void	pd_InitBrush( pdBrush *, int, int, int );
void	pd_InitGraphicsState( pd_GraphicsState * );
bool	pd_PushGraphicsState( pd_PrinterInfo * );
bool	pd_PopGraphicsState( pd_PrinterInfo * );

int		pd_GetNewPathID( pd_PrinterInfo * );

/* for Data Out */
bool	pd_DvWrite( pd_PrinterInfo *, char *, size_t );
bool	pd_DvWriteStr( pd_PrinterInfo *, char * );

/* for Printer Control */
void	pd_SetDefDevInfo( pd_PrinterInfo * );
bool	pd_SetJobInfo( pd_PrinterInfo *, char * );
bool	pd_SetDocInfo( pd_PrinterInfo *, char * );
bool	pd_SetPageInfo( pd_PrinterInfo *, char * );

#endif
