/* Copyright (c) 2004 E-D Inc.  All rights reserved.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
/*******************************************************************

  OpenPrinting Vector Printer Driver API : for Canon

 ==================================================================

  Define : PDAPI

********************************************************************/
#ifndef	_PDAPI_H
#define	_PDAPI_H

/* Return Values */
#define	pdOK		0
#define	pdERROR		-1

/* Error Codes */
#define	pdFATALERROR	-1
#define	pdBADREQUEST	-2
#define	pdBADCONTEXT	-3
#define	pdNOTSUPPORTED	-4
#define	pdJOBCANCELED	-5
#define	pdPARAMERROR	-6

/* Fix */
#define	pdFIX_FRACT_WIDTH	8
#define	pdFIX_FRACT_DENOM	(1<<pdFIX_FRACT_WIDTH)
#define	pdFIX_FRACT_TBMSK	(1<<(pdFIX_FRACT_WIDTH-1))
#define	pdFIX_FLOOR_WIDTH	(sizeof(int)*8-pdFIX_FRACT_WIDTH)
typedef struct _pdFix {
        unsigned int    fract : pdFIX_FRACT_WIDTH;
        signed int      floor : pdFIX_FLOOR_WIDTH;
} pdFix;

#define	itoFix(i,fix)	(fix.fract=0,fix.floor=i)
#define	ftoFix(f,fix)	(fix.fract=(f-floor(f))*pdFIX_FRACT_DENOM,fix.floor=floor(f))
#define	Fixtoi(fix)		(fix.floor+((fix.fract&pdFIX_FRACT_TBMSK)?1:0))			// round
#define	Fixtof(fix)		((double)fix.floor+(double)fix.fract/pdFIX_FRACT_DENOM)

/* Basic Types */
typedef struct _pdPoint {
	pdFix	x, y;
} pdPoint;

typedef struct _pdRectangle {
	pdPoint	p0;
	pdPoint	p1;
} pdRectangle;

typedef struct _pdRoundRectangle {
	pdPoint	p0;
	pdPoint	p1;
	pdFix		xellipse, yellipse;
} pdRoundRectangle;

/* Image Formats */
typedef enum _pdImageFormat {
	pdiformatRaw  = 0,
	pdiformatRLE  = 1,
	pdiformatJPEG = 2,
	pdiformatPNG  = 3
} pdImageFormat;

/* Color Presentation */
typedef enum _pdColorMapping {
	pdcmapDirect  = 0,
	pdcmapIndexed = 1
} pdColorMapping;

typedef enum _pdColorSpace {
	pdcspaceBW            = 0,
	pdcspaceDeviceGray    = 1,
	pdcspaceDeviceCMY     = 2,
	pdcspaceDeviceCMYK    = 3,
	pdcspaceDeviceRGB     = 4,
	pdcspaceStandardRGB   = 5,
	pdcspaceStandardRGB64 = 6
} pdColorSpace;

/* Fill, Paint, Clip */
typedef enum _pdFillMode {
	pdfillModeEvenOdd = 0,
	pdfillModeWinding = 1
} pdFillMode;

typedef enum _pdPaintMode {
	pdpaintModeOpaque      = 0,
	pdpaintModeTransparent = 1
} pdPaintMode;

typedef enum _pdClipRule {
	pdclipRuleEvenOdd = 0,
	pdclipRuleWinding = 1
} pdClipRule;

/* Line */
typedef enum _pdLineStyle {
	pdlineStyleSolid = 0,
	pdlineStyleDash  = 1
} pdLineStyle;

typedef enum _pdLineCap {
	pdlineCapButt   = 0,
	pdlineCapRound  = 1,
	pdlineCapSquare = 2
} pdLineCap;

typedef enum _pdLineJoin {
	pdlineJoinMiter = 0,
	pdlineJoinRound = 1,
	pdlineJoinBevel = 2
} pdLineJoin;

/* Brush */
typedef enum _pdBrushDataType {
	pdbdtypeNormal = 0
} pdBrushDataType;

typedef struct _pdBrushData {
	pdBrushDataType	type;
	int		width, height, pitch;
	char	data[];
} pdBrushData;

typedef struct _pdBrush {
	pdColorSpace	colorSpace;
	int			color[4];		// aRGB quadruplet
	int			xorg, yorg;		// brush origin
								// ignored for LIPS_SetBgColor
	pdBrushData	*pbrush;		// pointer to brush data
								// solid brush used, if null
} pdBrush;

/* Misc. Flags */
#define	pdArc				0
#define	pdChord				1
#define	pdPie				2
#define	pdClockwise			0
#define	pdCounterclockwise	1
#define	pdPathClose			0	// Close path upon LIPS_LinePath
#define	pdPathOpen			1	// Do not close path upon LIPS_LinePath

/* CTM */
typedef struct _pdCTM {
	float a, b, c, d, e, f;
} pdCTM;

/* Device Information and Capabilites */
typedef enum _pdQueryInfoFlags {
	pdQFDeviceResolution = 0x00000001,
	pdQFMediaSize        = 0x00000002,
	pdQFPageRotation     = 0x00000004,
	pdQFMediaNUp         = 0x00000008,
	pdQFMediaDuplex      = 0x00000010,
	pdQFMediaSource      = 0x00000020,
	pdQFMediaDestination = 0x00000040,
	pdQFMediaType        = 0x00000080,
	pdQFMediaCopy        = 0x00000100,
	pdQFPrintRegion      = 0x00010000	// only for QueryDeviceInfo
} pdQueryInfoFlags;

/* Set Error Cord */
extern int	errorno;

/** API Entry **/
/* Creating and Managing Print Contexts */
int OpenPrinter( int outputFD, char *printerModel, int *nApiEntry, int (**apiEntry[])() );
int ClosePrinter( int printerContext );

/* Job Control Operations */
int QueryDeviceCapability( int printerContext, int queryflag, int buflen, char *infoBuf );
int QueryDeviceInfo( int printerContext, int queryflag, int buflen, char *infoBuf );

/** API Entry No. **/
enum {
	/* Creating and Managing Print Contexts */
	pdAPIn_OpenPrinter,
	pdAPIn_ClosePrinter,

	/* Job Control Operations */
	pdAPIn_StartJob,
	pdAPIn_EndJob,
	pdAPIn_StartDoc,
	pdAPIn_EndDoc,
	pdAPIn_StartPage,
	pdAPIn_EndPage,
	pdAPIn_QueryDeviceCapability,
	pdAPIn_QueryDeviceInfo,

	/* Graphics State Object Operations */
	pdAPIn_ResetCTM,
	pdAPIn_SetCTM,
	pdAPIn_GetCTM,
	pdAPIn_InitGS,
	pdAPIn_SaveGS,
	pdAPIn_RestoreGS,
	pdAPIn_QueryColorSpace,
	pdAPIn_SetColorSpace,
	pdAPIn_GetColorSpace,
	pdAPIn_QueryROP,
	pdAPIn_SetROP,
	pdAPIn_GetROP,
	pdAPIn_SetFillMode,
	pdAPIn_GetFillMode,
	pdAPIn_SetAlphaConstant,
	pdAPIn_GetAlphaConstant,
	pdAPIn_SetLineWidth,
	pdAPIn_GetLineWidth,
	pdAPIn_SetLineDash,
	pdAPIn_GetLineDash,
	pdAPIn_SetLineDashOffset,
	pdAPIn_GetLineDashOffset,
	pdAPIn_SetLineStyle,
	pdAPIn_GetLineStyle,
	pdAPIn_SetLineCap,
	pdAPIn_GetLineCap,
	pdAPIn_SetLineJoin,
	pdAPIn_GetLineJoin,
	pdAPIn_SetMiterLimit,
	pdAPIn_GetMiterLimit,
	pdAPIn_SetPaintMode,
	pdAPIn_GetPaintMode,
	pdAPIn_SetStrokeColor,
	pdAPIn_SetFillColor,
	pdAPIn_SetBgColor,

	/* Path Operations */
	pdAPIn_NewPath,
	pdAPIn_EndPath,
	pdAPIn_StrokePath,
	pdAPIn_FillPath,
	pdAPIn_StrokeFillPath,
	pdAPIn_SetClipPath,
	pdAPIn_ResetClipPath,
	pdAPIn_SetCurrentPoint,
	pdAPIn_LinePath,
	pdAPIn_PolygonPath,
	pdAPIn_RectanglePath,
	pdAPIn_RoundRectanglePath,
	pdAPIn_BezierPath,
	pdAPIn_ArchPath,

	/* Text Operations */
	pdAPIn_DrawBitmapText,

	/* Bitmap Image Operations */
	pdAPIn_DrawImage,
	pdAPIn_StartDrawImage,
	pdAPIn_TransferDrawImage,
	pdAPIn_EndDrawImage,

	/* Scan Line Operations */
	pdAPIn_StartScanline,
	pdAPIn_Scanline,
	pdAPIn_EndScanline,

	/* Raster Image Operations */
	pdAPIn_StartRaster,
	pdAPIn_TransferRasterData,
	pdAPIn_SkipRaster,
	pdAPIn_EndRaster,

	/* Stream Data Operations */
	pdAPIn_StartStream,
	pdAPIn_TransferStreamData,
	pdAPIn_EndStream,

	/* API Entry Count */
	pdAPI_COUNT
};

#endif
