/* EpsonPageColor.c -- Vector Driver for EPSON Color Laser Printer.
 * Copyright (C) 2003 EPSON KOWA Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: EpsonPageColor.c,v 1.94 2004/02/23 09:48:11 kitayama Exp $ */

/**
 * @file EpsonPageColor.c
 * @brief EPSON ESC-Page Color ɥ饤
 * @author EPSON KOWA Corporation.
 * @date Mon Feb 23 2004
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* system include files */
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
/* local */
#include "pdapi.h"

/**
 * OSˤäƷɤ¸ߤ
 */
typedef int bool;

#ifndef FALSE
#define FALSE 0
#endif

#ifndef TRUE
#define TRUE (!FALSE)
#endif

#ifndef NULL
#ifdef __cplusplus
#define NULL 0
#else
#define NULL (void *)0
#endif
#endif

#define	GS		(0x1d)
#define	ESC_GS		"\035"
#define	ESC_FF		"\014"
#define	ESC_CR		"\015"
#define	ESC_LF		"\012"
#define	ESC_BS		"\010"

#define	POINT					72
#define	MMETER_PER_INCH				25.4

#define	ESCPAGE_DEFAULT_WIDTH			(4840)
#define	ESCPAGE_DEFAULT_HEIGHT			(6896)

#if 0
#define	ESCPAGE_DEFAULT_WIDTH			(4720 + 5. / (MMETER_PER_INCH / POINT))
#define	ESCPAGE_DEFAULT_HEIGHT			(6772 + 5. / (MMETER_PER_INCH / POINT))

#define	ESCPAGE_DEFAULT_WIDTH			(4960 - 5. / (MMETER_PER_INCH / POINT))
#define	ESCPAGE_DEFAULT_HEIGHT			(7016 - 5. / (MMETER_PER_INCH / POINT))
#endif

#define	ESCPAGE_LEFT_MARGIN_DEFAULT		5. / (MMETER_PER_INCH / POINT)
#define	ESCPAGE_BOTTOM_MARGIN_DEFAULT		5. / (MMETER_PER_INCH / POINT)
#define	ESCPAGE_RIGHT_MARGIN_DEFAULT		5. / (MMETER_PER_INCH / POINT)
#define	ESCPAGE_TOP_MARGIN_DEFAULT		5. / (MMETER_PER_INCH / POINT)

#define	BUFCHECK(x)	\
	if (strlen(obuf) > x) {	\
	    (void)DebugPrint("overflow %d : LINE %d\n", strlen(obuf), __LINE__); \
	}

#define ESCPAGE_OPTION_MANUALFEED		"ManualFeed"		/* ѻȥ쥤 */
#define ESCPAGE_OPTION_CASSETFEED		"Casset"		/* ȥ쥤 */
#define ESCPAGE_OPTION_FACEUP			"FU"			/* ӻȥ쥤 Face Up */
#define ESCPAGE_OPTION_FACEDOWN			"FD"			/* ӻȥ쥤 Face Down */
#define ESCPAGE_OPTION_DUPLEX			"Duplex"		/* ξ̰ */
#define ESCPAGE_OPTION_DUPLEX_TUMBLE		"Tumble"		/* Ȥ */
#define ESCPAGE_OPTION_MEDIATYPE		"MediaType"		/*  */
#define	ESCPAGE_OPTION_RIT			"RITOff"		/* RIT */
#define	ESCPAGE_OPTION_LANDSCAPE		"Landscape"		/* LANDSCAPE */
#define ESCPAGE_OPTION_TONERDENSITY		"TonerDensity"		/* ȥʡǻ */
#define ESCPAGE_OPTION_TONERSAVING		"TonerSaving"		/* ȥʡ */
#define ESCPAGE_OPTION_COLLATE			"Collate"		/*  */

#define ESCPAGE_TUMBLE_DEFAULT			FALSE			/* Long age */
#define ESCPAGE_RIT_DEFAULT			FALSE
#define ESCPAGE_FACEUP_DEFAULT			FALSE
#define ESCPAGE_FACEUP_DEFAULT			FALSE

#define ESCPAGE_MEDIATYPE_DEFAULT		0			/* NORMAL */
#define ESCPAGE_MEDIACHAR_MAX			32

#define ESCPAGE_MANUALFEED_DEFAULT		FALSE
#define ESCPAGE_CASSETFEED_DEFAULT		0

#define	ESCPAGE_DPI_MIN				60
#define	ESCPAGE_DPI_MAX				600
#define	ESCPAGE_DPI_SUPERFINE			1200

#define ESCPAGE_A3_HEIGHT			1190
#define ESCPAGE_A3_WIDTH			842
#define ESCPAGE_POSTCARD_HEIGHT			419
#define ESCPAGE_POSTCARD_WIDTH			284
#define ESCPAGE_LETTER_HEIGHT			792
#define ESCPAGE_LETTER_WIDTH			612
#define ESCPAGE_LEDGER_HEIGHT			1224
#define ESCPAGE_LEDGER_WIDTH			ESCPAGE_LETTER_HEIGHT

#define ESCPAGE_HEIGHT_MAX			ESCPAGE_A3_HEIGHT
#define ESCPAGE_WIDTH_MAX			ESCPAGE_A3_WIDTH
#define ESCPAGE_HEIGHT_MIN			ESCPAGE_POSTCARD_HEIGHT
#define ESCPAGE_WIDTH_MIN			ESCPAGE_POSTCARD_WIDTH

#define EP_COMPRESS5				(5 << 16)
#define EP_COMPRESS20				(20 << 16)
#define EP_COMPRESS30				(30 << 16)

#define RES1200					ESCPAGE_DPI_SUPERFINE
#define RES600					ESCPAGE_DPI_MAX
#define RES300					300

#define JPN					TRUE
#define ENG					FALSE


#define ESCPAGE_DEFAULT_PRINTER         "default" /* ǥեȥץ̾ */
#define ESCPAGE_OPTION_FACE_UP		"FACEUP" /* ӻȥ쥤 Face Up */
#define ESCPAGE_OPTION_FACE_DOWN	"FACEDOWN" /* ӻȥ쥤 Face Down */

#define ESCPAGE_OPTION_ON		"ON"
#define ESCPAGE_OPTION_OFF		"OFF"
#define ESCPAGE_OPTION_YES		"YES"
#define ESCPAGE_OPTION_NO		"NO"

/** ѻ極ơ֥ */
typedef struct PaperTable_s
{
    int         width;                  /**< paper width (unit: point) */
    int         height;                 /**< paper height (unit: point) */
    int         escpage;                /**< number of papersize in ESC/PAGE */
    char        *name;                  /**< Paper Name */
} PaperTable;

/** ѻ極 */
static const PaperTable EpPaperTable[] =
{
    {595, 842, 14, "A4"},               /**< A4 */
    {516, 729, 25, "B5"},               /**< B5 JIS */
    {842, 1190, 13, "A3"},              /**< A3 */
#if 0
    {421, 595, 15, "A5"},               /**< A5 */
    {709, 1002, 24, "B4"},              /**< B4 */
    {729, 1032, 24, "B4"},              /**< B4 JIS */
    {612, 792, 30, "LT"},               /**< Letter */
    {396, 612, 31, "HLT"},              /**< Half Letter */
    {612, 1008, 32, "LGL"},             /**< Legal */
    {522, 756, 33, "EXE"},              /**< Executive */
    {612, 936, 34, "GLG"},              /**< Government Letter (LPD.1) */
    {576, 756, 35, "GLT"},              /**< Government Legal (LPD.1) */
    {792, 1224, 36, "B"},               /**< Ledger */
    {593, 935, 37, "F4"},               /**< F4 */
    {284, 419, 38, "POSTCARD"},         /**< PostCard */
    {298, 666, 64, "YOU4"},             /**< Japanese Envelope You4 */
    {933, 1369, 72, "A3PLUS"},          /**< A3 NOBI */
    {279, 540, 80, "MON"},              /**< Monarch */
    {297, 684, 81, "C10"},              /**< Commercial 10 */
    {312, 624, 90, "DL"},               /**< DL */
    {459, 649, 91, "C5"},               /**< Envelope C5 */
    {501, 709, 99, "IB5"},              /**< Envelope ISO B5 */
#endif /* 0 */
    {0, 0, -1, NULL}		/**< Undefined */
};

/** Vector Driver API Proc. Entries */
typedef struct _OPVP_api_procs {
        int     (*OpenPrinter)(int,char *,int *,struct _OPVP_api_procs **);
        int     (*ClosePrinter)(int);
        int     (*StartJob)(int,char *);
        int     (*EndJob)(int);
        int     (*StartDoc)(int,char *);
        int     (*EndDoc)(int);
        int     (*StartPage)(int,char *);
        int     (*EndPage)(int);
	int	(*QueryDeviceCapability)(int, int, int, char *);
	int	(*QueryDeviceInfo)(int, int, int, char *);
        int     (*ResetCTM)(int);
        int     (*SetCTM)(int,CTM *);
        int     (*GetCTM)(int,CTM *);
        int     (*InitGS)(int);
        int     (*SaveGS)(int);
        int     (*RestoreGS)(int);
        int     (*QueryColorSpace)(int,ColorSpace *,int *);
        int     (*SetColorSpace)(int,ColorSpace);
        int     (*GetColorSpace)(int,ColorSpace *);
        int     (*QueryROP)(int,int *,int *);
        int     (*SetROP)(int,int);
        int     (*GetROP)(int,int *);
        int     (*SetFillMode)(int,FillMode);
        int     (*GetFillMode)(int,FillMode *);
        int     (*SetAlphaConstant)(int,float);
        int     (*GetAlphaConstant)(int,float *);
        int     (*SetLineWidth)(int,Fix);
        int     (*GetLineWidth)(int,Fix *);
        int     (*SetLineDash)(int,Fix *,int);
        int     (*GetLineDash)(int,Fix *,int *);
        int     (*SetLineDashOffset)(int,Fix);
        int     (*GetLineDashOffset)(int,Fix *);
        int     (*SetLineStyle)(int,LineStyle);
        int     (*GetLineStyle)(int,LineStyle *);
        int     (*SetLineCap)(int,LineCap);
        int     (*GetLineCap)(int,LineCap *);
        int     (*SetLineJoin)(int,LineJoin);
        int     (*GetLineJoin)(int,LineJoin *);
        int     (*SetMiterLimit)(int,Fix);
        int     (*GetMiterLimit)(int,Fix *);
        int     (*SetPaintMode)(int,PaintMode);
        int     (*GetPaintMode)(int,PaintMode *);
        int     (*SetStrokeColor)(int,Brush *);
        int     (*SetFillColor)(int,Brush *);
        int     (*SetBgColor)(int,Brush *);
        int     (*NewPath)(int);
        int     (*EndPath)(int);
        int     (*StrokePath)(int);
        int     (*FillPath)(int);
        int     (*StrokeFillPath)(int);
        int     (*SetClipPath)(int,ClipRule);
	int	(*ResetClipPath)(int);
        int     (*SetCurrentPoint)(int,Fix,Fix);
        int     (*LinePath)(int,int,int,Point *);
        int     (*PolygonPath)(int,int,int *,Point *);
        int     (*RectanglePath)(int,int,Rectangle *);
        int     (*RoundRectanglePath)(int,int,RoundRectangle *);
        int     (*BezierPath)(int,int *,Point *);
        int     (*ArcPath)(int,int,int,Fix,Fix,Fix,Fix,
                           Fix,Fix,Fix,Fix);
        int     (*DrawBitmapText)(int,int,int,int,void *);
        int     (*DrawImage)(int,int,int,int,
                             ImageFormat,Rectangle,int,void *);
        int     (*StartDrawImage)(int,int,int,int,
                                  ImageFormat,Rectangle);
        int     (*TransferDrawImage)(int,int,void *);
        int     (*EndDrawImage)(int);
        int     (*StartScanline)(int,int);
        int     (*Scanline)(int,int,int *);
        int     (*EndScanline)(int);
        int     (*StartRaster)(int,int);
        int     (*TransferRasterData)(int,int,unsigned char *);
        int     (*SkipRaster)(int,int);
        int     (*EndRaster)(int);
        int     (*StartStream)(int);
        int     (*TransferStreamData)(int,int,void *);
        int     (*EndStream)(int);
} OPVP_api_procs;		/**< _OPVP_api_procs̾ */

/** ꥹȤΥΡ */
typedef struct ItemList_s {
	void *item;		/**< Ρɤݻǡ(ƥ) */
	size_t length;		/**< itemĹ */
	struct ItemList_s *next; /**< ΥΡɤؤΥݥ */
} ItemList;

/** ˴ؤƤξ빽¤Ρ
 * @note ­ʤФϡɲä
 */
typedef struct GraphicsState_s {
	/* Graphic State Object Operations */
	CTM		ctm;	/**< CTM */
	ColorSpace	colorSpace; /**< 顼ڡ */
	FillMode 	fillMode; /**< եμ */
	Fix 		lineWidth; /**< 饤 */
	LineCap 	lineCap; /**< 饤νüμ */
	LineJoin 	lineJoin; /**< 饤ܹμ */
	Brush 		strokeBrush; /**< ȥŬѤ֥饷ѥ */
	Brush 		fillBrush; /**< եŬѤ֥饷ѥ */
	LineStyle	lineStyle; /**< 饤μ */
	Fix		lineDash[255]; /**< ޤϹĹ255 */
	int		dashNum; /**< åѥο */
	Fix		dashOffset; /**< åѥΥեå */
	Fix		miterLimit; /**< ޥߥå */
	int		rop;	/**< ROP */
	int		ropChanged; /**< ROPͤѹե饰 */
	PaintMode	paintMode; /**< ɤ٤μ */
	int		strokePatternChanged; /**< ȥѥѹե饰 */
	int		fillPatternChanged; /**< եѥѹե饰 */
	int		bgColor[4];		/**< aRGB quadruplet */
	/** Path */
	bool        pathActive; /**< PathǤTRUEǤʤFALSE */
	ItemList    *pathList; /**< PathޥɤǼꥹȤƬɥ쥹 */
	ItemList    *pathEndPoint; /**< PathޥɥꥹȤνüɥ쥹 */
	/** Image */
	bool        imageActive; /**< ᡼ǤTRUEǤʤFALSE */
	ImageFormat imageFormat; /**< ᡼Υեޥå */
	int         imageWidth; /**< ᡼β */
	int         imageHeight; /**< ᡼ν */
	int         imageDepth; /**< ᡼ο */
} GraphicsState;

/** GraphicsState δѥꥹ
 *  @note next ǤɤäǸ夬Ȥ GraphicsState ɽ
 */
typedef struct GraphicsStateList_s {
	GraphicsState			*gstate; /**< Graphics State */
	struct GraphicsStateList_s	*next; /**< ΥΡɤؤΥݥ */
} GraphicsStateList;

/** Job δѹ¤Ρ
 * @todo ­ʤФϡɲä
 */
typedef struct JobInfo_s {
	GraphicsStateList		gstateList;	/**< Graphics State */
} JobInfo;

/** ץ󥿤ΥǥХɽ
 * @noteĤ printerContext ˣĤ¸ߤ롣
 */
typedef struct DeviceInfo_s {
	unsigned char	*printerName;	/**< ץ̾ */
	int		resolution;	/**< ֲ */
	int		maxRes;		/**<  */
	int		country;	/**< ̾ */
	char		*duplex;	/**< ξ̰ */
	char		*face;		/**< ե */

} DeviceInfo;


/** ץ󥿤Ƥξ뤿ι¤Τɽ
 * @note Ĥ printerContext ˣĤ¸ߤ
 */
typedef struct Printer_s {
	DeviceInfo		dev;			/**< Device */
	JobInfo			job;			/**< job infomation */
	int			outputFD;		/**< ǡν */
	bool			jobStarted;		/**< StartJobλƤ뤫 */
//	PageInfo		page;			/**< page information */
//	DocInfo			doc;			/**< document information */
} Printer;

/** Printer ΥꥹȤ뤿ι¤Τɽ */
typedef struct PrinterList_s {
	Printer			printer;		/**< Device */
	int			printerContext;		/**< printerContext */
	struct PrinterList_s	*next; /**< ΥΡɤؤΥݥ */
} PrinterList;


/* ---------------- global valiable(s) ---------------- */
/** Printer δ */
PrinterList		*gPrinterList = NULL;


/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

/* static functions */
#ifdef __cplusplu
extern "C" {
#endif /* __cplusplus */
static void DebugPrint(const char *fmt, ...);
static ItemList *NewList(void *item, size_t length);
static void DeleteList(ItemList *node);
static ItemList *AddList(void *list, void *item, size_t length);
static void DeleteListAll(ItemList *node);
static int Write(int printerContext, const void *buf, size_t nBytes);
static int CheckWrite(int printerContext, const void *buf, size_t nBytes);

static Printer *GetPrinter(int printerContext);
static GraphicsState *GetGraphicsState(int printerContext);
static int NewPrinterContext(void);
static DeviceInfo *GetDeviceInfo(int printerContext);

static int WriteLineAttributes(int printerContext);
static int WriteLineKind(int printerContext);
static int WriteMiterLimit(int printerContext);
static int CopyBrush(Brush* dest, Brush* src);

static int SetPrinterName(int printerContext, char *printerModel);

static int RowBytes8(int x);
static int CopyAreaByBits(char *src, char *dst, int sx, int sy, int dx, int dy, int copydotsH, int copydotsW, int width);
static int ShiftOrigin(int xorg, int yorg, int width, int height, char *srcData, char *distData);

static int StartRaws(int printerContext, int sourceWidth, int sourceHeight, int colorDepth, int destinationWidth, int destinationHeight);
static int WriteRaws(int printerContext, int ndata, void *rawsData);
static int EndRaws(int printerContext);
static int ToByteBoundary(int srcWidthBytes, int distWidthBytes, int height, char *srcData, char *distData);
#ifdef __cplusplus
}
#endif /* __cplusplus */

/** ɥ饤ФǡݡȤƤؿꥹ */
/* static const int (*VectorProcs[])() = */
static const void *VectorProcs[] =
{
	/* Creating and Managing Print Contexts */
	OpenPrinter,
	ClosePrinter,

	/* Job Control Operations */
	StartJob,
	EndJob,
	StartDoc,
	EndDoc,
	StartPage,
	EndPage,
	NULL, /* QueryDeviceCapability, */
	NULL, /* QueryDeviceInfo, */
	/* Graphics State Object Operations */
	ResetCTM,
	SetCTM,
	GetCTM,
	InitGS,
	SaveGS,
	RestoreGS,
	QueryColorSpace,
	SetColorSpace,
	GetColorSpace,
	QueryROP,
	SetROP,
	GetROP,
	SetFillMode,
	GetFillMode,
	NULL, /* SetAlphaConstant, */
	NULL, /* GetAlphaConstant, */
	SetLineWidth,
	GetLineWidth,
	SetLineDash,
	GetLineDash,
	SetLineDashOffset,
	GetLineDashOffset,
	SetLineStyle,
	GetLineStyle,
	SetLineCap,
	GetLineCap,
	SetLineJoin,
	GetLineJoin,
	SetMiterLimit,
	GetMiterLimit,
	SetPaintMode,
	GetPaintMode,
	SetStrokeColor,
	SetFillColor,
	SetBgColor,

	/* Path Operations */
	NewPath,
	EndPath,
	StrokePath,
	FillPath,
	StrokeFillPath,
	SetClipPath,
	ResetClipPath,
	SetCurrentPoint,
	LinePath,
	NULL, /* PolygonPath, */
	RectanglePath,
	RoundRectanglePath,
	BezierPath,
	NULL, /* ArchPath, */

	/* Text Operations */
	DrawBitmapText,

	/* Bitmap Image Operations */
	NULL, /* DrawImage, */
	StartDrawImage,
	TransferDrawImage,
	EndDrawImage,

	/* Scan Line Operations */
	NULL, /* StartScanline, */
	NULL, /* Scanline, */
	NULL, /* EndScanline, */

	/* Raster Image Operations */
	StartRaster, /* StartRaster, */
	TransferRasterData, /* TransferRasterData, */
	NULL, /* SkipRaster, */
	EndRaster, /* EndRaster, */

	/* Stream Data Operations */
	NULL, /* StartStream, */
	NULL, /* TransferStreamData, */
	NULL, /* EndStream */
};

/** 顼ɳǼѤΥХѿ */
int	errorno;

/* ----------------------------------------------------------------

    Static Functions

 * ---------------------------------------------------------------- */

/** Debugѥåϴؿ
 * @param *fmt	printfؿƱΥեޥå
 * @param ...	fmt
 * @return ̵
 */
static void
DebugPrint(const char *fmt, ...)
{
#ifdef DEBUG
	va_list ap;

	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	fflush(stderr);
#endif
	return;
}


/** ꥹȤΥΡɤ
 * @param *item		item(data)ǼؤΥݥ
 * @param length	itemΥ
 * @return		кΡɤؤΥݥ󥿤򡢼ԤʤNULL֤
 */
static ItemList *NewList(void *item, size_t length)
{
	ItemList *node;

	if (item == NULL) {
		/* ꥹȤ֤꤬NULLǤ */
		errorno = FATALERROR;
		return (NULL);
	}

	node = (ItemList *)malloc(sizeof(ItemList));
	if (node == NULL) {
		errorno = FATALERROR;
		return (NULL);
	}

	if (length == 0) {
		node->item = NULL;
	} else {
		node->item = (ItemList *)malloc(length);
		if (node->item == NULL) {
			free(node);
			errorno = FATALERROR;
			return (NULL);
		}
		node->item = memcpy(node->item, item, length);
	}

	node->length = length;
	node->next = NULL;

	return (node);
}


/** ꥹȤΥΡɤ
 * @param *node	ItemList¤ΤؤΥݥ
 * @return	̵
 */
static void DeleteList(ItemList *node)
{
	if (node != NULL) {
		if (node->item != NULL) {
			free(node->item);
		}
		free(node);
	}
	return;
}


/** ΡɤꥹȤɲä
 * @param *list		listؤΥݥ
 * @param *item		item(data)ǼؤΥݥ
 * @param length	itemΥ
 * @return		кΡɤؤΥݥ󥿤򡢼ԤʤNULL֤
 */
static ItemList *AddList(void *list, void *item, size_t length)
{
	ItemList *node;

	if (list == NULL) {
		errorno = FATALERROR;
		return (NULL);
	}

	node = list;
	while (node->next != NULL) {
		node = node->next;
	}

	node->next = NewList(item, length);
	return (node->next);
}


/** ꥹȤ˴ޤޤƤΥΡɤ
 * @param *node	ꥹȤƬItemList¤ΤؤΥݥ
 * @return	̵
 */
static void DeleteListAll(ItemList *node)
{
	if (node == NULL) {
		errorno = FATALERROR;
		return;
	}

	while (node != NULL) {
		ItemList *next;

		next = node->next;
		DeleteList(node);
		node = next;
	}
	return;
}


/** printerContext ǻꤵ줿ץ󥿾ؤΥݥ󥿤֤
 * @param printerContext оݤΥץ󥿥ƥ
 * @retval  ץ󥿾ؤΥݥ
 * @retval  NULL
 */
static Printer *GetPrinter(int printerContext)
{
	PrinterList		*list;

	// DebugPrint("pdapi: %s,%d (printerContext=%d)\n", __FUNCTION__, __LINE__, printerContext);

	for (list = gPrinterList; list != NULL; list = list->next) {
		if (list->printerContext == printerContext) {
			break;
		}
	}
	if (list == NULL) {
		errorno = BADCONTEXT;
		return(NULL);
	}
	return(&(list->printer));
}


/** printerContext ǻꤵ줿 Graphics State ؤΥݥ󥿤֤
 * @param printerContext оݤΥץ󥿥ƥ
 * @retval  ꤵ줿 Graphics State ؤΥݥ
 * @retval  NULL
 */
static GraphicsState *GetGraphicsState(int printerContext)
{
	Printer				*printer;
	GraphicsStateList	*gl;

	printer = GetPrinter(printerContext);
	if (printer == NULL) {
		return(NULL);
	}

	for (gl = &(printer->job.gstateList); gl->next != NULL; gl = gl->next) {
		;
	}

	return(gl->gstate);
}


/** graphics state 򥳥ԡ
 * @param *dst ԡ
 * @param *src ԡ
 * @return ʤ
 */
static void CopyGraphicsState(GraphicsState *dst, GraphicsState *src)
{
	BrushData		*strokeBrush;
	BrushData		*fillBrush;

	/*  */
	strokeBrush = dst->strokeBrush.pbrush;
	fillBrush = dst->fillBrush.pbrush;

	memcpy(dst, src, sizeof(GraphicsState));

	/*  */
	dst->strokeBrush.pbrush = strokeBrush;
	dst->fillBrush.pbrush = fillBrush;

	/* (TBD)
	 * ǡѥꥹȤɤΤ褦˰Ĵɬ
	 */
	/*  */
	dst->pathList = NULL;
	dst->pathEndPoint = NULL;

	return;
}


/** create a new graphics state
 * @param ʤ
 * @retval  ˺줿 Graphics State ؤΥݥ
 * @retval  NULL
 */
static GraphicsState *NewGraphicsState(void)
{
	GraphicsState	*gs;
	BrushData		*brushData;

	gs = malloc(sizeof(GraphicsState));
	if (gs == NULL) {
		errorno = FATALERROR;			/* (TBD) */
		return(NULL);
	}

	brushData = malloc(sizeof(BrushData));
	if (brushData == NULL) {
		/* ­ */
		errorno = FATALERROR;			/* (TBD) */

		free(gs);				/* ΰ */

		return(NULL);
	}
	gs->strokeBrush.pbrush = brushData;

	brushData = malloc(sizeof(BrushData));
	if (brushData == NULL) {
		/* ­ */
		errorno = FATALERROR;			/* (TBD) */

		free(gs->strokeBrush.pbrush);		/* ΰ */
		free(gs);

		return(NULL);
	}
	gs->fillBrush.pbrush = brushData;
	gs->pathList = NULL;

	return(gs);
}


/** delete a specified graphics state
 * @param *gs оݤ GraphicsState
 * @return ʤ
 */
static void DeleteGraphicsState(GraphicsState *gs)
{
	if (gs == NULL) {
		return;
	}

	DeleteListAll(gs->pathList);				/* ѥκ */

	free(gs->strokeBrush.pbrush);				/* ֥饷κ */
	free(gs->fillBrush.pbrush);
	free(gs);					/* ˺Ƥޤ  _o_ */

	return;
}


/** push a graphics state and create a new graphics state
 * ꤵ줿 printerContext ˥ȤƱ Graphics State ɲä롣
 * <BR>
 *	(TBD)
 *	ͽǤ "n" ȤʤäƤ뤬ϤäꤷƤʤΤǡ push ̤å
 * @param printerContext оݤΥץ󥿥ƥ
 * @retval  OK
 * @retval  ERROR
 */
static int PushGraphicsState(int printerContext)
{
	Printer				*printer;
	GraphicsStateList	*gl;
	GraphicsState		*gs;

	printer = GetPrinter(printerContext);
	if (printer == NULL) {
		return(ERROR);
	}

	for (gl = &(printer->job.gstateList); gl->next != NULL; gl = gl->next) {
		;
	}
	gl->next = malloc(sizeof(GraphicsStateList));
	if (gl->next == NULL) {
		errorno = JOBCANCELED;
		return(ERROR);
	}
	gl->next->next = NULL;

	gs = NewGraphicsState();
	if (gs == NULL) {
		free(gl->next);					/*  */
		gl->next = NULL;
		errorno = JOBCANCELED;
		return(ERROR);
	}
	gl->next->gstate = gs;

	/*  GraphicsState ˥Ȥ GraphicsState 򥳥ԡ */
	CopyGraphicsState(gs, gl->gstate);

	return(OK);
}


/** pop a graphics state
 * ꤵ줿 printerContext  Graphics State  pop 롣
 * Ȥ Graphics State Ϻ롣
 * @param printerContext оݤΥץ󥿥ƥ
 * @retval  OK
 * @retval  ERROR
 */
static int PopGraphicsState(int printerContext)
{
	Printer				*printer;
	GraphicsStateList	*gl;
	GraphicsStateList	*prev;

	printer = GetPrinter(printerContext);
	if (printer == NULL) {
		return(ERROR);
	}

	prev = &(printer->job.gstateList);
	for (gl = prev; gl->next != NULL; gl = gl->next) {
		prev = gl;
	}
	if (prev == &(printer->job.gstateList)) {
		/* 顼: pop Ǥʤ */
		errorno = PARAMERROR;
		return(ERROR);
	}

	/* (TBD)
	 * ˸ɬפ?
	 * ä ѥꥹȤΰ
	 */

	prev->next = NULL;
	DeleteGraphicsState(gl->gstate);
	free(gl);

	return(OK);
}


/** delete a printer device information.
 * @param *dev оݤΥǥХ
 * @return ʤ
 */
static void DeleteDevice(DeviceInfo *dev)
{

	/* Ϻ٤Τϲʤ */

	return;
}


/** delete a job information.
 * @param *dev оݤ job
 * @return ʤ
 */
static void DeleteJob(JobInfo *job)
{
	GraphicsStateList	*gl;

	/* GraphicsState ϥꥹȤˤʤäƤǽ */
	gl = &(job->gstateList);
	for (gl = gl->next; gl != NULL; gl = gl->next) {
		DeleteGraphicsState(gl->gstate);
	}
	job->gstateList.next = NULL;

	return;
}


/** printerContext ǻꤵ줿ץ󥿾Ѥδ
 * @param printerContext оݤΥץ󥿥ƥ
 * @retval  OK
 * @retval  ERROR
 */
static int DeletePrinterContext(int printerContext)
{
	PrinterList		*list;
	PrinterList		*prev;

	for (list = gPrinterList; list != NULL; list = list->next) {
		if (list->printerContext == printerContext) {
			break;
		}
		prev = list;
	}
	if (list == NULL) {
		errorno = BADCONTEXT;
		return(ERROR);
	}


	if (list == gPrinterList) {
		/* Ƭ */
		gPrinterList = gPrinterList->next;
	} else {
		prev->next = list->next;
	}

	/* Τ list  */
	/* dev ʲ */
	DeleteDevice(&(list->printer.dev));

	/* job ʲ */
	DeleteJob(&(list->printer.job));

	free(list);

	return(OK);
}


/** create a new printer context.
 * @param ʤ
 * @retval  ˺줿ؤ printerContext 
 * @retval ERROR 
 */
static int NewPrinterContext(void)
{
	PrinterList			*list;
	PrinterList			*next;
	GraphicsStateList	*gl;
	int					printerContext;
	int					result;

	/* device ΰγ */
	list = malloc(sizeof(PrinterList));
	if (list == NULL) {
		/* ­ */
		errorno = FATALERROR;			/* (TBD) */
		return(ERROR);
	}

	list->next = NULL;

	if (gPrinterList == NULL) {
		/*  */
		gPrinterList = list;
		printerContext = 0;
		list->printerContext = printerContext;
	} else {
		/* ܰʹ */
		for (next = gPrinterList; next->next != NULL; next = next->next) {
			;
		}
		printerContext = next->printerContext;
		printerContext++;
		list->printerContext = printerContext;
		next->next = list;
	}


	gl = &(list->printer.job.gstateList);
	gl->gstate = NewGraphicsState();
	if (gl->gstate == NULL) {
		/* ­ */
		errorno = FATALERROR;			/* (TBD) */

		(void)DeletePrinterContext(printerContext);

		return(ERROR);
	}
	gl->next = NULL;
	list->printer.jobStarted = FALSE;

	/* Graphics State νԤʤ */
	result = InitGS(printerContext);
	if (result < 0) {
		/* 顼 (ǤȯʤȦǰΰ) */

		(void)DeletePrinterContext(printerContext);

		return(ERROR);
	}

	return(printerContext);
}


/** get a specified device information.
 * @param printerContext оݤΥץ󥿥ƥ
 * @retval  ץ󥿤ΥǥХؤΥݥ
 * @retval  NULL
 */
static DeviceInfo *GetDeviceInfo(int printerContext)
{
	Printer		*printer;

	printer = GetPrinter(printerContext);
	if (printer == NULL) {
		return(NULL);
	}

	return(&(printer->dev));
}


/** 饤°ѹ륳ޥɤ
 * @param printerContext OpenPrinter ֵѤ줿ץ󥿥ƥ
 *			 ͤꤹ
 * @return ˽λȥץ󥿤ΥǥХؤΥݥ󥿤֤롣
 *	   顼ä NULL ֤롣
 *
 * @note ꤵ줿printerContextؤǥХˡGraphicsStatesĥ饤°ͤ
 *	 Ѥơ饤°ѹޥɤ롣
 *	 δؿϡSetLineWidth(), SetLineCap(), SetLineJoin()Τ줾줫ƤӽФ롣
 */
static int WriteLineAttributes(int printerContext)
{
	GraphicsState	*gs;
	Fix		width;
	LineCap		cap;
	LineJoin	join;
	int		escWidth;
	int 		escCap;
	int 		escJoin;

	DebugPrint("pdapi: %s,%d printerContext=%d\n", __FUNCTION__, __LINE__, printerContext);
	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return(ERROR);
	}

	/* °ѹΥɤץ󥿤 */

	/* widthͤϡwidth/720(inch) */
	if (GetLineWidth(printerContext, &width) != OK) {
		return(ERROR);
	}
	{
		float f;
		Fix2f(width, f);
		//DebugPrint("width = %f\n", f);
	}



	escWidth = Fix2i(width);
	{
		float f;
		Fix2f(width, f);
		//DebugPrint("width = %f\n", f);
	}

	if ((escWidth < 0) || (escWidth == 0)) {
		/* ͤϥ꡼롢0°ѹԤʤʤ?
		   פ롣Ǥ褤(TBD)*/
		escWidth = 1;
	}
	{
		float f;
		Fix2f(width, f);
		//DebugPrint("width = %f\n", f);
	}

	/* lineCapμ */
	if (GetLineCap(printerContext, &cap) != OK) {
		return(ERROR);
	}
	switch (cap) {
		case lineCapSquare:
			escCap = 2;
			break;
		case lineCapRound:
			escCap = 1;
			break;
		case lineCapButt:
		default:
			/* defaultSquare */
			escCap = 0;
			break;
	}

	/* lineJoinμ */
	if (GetLineJoin(printerContext, &join) != OK) {
		return(ERROR);
	}
	switch (join) {
		case lineJoinRound:
			escJoin = 1;
			break;
		case lineJoinBevel:
			escJoin = 2;
			break;
		case lineJoinMiter:
		default:
			/* defaultMiter */
			escJoin = 3;
			break;
	}

	/**
	 * @note
	 * ESC/Page Ǥüܹ1ĤΥޥɤˤʤäƤ
	 * ΤǡޤȤ롣ɤ줫1Ĥ°ǤѹС
	 * ޥɤꤲɬפ롣
	 */
	{
		char obuf[128];

		memset(obuf, 0, 128);
		(void)sprintf(obuf, ESC_GS "%d;%d;%dlwG", escWidth, escCap, escJoin);
		DebugPrint("pdapi: %s,%d comm=[%s]\n", __FUNCTION__, __LINE__, obuf+1);
		if (CheckWrite(printerContext, obuf, strlen(obuf)) != OK) {
			return(ERROR);
		}

		/*
		 * MiterLimitͤΥޥϡʲ2ĤΥߥ󥰤ɬס
		 *  1. join=mitermiterlimitͤѹ줿(WriteLineAttributes())
		 *  2. join=miter򤵤줿(SetMiterLimit())
		 */
		if (join == lineJoinMiter) {
			return WriteMiterLimit(printerContext);
		}
	}
	return(OK);
}


/** ѹ륳ޥɤ
 * @param printerContext	OpenPrinter ֵѤ줿ץ󥿥ƥ
 *				ͤꤹ롣
 * @return ˽λȥץ󥿤ΥǥХؤΥݥ󥿤֤롣
 *	   顼ä NULL ֤롣
 *
 * @note ꤵ줿printerContextؤǥХˡGraphicsStates
 *       Ѥơѹޥɤ롣δؿϡ
 *       SetLineStyle(), SetLineDash(), SetLineDashOffset()Τ줾줫
 *       ӽФ롣
 */
static int WriteLineKind(int printerContext)
{
	GraphicsState *gs;
	LineStyle style;
	int num;
	int lineType;
	int lineNum;

	DebugPrint("pdapi: %s,%d printerContext=%d\n", __FUNCTION__, __LINE__, printerContext);

	// GraphicsStateμ
	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return(ERROR);
	}

	if (GetLineStyle(printerContext, &style) != OK) {
		return(ERROR);
	}

	/* Dash̵ξϡ¢0֤Ѥ */
	lineType = 0;
	lineNum = 0;
	num = gs->dashNum;
	if ((style == lineStyleDash) && (num > 0)) {
		/* LineDashѥϿޥȯɬ */
		char tmp[64];
		char obuf[1024]; /* ­?(TBD) */
		char pattern_suffix[1024];
		char *p;
		int i;
		int offset;

		obuf[0] = '\0';
		pattern_suffix[0] = '\0';

		/* Ͽֹ0Ѥ */
		sprintf(tmp, ESC_GS "%d;", 0);
		strcat(obuf, tmp);

		DebugPrint("DashNum=%d\n", num);

		/* Τoffsetͤ򻻽Ф */
		offset = Fix2i(gs->dashOffset);
		if (offset != 0) {
			int size = 0;

			for (i = 0; i < num; i++) {
				size += Fix2i(gs->lineDash[i]);
			}
			if (size != 0) {
				offset %= size;
				offset = (offset < 0) ? (size - offset) : offset;
			} else {
				offset = 0;
			}
		}

		/* ѥϿ */
		for (i = 0; i < num; i++) {
			int length;

			length = Fix2i(gs->lineDash[i]);
			if (offset == 0) {
				/* Ƭ֤Ͽ */
				sprintf(tmp, "%d;", length);
				strcat(obuf, tmp);
			} else if (length < offset) {
				/* 夫³٤̥ХåեϿ */
				sprintf(tmp, "%d;", length);
				strcat(pattern_suffix, tmp);
				offset -= length;
			} else {
				/* ǤĤʬ䤷Ƥ줾prefixsuffix */
				int rest;

				sprintf(tmp, "%d;", offset);
				strcat(pattern_suffix, tmp);

				rest = length - offset;
				if (rest != 0) {
					/* ;꤬ϡǸ0ɲ */
					sprintf(tmp, "%d;", 0);
					strcat(pattern_suffix, tmp);
					if (i % 2) {
						/* OFFӥåȤ鳫 */
						sprintf(tmp, "%d;", 0);
						strcat(obuf, tmp);
					}
					/* ;äʬƬϿ */
					sprintf(tmp, "%d;", rest);
					strcat(obuf, tmp);
				} else if ((i % 2) == 0) {
					/* OFFӥåȤ鳫 */
					sprintf(tmp, "%d;", 0);
					strcat(obuf, tmp);
				}
				offset = 0;
			}
		}
		/* offset줿ʬ夫³ */
		if (pattern_suffix[0] != '\0') {
			strcat(obuf, pattern_suffix);
		}
		p = obuf + strlen(obuf) - 1;
		strcpy(p, "dlG");
		DebugPrint("DashCommand[%s]\n", obuf);

		/* 񤭤 */
		if (CheckWrite(printerContext, obuf, strlen(obuf)) != OK) {
			return(ERROR);
		}
		/* Ͽ0֤Ѥ */
		lineType = 1;
		lineNum = 0;
	}

	/* 򥳥ޥɤ */
	{
		char obuf[64];
		(void)sprintf(obuf, ESC_GS "%d;%dlpG", lineType, lineNum);
		DebugPrint("LineKindCommand[%s]\n", obuf);
		return CheckWrite(printerContext, obuf, strlen(obuf));
	}
}


/** ޥߥåȤꥳޥɤ
 * @param printerContext	OpenPrinter ֵѤ줿ץ󥿥ƥ
 *		                ͤꤹ롣
 * @return ˽λȥץ󥿤ΥǥХؤΥݥ󥿤֤롣
 *	   顼ä NULL ֤롣
 * @note
 *  MiterLimitͤΥޥϡʲ2ĤΥߥ󥰤ɬס
 *  	-# join=mitermiterlimitͤѹ줿
 *  	-# join=miter򤵤줿
 *  
 *  ޥߥåͤꤹˤ lwG ˤ ܹ(n3)  3 ˤʤ
 *  Ƥɬפ롣
 */
static int WriteMiterLimit(int printerContext)
{
	char obuf[128];
	Fix miterLimit;
	int limit;

	DebugPrint("pdapi: %s,%d printerContext=%d\n", __FUNCTION__, __LINE__, printerContext);

	/* ߤmiterlimitμ */
	if (GetMiterLimit(printerContext, &miterLimit) != OK) {
		return(ERROR);
	}

	/* miterlimit0.?ξ硢0ʤäƤޤΤǶŪ1ˤ */
	limit = Fix2i(miterLimit);
	if (limit == 0) {
		limit = 1;
	}
	DebugPrint("%s: %s,%d miterLimit=%x, limit=%x, %d\n",
			   __FILE__, __FUNCTION__, __LINE__, miterLimit, limit, limit );

    (void)sprintf(obuf, ESC_GS "1;%dmlG", limit);
	return CheckWrite(printerContext, obuf, strlen(obuf));
}


/** ǡԤ
 * @param printerContext	OpenPrinter ֵѤ줿ץ󥿥ƥ
 *				ͤꤹ롣
 * @param buf			񤭹ߤԤǡؤΥݥ
 * @param nBytes		񤭹ߤԤǡΥ
 * @return ˽λOK ֤롣顼ä-1 ֤졢
 *	   errorno ˥顼ɤǼ롣
 * @note ƤΥǡ񤭽Фޤǡܴؿϥ֥åޤ
 */
static int Write(int printerContext, const void *buf, size_t nBytes)
{
	Printer *pl;
	int fd;

	pl = GetPrinter(printerContext);
	if (pl == NULL) {
		/* Printer ContextǤ */
		DebugPrint("pdapi: %s,%d\n", __FUNCTION__, __LINE__);
		return (ERROR);
	}

	fd = pl->outputFD;
	while (0 < nBytes) {
		ssize_t wBytes;

		wBytes = write(fd, buf, nBytes);
		if (wBytes < 0) {
			/* ƥॳ륨顼 */
			if((errno != EINTR) && (errno != EAGAIN) && (errno != EIO)) {
				/* ʤ顼ȯ */
				DebugPrint("pdapi: %s,%d\n", __FUNCTION__, __LINE__);
				errorno = FATALERROR;
				return (ERROR);
			}
		} else {
			/* 񤭹 */
			buf += wBytes;
			nBytes -= wBytes;
		}
	}
	return (OK);
}


/** StartJob()λƤ硢ǡν񤭹(Write())Ԥ
 *  λƤʤ硢ʤˤ⤻ｪλ롣
 * @param printerContext	OpenPrinter ֵѤ줿ץ󥿥ƥ
 *				ͤꤹ롣
 * @param buf			񤭹ߤԤǡؤΥݥ
 * @param nBytes		񤭹ߤԤǡΥ
 *
 * @return ˽λOK ֤롣顼ä-1 ֤졢
 *	   errorno ˥顼ɤǼ롣
 */
static int CheckWrite(int printerContext, const void *buf, size_t nBytes) {

	Printer *pPrinter;

	pPrinter = GetPrinter(printerContext);
	if (pPrinter == NULL) {
		return(ERROR);
	}
	/* StartJobϽλƤ뤫 */
	if (!pPrinter->jobStarted) {
		/* ޤޥɤǤˤʤ */
		return (OK);
	}
	return Write(printerContext, buf, nBytes);
}


/** Brush¤ΤΥԡԤʤ
 * @param pDest	ԡBrush¤
 * @param pSrc	ԡBrush¤
 * @return ˽λOK ֤롣顼ä-1 ֤졢
 *	   errorno ˥顼ɤǼ롣
 * @note
 *	Brush¤BrushData¤ΤƤ⥳ԡ롣
 *	pDestϡ˳ݤ줿¤ΤƬɥ쥹Ǥ뤳ȡ
 *	BrushData.dataˤĤƤϡ̤ݡȡNULLꤵ롣
 */
static int CopyBrush(Brush *pDest, Brush *pSrc)
{
	BrushData *pSrcBD;
	BrushData *pDestBD;

	/* ͤɤĤȯʤ */
	if (pDest == NULL || pSrc == NULL) {
		DebugPrint("pdapi: %s,%d pDest=%p or pSrc=%p is NULL.\n", __FUNCTION__, __LINE__, pDest, pSrc);
		errorno = FATALERROR;
		return(ERROR);
	}

	pDestBD = pDest->pbrush;
	pSrcBD = pSrc->pbrush;

	/* ֥饷ǡꤵƤСԡ */
	if (pSrcBD != NULL) {
		/* ԡ˥֥饷ǡѤΥ꤬ݤƤʤСݤ */
		if (pDestBD == NULL) {
			pDestBD = (BrushData *)malloc(sizeof(BrushData));
			if (pDestBD == NULL) {
				/* memoryݼ */
				return (ERROR);
			}
			pDest->pbrush = pDestBD;
		}
		pDestBD->type = pSrcBD->type;
		pDestBD->width = pSrcBD->width;
		pDestBD->height = pSrcBD->height;
		pDestBD->pitch = pSrcBD->pitch;
		pDestBD->data = NULL; /* (TBD) Not Supported this Version. */

	} else {
		/* ԡ˥֥饷ǡѤΥ꤬ݤƤС */
		if (pDestBD != NULL) {
			free (pDestBD);
			pDest->pbrush = NULL;
		}
	}

	pDest->colorSpace = pSrc->colorSpace;
#if 0
	memcpy(pDest->color, pSrc->color, 4);
#else
	pDest->color[0] = pSrc->color[0];
	pDest->color[1] = pSrc->color[1];
	pDest->color[2] = pSrc->color[2];
	pDest->color[3] = pSrc->color[3];
#endif
	pDest->xorg = pSrc->xorg;
	pDest->yorg = pSrc->yorg;

	return(OK);
}


/** set printer name.
 * ץ̾Τ Graphics State ꤷ̾ΤˤäɬפԤʤ
 * ץ̾Τ򥳥ԡƥ˻ġ (ɬפΤʡ)
 * @param printerContext	оݤΥץ󥿥ƥ
 * @param *printerModel ץ󥿥ǥ̾ꤹ(UTF-8)NULL ξ
 *                      ϥɥ饤ФꤹǥեȤΥǥȤʤ롣
 * @retval  OK
 * @retval  ERROR
 */
static int SetPrinterName(int printerContext, char *printerModel)
{

	/** Ƶθͭξν
	 * (TBD)ꤷƤ뤬¾ˤ⤢Ф˵Ҥ */
	struct PrinterDeviceInfo_s {
		char	*name;				/**< ץ̾ (Glue Ϥ̾ΤƱ) */
		int	maxRes;				/**<  */
		int	country;			/**< ̾ */
		char	*duplex;			/**< ξ̰ */
		char	*face;				/*<* ե */
	} PrinterDeviceList[] = {
		{ "lp8000c", RES600, JPN, ESCPAGE_OPTION_OFF, ESCPAGE_OPTION_FACE_UP },
		{ "lp8200c", RES600, JPN, ESCPAGE_OPTION_OFF, ESCPAGE_OPTION_FACE_UP },
		{ "lp8300c", RES600, JPN, ESCPAGE_OPTION_ON, ESCPAGE_OPTION_FACE_UP },
		{ "lp8500c", RES600, JPN, ESCPAGE_OPTION_ON, ESCPAGE_OPTION_FACE_UP },
		{ "lp8800c", RES600, JPN, ESCPAGE_OPTION_ON, ESCPAGE_OPTION_FACE_UP },
		{ "lp9500c", RES600, JPN, ESCPAGE_OPTION_OFF, ESCPAGE_OPTION_FACE_DOWN },
		{ "lp3000c", RES600, JPN, ESCPAGE_OPTION_ON, ESCPAGE_OPTION_FACE_DOWN },
		{ "alc8500", RES600, ENG, ESCPAGE_OPTION_ON, ESCPAGE_OPTION_FACE_UP },
		{ "alc8600", RES600, ENG, ESCPAGE_OPTION_ON, ESCPAGE_OPTION_FACE_UP },
		{ "alc2000", RES600, ENG, ESCPAGE_OPTION_ON, ESCPAGE_OPTION_FACE_DOWN },
		{ "alc4000", RES1200, ENG, ESCPAGE_OPTION_ON, ESCPAGE_OPTION_FACE_DOWN },
		{ "alc1900", RES600, ENG, ESCPAGE_OPTION_ON, ESCPAGE_OPTION_FACE_DOWN },

		/* default 
		 *	κǸˤʤФʤʤ
		 */
		/* FIXME: ǤΥǥե */
		{ ESCPAGE_DEFAULT_PRINTER, RES300, JPN, ESCPAGE_OPTION_OFF, ESCPAGE_OPTION_FACE_DOWN }
	};

	DeviceInfo	*dev;
	int			i;

	DebugPrint("pdapi: %s,%d (printerContext=%d, printerModel=%s\n", __FUNCTION__, __LINE__, printerContext, printerModel);

	dev = GetDeviceInfo(printerContext);
	if (dev == NULL) {
		return(ERROR);
	}

	/* ץ̾ */
	/* ˱ơŬڤͤˤ */
#if 0
	for (i = 0; i < (sizeof(PrinterDeviceList) / sizeof(struct PrinterDeviceInfo_s)); i++) {
		if ((printerModel == NULL) ||
			(PrinterDeviceList[i].name == NULL) ||
			(strcmp(printerModel, PrinterDeviceList[i].name) == 0)) {
			if ((printerModel == NULL) ||
				(PrinterDeviceList[i].name == NULL)) {
				dev->printerName = "default";
				i = sizeof(PrinterDeviceList) - 1;
			} else {
				dev->printerName = PrinterDeviceList[i].name;
			}
			dev->maxRes =  PrinterDeviceList[i].maxRes;
			dev->country  = PrinterDeviceList[i].country;
			dev->duplex = PrinterDeviceList[i].duplex;
			dev->face = PrinterDeviceList[i].face;
		}
	}
#else
	for (i = 0; strcmp(PrinterDeviceList[i].name, ESCPAGE_DEFAULT_PRINTER) != 0; i++) {
		if (printerModel != NULL
		    && strcmp(printerModel, PrinterDeviceList[i].name) == 0) {
			break;
		}
	}

	dev->printerName = PrinterDeviceList[i].name;
	dev->maxRes = PrinterDeviceList[i].maxRes;
	dev->country  = PrinterDeviceList[i].country;
	dev->duplex = PrinterDeviceList[i].duplex;
	dev->face = PrinterDeviceList[i].face;
#endif
	return(OK);
}


/** Ȥ GraphicsState ǥץ󥿤 () ꤹ롣
 * RestoreGS ξʤɤ˸ƤФ롣
 * @param printerContext оݤΥץ󥿥ƥ
 * @retval  OK
 * @retval  ERROR
 */
static int OutputCurrentGS(int printerContext)
{
	GraphicsState	*gs;
	int				result;

	// Get pointer to current GraphicsState.
	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return(ERROR);
	}

	/*
	 * (TBD)ץ󥿤ѹʤФʤʤ
	 * ¤ΤΥФѤäƤפʥƥʥ󥹥ե꡼Υɤˤ
	 * ƤӽФν֤ѹƤɤ???
	 */
	result = SetFillMode(printerContext, gs->fillMode);
	if (result < 0) {
		return(ERROR);
	}
	result = SetLineWidth(printerContext, gs->lineWidth);
	if (result < 0) {
		return(ERROR);
	}
	result = SetLineCap(printerContext, gs->lineCap);
	if (result < 0) {
		return(ERROR);
	}
	result = SetLineJoin(printerContext, gs->lineJoin);
	if (result < 0) {
		return(ERROR);
	}
	result = SetStrokeColor(printerContext, &gs->strokeBrush);
	if (result < 0) {
		return(ERROR);
	}
	result = SetFillColor(printerContext, &gs->fillBrush);
	if (result < 0) {
		return(ERROR);
	}
	result = SetROP(printerContext, gs->rop);
	if (result < 0) {
		return(ERROR);
	}
	result = SetLineDash(printerContext, gs->lineDash, gs->dashNum);
	if (result < 0) {
		return(ERROR);
	}
	result = SetLineDashOffset(printerContext, gs->dashOffset);
	if (result < 0) {
		return(ERROR);
	}
	result = SetLineStyle(printerContext, gs->lineStyle);
	if (result < 0) {
		return(ERROR);
	}
	result = SetMiterLimit(printerContext, gs->miterLimit);
	if (result < 0) {
		return(ERROR);
	}
	result = SetPaintMode(printerContext, gs->paintMode);
	if (result < 0) {
		return(ERROR);
	}
	{
		Brush brush;

		brush.colorSpace = gs->colorSpace;
		brush.color[0] = gs->bgColor[0];
		brush.color[1] = gs->bgColor[1];
		brush.color[2] = gs->bgColor[2];
		brush.color[3] = gs->bgColor[3];
		brush.xorg = 0;
		brush.yorg = 0;
		brush.pbrush = NULL;

		result = SetBgColor(printerContext, &brush);
		if (result < 0) {
			return(ERROR);
		}
	}

	return(OK);
}

/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/*
 * ʹߤ VDAPI
 *	ƤФ
 */
/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

/* ----------------------------------------------------------------
 *
 * Creating and Managing Print Contexts
 *
 * ---------------------------------------------------------------- */

/** PDAPI˽
 * @note  apiEntry Ϥɤ */
/* int OpenPrinter(int outputFD, char *printerModel, int *nApiEntry, int *apiEntry[]) */
/* int OpenPrinter(int outputFD, char *printerModel, int *nApiEntry, int (*apiEntry[])()) */
int OpenPrinter(int outputFD, char *printerModel, int *nApiEntry, void *apiEntry[])
{
	int			printerContext;
	Printer		*printer;
	int			result;

	DebugPrint("pdapi: %s,%d outputFD=%d, printerModel=%s, nApiEntry=%p, apiEntry=%p\n",
			   __FUNCTION__,
			   __LINE__,
			   outputFD,
			   printerModel,
			   nApiEntry,
			   apiEntry);

	/* device ΰγ */
	printerContext = NewPrinterContext();
	if (printerContext < 0) {
		/* 顼 (餯­) */
		return(ERROR);
	}

	/*
	 * ǥ顼å򤷤ʤФʤʤ
	 * ΰǤϽ褽ˤʤ
	 *
	 * GS-LPDǤ
	 *		code = escv_range_check(dev);
	 * ѻ極䡢٤ΥåԤʤäƤ롣
	 */


	/*
	 * ͤ
	 */

	/* ؿȥؤϿ */
	*apiEntry = VectorProcs;
	*nApiEntry = sizeof(VectorProcs) / sizeof(int);

	/* ץ̾ꤹ */
	result = SetPrinterName(printerContext, printerModel);
	if (result < 0) {
		/* 顼 */
		return(ERROR);
	}

	/*  */
	printer = GetPrinter(printerContext); /* #### Ǥϥ顼ȯʤΤǥåʤ */
	printer->outputFD = outputFD;

	return(printerContext);
}


/** PDAPI˽ */
int ClosePrinter(int printerContext)
{
	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);
	return (DeletePrinterContext(printerContext));
}




/* ----------------------------------------------------------------
 *
 * Job Control Operations
 *
 * ---------------------------------------------------------------- */

/** ֤򳫻Ϥ
 * ѥ᡼jobInfoϤ줿°ᤷ(TBD)
 * ꤵ줿°˹碌EJLޥɤ롣
 * Jobñ̤٤Esc/Pageޥɤ⤳롣
 * Esc/PageޥɤΤĤϡѥ᡼jobInfo˰¸롣
 * (TBD) jobInfoν
 *   @<resolution@>?;@<pagesize@>?
 * resolutionȤơʸ'600x600', '300x300'ǽ
 * pagesizeȤơʸ'A4', 'B5'ǽ
 * resolution, pagesizeȤ˾άǽǤ롣
 *
 * @param printerContext OpenPrinter()ץ󥿡ƥȤꤹ
 * @param *jobInfo ֤°ꤹ롣NULLξϤ٤ƤΥ°ϥɥ饤ФΥǥեͤѤ롣
 * @retval OK ｪλ
 * @retval ERROR 顼ȯǤ
 */
int StartJob(int printerContext, char *jobInfo)
{
	DeviceInfo	*dev;
	char	obuf[128];
	char	*p;
	int resolution;
	int pageSize;

	// for test
	//jobInfo = "600x600;";
	//jobInfo = ";B5";
	//jobInfo = "600x600;";
	//jobInfo = "600x600;";
	//jobInfo = "600x600;";
	//jobInfo = NULL;

	DebugPrint("pdapi: %s,%d (printerContext=%d, jobInfo=[%s])\n",
			   __FUNCTION__,
			   __LINE__,
			   printerContext,
			   jobInfo);

	resolution = 300; // default
	pageSize = 0; // default(EpPaperTable0)

	if (jobInfo != NULL) {
		int i;
		char* p;
        char* q;
		char tmp[16]; // TBD

        // resolutionʸμ
        p = &jobInfo[0];
        q = &tmp[0];

        while ((*p != '\0') && (*p != ';')) {
			*q++ = *p++;
        }
        *q = '\0';

		//
		if (strcmp(tmp, "600x600") == 0) {
			resolution = 600;
		}
		// else resolution = default;


        if (p != NULL) {
			p++;
        }

        q = &tmp[0];
        while ((*p != '\0') && (*p != ';')) {
			*q++ = *p++;
        }
        *q = '\0';

		//
		i = 0;
		for (i = 0; EpPaperTable[i].name != NULL; i++) {
			DebugPrint("serching paper size - %d [%s] [%s]\n",
					   i,
					   EpPaperTable[i].name,
					   tmp);
			if (strcmp(EpPaperTable[i].name, tmp) == 0) {
				pageSize = i;
			}
		}
	}

	DebugPrint("pdapi: %s,%d resolution=%d, pagesize=%d[%s]\n",
			   __FUNCTION__,
			   __LINE__,
			   resolution,
			   pageSize,
			   EpPaperTable[pageSize].name);

	dev = GetDeviceInfo(printerContext);
	if (dev == NULL) {
		// errorno = BADCONTEXT()
		return(ERROR);
	}

	// ٤StartPage()ǤѤ뤿ᡢ
	// DeviceInfoꤷƤ
	dev->resolution = resolution;

	/**
	 * EJLޥɤϰǡ(Job)Ƭˤ֤Ǥʤ
	 * EJLޥɤΥѥ᡼jobInfo˰¸롣
	 * jobInfoƤ˹碌EJLѹ롣
	 */
	/**
	 * ޥΤWriteǥ顼ȯ
	 * ³ԲǽǤΤǡʹߤνϹԤʤʤ
	 */

	/* 귿 */
	p = "\033\001@EJL \012";
	if (ERROR == Write(printerContext, p, strlen(p))) {
		return(ERROR);
	}

	/* STARTJOBޥ */
	sprintf(obuf, "@EJL STARTJOB ID=\"OPVP-%s\"\012", dev->printerName);
	if (ERROR == Write(printerContext, obuf, strlen(obuf))) {
		return(ERROR);
	}

	/* SELECTޥ */
	p = "@EJL SELECT LANGUAGE=ESC/PAGE\012";
	if (ERROR == Write(printerContext, p, strlen(p))) {
		return(ERROR);
	}

	/* SETޥɳ */
	p = "@EJL SET";
	if (ERROR == Write(printerContext, p, strlen(p))) {
		return(ERROR);
	}

	/*  */
	//sprintf(obuf, " RESOLUTION=%d", dev->maxRes);
	sprintf(obuf, " RESOLUTION=%d", resolution);
	if (ERROR == Write(printerContext, obuf, strlen(obuf))) {
		return(ERROR);
	}

	/* ȥ쥤 */
	sprintf(obuf, " OUTPUTUNIT=%s", dev->face);
	if (ERROR == Write(printerContext, obuf, strlen(obuf))) {
		return(ERROR);
	}

	/* */
	p = " PAPERUNIT=AUTO";
	if (ERROR == Write(printerContext, p, strlen(p))) {
		return(ERROR);
	}

	/* ξ̰On/Off */
	sprintf(obuf, " DUPLEX=%s", dev->duplex);
	if (ERROR == Write(printerContext, obuf, strlen(obuf))) {
		return(ERROR);
	}

	/* */
	p = " QUANTITY=1";
	if (ERROR == Write(printerContext, p, strlen(p))) {
		return(ERROR);
	}

	/*  */
	p = " COPIES=1";
	if (ERROR == Write(printerContext, p, strlen(p))) {
		return(ERROR);
	}

	/* RIT(RET) */
	p = " RITECH=ON";
	if (ERROR == Write(printerContext, p, strlen(p))) {
		return(ERROR);
	}

	/* ѻ楿 */
	p = " PAPERTHICKNESS=NORMAL";
	if (ERROR == Write(printerContext, p, strlen(p))) {
		return(ERROR);
	}

	/* ѻ極 */
	//p = " PAGESIZE=A4";
	sprintf(obuf, " PAGESIZE=%s", EpPaperTable[pageSize].name);
	if (ERROR == Write(printerContext, obuf, strlen(obuf))) {
		return(ERROR);
	}

	/*  */
	p = " ZOOM=OFF";
	if (ERROR == Write(printerContext, p, strlen(p))) {
		return(ERROR);
	}

	/* 顼 */
	p = " ERRORCODE=ON";
	if (ERROR == Write(printerContext, p, strlen(p))) {
		return(ERROR);
	}

	/* */
	p = " SIZEIGNORE=OFF";
	if (ERROR == Write(printerContext, p, strlen(p))) {
		return(ERROR);
	}

	/*  */
	p = " SKIPBLANKPAGE=YES";
	if (ERROR == Write(printerContext, p, strlen(p))) {
		return(ERROR);
	}

	/* 奪եå */
	p = " TOPOFFSET=0";
	if (ERROR == Write(printerContext, p, strlen(p))) {
		return(ERROR);
	}

	/* եå */
	p = " LEFTOFFSET=0" "\012"; // SETޥɽλ
	if (ERROR == Write(printerContext, p, strlen(p))) {
		return(ERROR);
	}

	/* 쥹 */
	p = "@EJL ENTER LANGUAGE=ESC/PAGE-COLOR\012";
	if (ERROR == Write(printerContext, p, strlen(p))) {
		return(ERROR);
	}

	/* Jobñ̤٤ESC/PAGEޥ */
	{
		int i;
		char *cmd1[] = {
			ESC_GS "rhE",			//ϡɥꥻå
			ESC_GS "0pddO",			//ǥХ
			ESC_GS "0;0cmmE",		///
			ESC_GS "1;2;3ccmE",		//ˡ
			ESC_GS "2;2;5ccmE", 	//ˡ(2;2;3 -> 2;2;5ѹ)
			ESC_GS "3;2;4ccmE", 	//ˡ
			ESC_GS "1;1raE",		//꡼˥
			ESC_GS "2;1raE",		//꡼˥
			ESC_GS "3;1raE"			//꡼˥
		};

		/* Ĵ(ǡ 7byte) */
		char cmd2[] = ESC_GS "7;0;2;0cam{E\012\000\000\000\000\000\000";

		/**
		 * ٤ξϡǾñꡢ˱ƶ롣
		 * Ǿñ(muE)ϡ300dpiξ0.24, 600dpiξ0.12, 1200dpiξ0.06ꤹ롣
		 * (loE)ϡ300dpiξ(60, 60)600dpiξ(120, 120)ꤹ롣
		 */
		char *cmd3_600[] = {
			//ESC_GS "120;120loE",	//(ٰ¸) StartPage()ǽФ
			ESC_GS "0;0.12muE",		//Ǿñ(ٰ¸)
			ESC_GS "3;600;600drE",	//ɥåȥѥ(ٰ¸)
			ESC_GS "2;600;600drE",	//ɥåȥѥ(ٰ¸)
			ESC_GS "1;600;600drE",	//ɥåȥѥ(ٰ¸)
			ESC_GS "0;600;600drE",	//ɥåȥѥ(ٰ¸)
		};

		char *cmd3_300[] = {
			//ESC_GS "60;60loE",		//(ٰ¸) StartPage()ǽФ
			ESC_GS "0;0.24muE",		//Ǿñ(ٰ¸)
			ESC_GS "3;300;300drE",	//ɥåȥѥ(ٰ¸)
			ESC_GS "2;300;300drE",	//ɥåȥѥ(ٰ¸)
			ESC_GS "1;300;300drE",	//ɥåȥѥ(ٰ¸)
			ESC_GS "0;300;300drE",	//ɥåȥѥ(ٰ¸)
		};

		char *cmd4[] = {
			//ESC_GS "0;0cmmE",		//(cmd1ѤʤΤǺ)
			//ESC_GS "1tsE",		//(ΥѥޥɤʤΤǺ)
			ESC_GS "0alfP",			//ư(ʸ)
			ESC_GS "0affP",			//ưڡ(ʸ)
			ESC_GS "0;0;0clfP",		//CR/LF/FFư(ʸ)
			ESC_GS "0pmP",			//ְ߰ư⡼(ʸ)
			//ESC_GS "1024ibI",		//ӥåȥ᡼Ͽ(ŤޥɤʤΤǺ)
			ESC_GS "2cmE",			//åץ⡼ɻ/ ѹ('2cmE' -> '1cmE')
			//ESC_GS "0bcI",		//ӥåȥ᡼ǡ̷(äפʤΤǺ)
			//ESC_GS "1;10mlG"		//޷ޥߥå
		};

		/* ѥåϿ */
		char cmd5[] = ESC_GS "8;1;2;2;2plr{E\377\377\377\377\000\000\000\000";

		char *cmd6[] = {
			ESC_GS "0sarG",		//޷ɸ⡼
			ESC_GS "2;204wfE"	//黻⡼
		};

		/* Ǥ EJL ޥɤΤߤФ */
		for (i = 0; i < (sizeof(cmd1) / sizeof(char *)); i++) {
			if (ERROR == Write(printerContext, cmd1[i], strlen(cmd1[i]))) {
				return(ERROR);
			}
		}

		if (ERROR == Write(printerContext, cmd2, sizeof(cmd2) - 1)) {
			return(ERROR);
		}

		/* resolution¸륳ޥɷ */
		if (resolution == 600) {
			for (i = 0; i < (sizeof(cmd3_600) / sizeof(char *)); i++) {
				if (ERROR == Write(printerContext, cmd3_600[i], strlen(cmd3_600[i]))) {
					return(ERROR);
				}
			}
		} else {
			for (i = 0; i < (sizeof(cmd3_300) / sizeof(char *)); i++) {
				if (ERROR == Write(printerContext, cmd3_300[i], strlen(cmd3_300[i]))) {
					return(ERROR);
				}
			}
		}

		for (i = 0; i < (sizeof(cmd4) / sizeof(char *)); i++) {
			if (ERROR == Write(printerContext, cmd4[i], strlen(cmd4[i]))) {
				return(ERROR);
			}
		}

		if (ERROR == Write(printerContext, cmd5, sizeof(cmd5))) {
			return(ERROR);
		}

		for (i = 0; i < (sizeof(cmd6) / sizeof(char *)); i++) {
			if (ERROR == Write(printerContext, cmd6[i], strlen(cmd6[i]))) {
				return(ERROR);
			}
		}
	}

	DebugPrint("pdapi: %s,%d\n", __FUNCTION__, __LINE__);
	/* StartJob()λ */
	{
		Printer* pPrinter;
		pPrinter = GetPrinter(printerContext);
		if (pPrinter == NULL) {
			return(ERROR);
		}
		pPrinter->jobStarted = TRUE;
		/* ʸ塢escpageʿĶ */
	}

	DebugPrint("pdapi: %s,%d\n", __FUNCTION__, __LINE__);
	/* ĤĤʤλ */
	return(OK);
}


/** ֤λ롣
 * escpageæФEJLνλϤ
 * @retval OK ｪλ
 * @retval ERROR 顼ȯ
 * @param printerContext OpenPrinter()˼ץ󥿥ƥ
 */
int EndJob(int printerContext)
{
	char *	page_cmd = ESC_GS "rhE"; /* ϡɥꥻå */
	char *	ejl_cmd = "\033\001@EJL \012@EJL ENDJOB \012\033\001@EJL \012";

	DebugPrint("pdapi: %s,%d(printerContext=%d)\n",
			   __FUNCTION__,
			   __LINE__,
			   printerContext);

	/* StartJob()ޤpageޥɤ϶ػߤ */
	{
		Printer *pPrinter;

		pPrinter = GetPrinter(printerContext);
		if (pPrinter == NULL) {
			return(ERROR);
		}
		pPrinter->jobStarted = FALSE;
		/* ʸ塢escpa	char	obuf[128];geŹ */
	}
	if (ERROR == Write(printerContext, page_cmd, strlen(page_cmd))) {
		return(ERROR);
	}
	if (ERROR == Write(printerContext, ejl_cmd, strlen(ejl_cmd))) {
		return(ERROR);
	}
	return(OK);
}



/** PDAPI˽(̵) */
int StartDoc(int printerContext, char *docInfo)
{
	int		result;
	result = 0;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	return(result);
}



/** PDAPI˽(̵) */
int EndDoc(int printerContext)
{
	int		result;
	result = 0;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	return(result);
}


/** PDAPI˽ */
int StartPage(int printerContext, char *pageInfo)
{
	DeviceInfo	*dev;
	char *p;

	dev = GetDeviceInfo(printerContext);
	if (dev == NULL) {
		return(ERROR);
	}

	/* ڡñ̤Υޥ */
	/**
	 * @note ꥳޥɤˤĤ
	 * ܥޥɤλϡ̤ɤΤ褦ʥǡ褦ˤ
	 * ˰¸̤ΰȤǡäƤ
	 * ǤСԤɬפ뤬餯
	 * ѻȤǡäƤ뤿ᡢǤ(0, 0)
	 * ꤹɤϤǤ롣
	*/
#if 0
	if (dev->resolution == 600) {
		p = ESC_GS "120;120loE";
		if (ERROR == Write(printerContext, p, strlen(p))) {
			return(ERROR);
		}
	} else {
		/* default resolution == 300 */
		p = ESC_GS "60;60loE";
		if (ERROR == Write(printerContext, p, strlen(p))) {
			return(ERROR);
		}
	}
#else
	p = ESC_GS "0;0loE";
#endif

	/* Graphics State νԤʤ */
	if (ERROR == InitGS(printerContext)) {
		return(ERROR);
	}

	return(OK);
}


/** ڡνλꤹ롣
 * escpageɥ饤ФǤϡΥߥ󥰤ǤESC_FF(0x14)ϤΤߤǤ롣
 * @param printerContext OpenPrinter()˼ץ󥿥ƥ
 * @retval OK ｪλ
 * @retval ERROR 顼ȯ
 */
int EndPage(int printerContext)
{
	/**
	 * @note PageEndꤲ٤ޥɤFormFeedǤ褤
	 * ڡȤѤΥѥ᡼Υǥեͤ
	 * ꤷϡStartPage˵Ҥ롣
	 */
	//char	*cmd = ESC_GS "3;0;0lwG" ESC_GS "1;10mlG" ESC_FF;
	char	*cmd = ESC_FF;

	DebugPrint("pdapi: %s,%d(printerContext=%d)\n",
			   __FUNCTION__,
			   __LINE__,
			   printerContext);

	return (Write(printerContext, cmd, strlen(cmd)));

}

/** PDAPI˽(̤ݡ) */
int QueryDeviceCapability(int printerContext, int queryflag, int buflen, char *infoBuf)
{
	int		result;
	result = 0;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	return(result);
}


/** PDAPI˽(̤ݡ) */
int QueryDeviceInfo(int printerContext, int queryflag, int buflen, char *infoBuf)
{
	int		result;
	result = 0;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	return(result);
}


/* ----------------------------------------------------------------
 *
 * Graphics State Object Operations
 *
 * ---------------------------------------------------------------- */

/** PDAPI˽ */
int ResetCTM(int printerContext)
{
	GraphicsState *gs;
	CTM ctm;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		/* GraphicsState̵⤷Printer Context
		   Ǥ */
		errorno = BADCONTEXT;
		return(ERROR);
	}

	/* ǥե */
	ctm.a = 1;
	ctm.b = 0;
	ctm.c = 0;
	ctm.d = 1;
	ctm.e = 0;
	ctm.f = 0;

	return SetCTM(printerContext, &ctm);
}

/** PDAPI˽ */
int SetCTM(int printerContext, CTM *pCTM)
{
	GraphicsState *gs;
	CTM *pCurCTM;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	if (pCTM == NULL) {
		errorno = PARAMERROR;
		return(ERROR);
	}

	/* Get pointer to current CTM. */
	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return(ERROR);
	}

	pCurCTM = &(gs->ctm);
#if 1
	{
		/* Copy CTM. */
		pCurCTM->a = pCTM->a;
		pCurCTM->b = pCTM->b;
		pCurCTM->c = pCTM->c;
		pCurCTM->d = pCTM->d;
		pCurCTM->e = pCTM->e;
		pCurCTM->f = pCTM->f;
	}
#else
	memcpy(pCurCTM, pCTM, sizeof(CTM));
#endif

	return(OK);
}


/** PDAPI˽ */
int GetCTM(int printerContext, CTM *pCTM)
{
	GraphicsState *gs;
	CTM *pCurCTM;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	if (pCTM == NULL) {
		errorno = PARAMERROR;
		return(ERROR);
	}

	/* Get pointer to current CTM. */
	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return(ERROR);
	}
	pCurCTM = &(gs->ctm);

#if 1
	{
		pCTM->a = pCurCTM->a;
		pCTM->b = pCurCTM->b;
		pCTM->c = pCurCTM->c;
		pCTM->d = pCurCTM->d;
		pCTM->e = pCurCTM->e;
		pCTM->f = pCurCTM->f;
	}
#else
	memcpy(pCTM, pCurCTM, sizeof(CTM));
#endif

	return(OK);
}

/** PDAPI˽
 * GraphicsStateǰƤ°ͤǥեȤͤꤹ롣ϡ
 * PDAPIAPIѤƹԤ
 */
int InitGS(int printerContext)
{

	DebugPrint("pdapi: %s,%d (printerContext=%d)\n", __FUNCTION__, __LINE__, printerContext);

	/* CTM */
	// (TBD) ǥե
	if (ResetCTM(printerContext) != OK) {
		return(ERROR);
	}

	/* ColorSpace */
	// (TBD) ǥե
	if (SetColorSpace(printerContext, cspaceStandardRGB) != OK) {
		return(ERROR);
	}

	/* FillMode */
	// (TBD) ǥե
	if (SetFillMode(printerContext, fillModeEvenOdd) != OK) {
		return(ERROR);
	}


	/* LineWidth */
	// (TBD) ǥե(1.0?)
	{
		Fix defaultLineWidth;
		i2Fix(1, defaultLineWidth);
		if (SetLineWidth(printerContext, defaultLineWidth) != OK) {
			return(ERROR);
		}
	}

	/* LineCap */
	// (TBD) ǥե
	if (SetLineCap(printerContext, lineCapButt) != OK) {
		return(ERROR);
	}

	/* LineJoin */
	// (TBD) ǥե
	if (SetLineJoin(printerContext, lineJoinMiter) != OK) {
		return(ERROR);
	}

	/* Brush (Stroke/Fill Common) */
	{
		// Create Default Brush.
		Brush brush;
		int defaultColor[4] = { 0, 0, 0, 0 };

		brush.colorSpace = cspaceStandardRGB;
#if 0
		memcpy(brush.color, defaultColor, 4); /* default color */
#else
		brush.color[0] = defaultColor[0];
		brush.color[1] = defaultColor[1];
		brush.color[2] = defaultColor[2];
		brush.color[3] = defaultColor[3];
#endif
		brush.pbrush = NULL;
		brush.xorg = 0;
		brush.yorg = 0;

		// Set Default Brush.
		if (SetStrokeColor(printerContext, &brush) != OK) {
			return(ERROR);
		}
		if (SetFillColor(printerContext, &brush) != OK) {
			return(ERROR);
		}
	}

	/* Line Style */
	// (TBD)ǥե͡Solidɤ
#if 1
	if (SetLineStyle(printerContext, lineStyleSolid) != OK) {
		return(ERROR);
	}
#else
	// ƥ
	if (SetLineStyle(printerContext, lineStyleDash) != OK) {
		return(ERROR);
	}
#endif

	/* Line Dash */
	// (TBD)ǥե
	// pdashNULLOK?(ͽˤä˻ʤ)
#if 1
	if (SetLineDash(printerContext, NULL, 0) != OK) {
		return(ERROR);
	}
#else
	// ǥХåDash
	{
		Fix dash[255];
		int dashNum;

		dashNum = 0;
		i2Fix(10, dash[dashNum]); dashNum++;
		i2Fix(2, dash[dashNum]); dashNum++;
		i2Fix(20, dash[dashNum]); dashNum++;
		i2Fix(2, dash[dashNum]); dashNum++;

		if (SetLineDash(printerContext, dash, dashNum) != OK) {
			return ERROR;
		}
	}
#endif

#if 0
	// ƥ
	if (SetLineStyle(printerContext, lineStyleDash) != OK) {
		return(ERROR);
	}
#endif

	/* Line Dash Offset */
	{
		Fix defaultLineDashOffset;
		i2Fix(0, defaultLineDashOffset);
		if (SetLineDashOffset(printerContext, defaultLineDashOffset) != OK) {
			return(ERROR);
		}
	}

	/* Miter Limit */
	{
		Fix miterLimit;
		i2Fix(10, miterLimit);
		if (SetMiterLimit(printerContext, miterLimit) != OK) {
			return(ERROR);
		}
	}

	/* rop */
	{
		// ǥեͤ204(escpageͤ)
		if (SetROP(printerContext, 204) != OK) {
			return (ERROR);
		}
	}

	DebugPrint("pdapi: %s,%d end of InitGS()\n", __FUNCTION__, __LINE__);
	return(OK);
}


/** PDAPI˽ */
int SaveGS(int printerContext)
{
	int		result;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	result = PushGraphicsState(printerContext);

	return(result);
}


/** PDAPI˽ */
int RestoreGS(int printerContext)
{
	int		result;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	result = PopGraphicsState(printerContext);
	if (result != OK) {
		return(result);
	}

	result = OutputCurrentGS(printerContext);

	return(result);
}


/** PDAPI˽
 * ݡȤ륫顼ڡcspaceStandardRGBΤ
 */
int QueryColorSpace(int printerContext, ColorSpace *pcspace, int *pnum)
{
	int i;
	int	num = 1;
	ColorSpace support[] = { cspaceStandardRGB };


	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	/* 顼ڡο֤ */
	if (*pnum < num) {
		*pnum = num;

		errorno = PARAMERROR;
		return (ERROR);
	}

	/* 顼ڡ֤ */
	*pnum = num;
	for (i = 0; i < num; i++) {
		pcspace[i] = support[i];
	}
	return (OK);
}


/** PDAPI˽ */
int SetColorSpace(int printerContext, ColorSpace cspace)
{
	GraphicsState *gs;

	DebugPrint("pdapi: %s,%d\n ", __FUNCTION__, __LINE__);

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return(ERROR);
	}
	gs->colorSpace = cspace;

	return(OK);
}


/** PDAPI˽ */
int GetColorSpace(int printerContext, ColorSpace *pcspace)
{
	GraphicsState *gs;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return(ERROR);
	}
	*pcspace = gs->colorSpace;

	return(OK);
}


/**ɥ饤ФǰȤǤROPΰGraphicsState롣
 * ٥ɥ饤ФϡROP3Τ٤Ƥ򥵥ݡȤ롣
 * /ѹǤΤϥ᡼ropΤߤǡ
 * ٥ƥrop/Ǥʤ
 * ٥ƥȤǤϾropP(240)Ѥ롣
 * GraphicsStateؤΥϹԤʤʤ
 *
 * @param printerContext OpenPrinter()ǼprinterContext
 * @param pnum ݡȤROPɤοɥ饤ФǤϾ256֤롣
 * @param prop ݡȤROPɤǼ
 * @retval OK ˽Ԥʤ줿
 * @retval -1 顼ȯ(pnum­Ƥ)
 */
int QueryROP(int printerContext, int *pnum, int *prop)
{
	DebugPrint("pdapi: %s,%d (printerContext=%d, pnum=%d, prop=%p)\n",
			   __FUNCTION__,
			   __LINE__,
			   printerContext,
			   *pnum,
			   prop);

	// ROPɤ䤤碌
	if (prop == NULL) {
		*pnum = 16;
		return (OK);
	}

	// ѲǽROPɤμ
	if (*pnum < 16) {
		errorno = PARAMERROR;
		return (ERROR);
	}

	// ROP2ϰϤΤߥݡȤ
	*pnum = 16;
	{
		int i;
		int rop2[] = {
			0x00, 0x05, 0x0A, 0x0F, 0x50, 0x55, 0x5A, 0x5F,
			0xA0, 0xA5, 0xAA, 0xAF, 0xF0, 0xF5, 0xFA, 0xFF };

		for (i = 0; i < 16; i++) {
			prop[i] = rop2[i];
		}
	}
	return (OK);
}

/** RasterOperationꤹ롣
 * ٥ɥ饤ФϤǻꤵ줿Raster Operation˽äơ
 * ᡼襳ޥɤ롣٥Ѥrop
 * δؿǤϻǤʤɥ饤ФǤϡ٥Ͼrop
 * P(240)Ѥ롣
 * GraphicsState˳Ǽropɤνͤ204(S)Ǥ롣
 *
 * @param printerContext OpenPrinter()줿printerContext
 * @param rop ROP3
 * @retval OK ٤Ƥν˽λ
 * @retval -1 ˥顼ȯ(printerContextꤵ줿)
 */
int SetROP(int printerContext, int rop)
{
	GraphicsState *gs;

	DebugPrint("pdapi: %s,%d (printerContext=%d, rop=%d)\n",
			   __FUNCTION__,
			   __LINE__,
			   printerContext,
			   rop);

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return(ERROR);
	}
	gs->rop = rop;

    // ropѹե饰ΩƤ
	gs->ropChanged = TRUE;

	return(OK);
}

/** GraphicsStateꤵƤRasterOperation롣
 * prop˸ꤵƤROP3ɤǼ롣
 * @param printerContext OpenPriter()ǼPrinterContext
 * @param prop ߤropͤǼΰؤΥݥ
 * @retval OK ˽Ԥʤ줿
 * @retval -1 ˥顼ȯ
 */
int GetROP(int printerContext, int *prop)
{
	GraphicsState *gs;

	DebugPrint("pdapi: %s,%d (printerContext=%d, prop=%p)\n",
		__FUNCTION__,
		__LINE__,
		printerContext,
		prop);

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return(ERROR);
	}
	*prop = gs->rop;

	return(OK);
}

typedef enum ROP_OPERATION_ {
	OPERATION_TEXT,
	OPERATION_VECTOR,
	OPERATION_IMAGE
} ROP_OPERATION;


/** ROPޥɤ롣
 * GraphicsState˳ǼƤropɤѤơ
 * ROPꥳޥ(wfE)롣
 * GraphicsStateropѹƤʤ
 * ޥԤʤʤ(OPARATION_IMAGEΤ)
 * (TBD)ȥå饹᡼(operation=3)ϥݡȤʤ
 *
 * @param printerContext OpenPrinter()ǼPrinterContextꤹ
 * @param operation ropѹ륪ڥ졼OPERATION_TEXT,
 *                  OPERATION_VECTOR, OPERATION_IMAGEΤ줫ꤹ롣
 * @param rop ropɡ
 *            operationOPERATION_TEXTλ240OPERATION_VECTORλ
 *            ROP2ߴROP3ɤǤʤƤϤʤʤ
 *            OPERATION_IMAGEξϡGraphicsState˳Ǽ줿ropɤ
 *            Ѥ뤿ᡢͤϻȤʤ
 * @retval OK ˥ޥ줿
 * @retval -1 顼ȯ
 */
#if 0
int WriteROP(int printerContext, ROP_OPERATION operation, int rop)
{
	int ropCode;
	int op;
	GraphicsState *gs;

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return (ERROR);
	}

	switch (operation) {
		case OPERATION_IMAGE:
			// ropѹƤʤ
			if (!gs->ropChanged) {
				// Ĥreturn
				return (OK);
			}
			if (GetROP(printerContext, &ropCode) != OK) {
				return (ERROR);
			}
			// Pageޥɰ¸
			op = 5;
			break;
		case OPERATION_TEXT:
			ropCode = rop;
			// ƥȤrop240Ǥʤ
			if (ropCode != 240) {
				errorno = FATALERROR;
				return (ERROR);
			}
			// Pageޥɰ¸
			op = 1;
			break;
		case OPERATION_VECTOR:
			ropCode = rop;
			// ٥OperationǤROP2ߴΥɤǤʤ
			// (TBD)ROP2ߴΥɤǤ뤫å
			// Pageޥɰ¸
			op = 2;
			break;
		default:
			errorno = FATALERROR;
			return (ERROR);
	}

	// ޥ
	{
		char obuf[128];

		memset(obuf, 0, 128);
		(void)sprintf(obuf, ESC_GS "%d;%d;wfE", op, ropCode);
		if (CheckWrite(printerContext, obuf, strlen(obuf)) != OK) {
			return (ERROR);
		}
	}

	// ǤСѹե饰ݤƤ
	if (operation == OPERATION_IMAGE) {
		gs->ropChanged = FALSE;
	}

	return (OK);
}
#else
int WriteROP(int printerContext)
{
	int ropCode;
	GraphicsState *gs;

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return (ERROR);
	}

	// ropѹƤʤ
	if (gs->ropChanged == FALSE) {
		// Ĥreturn
		return (OK);
	}
	if (GetROP(printerContext, &ropCode) != OK) {
		return (ERROR);
	}

	// ޥ
	{
		char obuf[128];

		memset(obuf, 0, 128);
		(void)sprintf(obuf,
			      ESC_GS "2;%d;wfE"
			      ESC_GS "3;%d;wfE"
			      ESC_GS "5;%d;wfE"
			      , ropCode, ropCode, ropCode);
		if (CheckWrite(printerContext, obuf, strlen(obuf)) != OK) {
			return (ERROR);
		}
	}

	// ǤСѹե饰ݤƤ
	gs->ropChanged = FALSE;

	return (OK);
}


#endif


/** PDAPI˽ */
int SetFillMode(int printerContext, FillMode fillmode)
{
	GraphicsState *gs;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return(ERROR);
	}
	gs->fillMode = fillmode;

	return(OK);
}


/** PDAPI˽ */
int GetFillMode(int printerContext, FillMode* pfillmode)
{
	GraphicsState *gs;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return(ERROR);
	}
	*pfillmode = gs->fillMode;

	return(OK);
}


/** PDAPI˽(̤ݡ) */
int SetAlphaConstant(int printerContext, float alpha)
{
	int		result;
	result = 0;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	return(result);
}


/** PDAPI˽(̤ݡ) */
int GetAlphaConstant(int printerContext, float *palpha)
{
	int		result;
	result = 0;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	return(result);
}


/** PDAPI˽ */
int SetLineWidth(int printerContext, Fix width)
{
	float			f;
	GraphicsState	*gs;

	Fix2f(width, f);
	DebugPrint("pdapi: %s,%d printerContext=%d, width=(%d/%f)\n", __FUNCTION__, __LINE__, printerContext, Fix2i(width), f);

	// GraphicsStateμ
	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return(ERROR);
	}

	// GraphicStateι
	gs->lineWidth = width;

	// Line°ѹޥɤ
	return WriteLineAttributes(printerContext);
}


/** PDAPI˽ */
int GetLineWidth(int printerContext, Fix *pwidth)
{
	GraphicsState *gs;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	// GraphicsStateμ
	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return(ERROR);
	}

	*pwidth = gs->lineWidth;
	return(OK);
}


/** PDAPI˽ */
int SetLineDash(int printerContext, Fix pdash[], int num)
{
	GraphicsState	*gs;
	int				dashNum;

	DebugPrint("pdapi: %s,%d pdash=%p, num=%d\n", __FUNCTION__, __LINE__, pdash, num);

	// num 0üͤǤ
	dashNum = num;
	if ((num % 2) != 0) {
		dashNum *= 2;
	}
	if ((dashNum < 0) || (dashNum >= 252)) {
		errorno = NOTSUPPORTED;
		return(ERROR);
	}

	// GraphicsStateμ
	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return(ERROR);
	}

	// dashLine0ꥢ
	{
		//DebugPrint("sizeof fix = %d.\n", sizeof(Fix));
//		memset(&gs->lineDash[0], 0, sizeof(Fix) * 255);
		memset(&gs->lineDash[0], 0, sizeof(gs->lineDash));
	}

	// GraphicsStateι
	{
		gs->dashNum = dashNum;
		if (gs->dashNum != 0) {
			int i;
			for (i = 0; i < num; i++) {
				// (TBD) Ĺϰϥå?
				// PageλͤǤϡֿ޷θ³͡פȤ롣
				gs->lineDash[i] = pdash[i];
				if ((num % 2) != 0) {
					gs->lineDash[i + num] = pdash[i];
				}
			}
		}
	}

#if 0
	// (TEST) dump gs->lineDash.
	{
		int i;

		DebugPrint("----------num=%d\n", gs->dashNum);
		for (i = 0; i < gs->dashNum; i++) {
			DebugPrint("dash %d = %d\n", i, Fix2i(gs->lineDash[i]));
		}
		DebugPrint("----------\n");
	}

#endif
	{
		LineStyle style;
		if (GetLineStyle(printerContext, &style) != OK) {
			return(ERROR);
		}

		if (style == lineStyleDash) {
			// StyleDashξ(=SolidǤʤ)
			// Dashѥѹ¨¤ȿǤʤƤϤʤʤ
			// ޥɬ
			// num=0ξǤ⡢Dash->SolidѹꤨΤɬ
			return WriteLineKind(printerContext);
		}
		// StyleSolidξ硢GraphicsStateιΤ߹Ԥʤ
		// ޥϹԤʤʤ
	}

	return(OK);
}


/** PDAPI˽ */
int GetLineDash(int printerContext, Fix pdash[], int *pnum)
{
	GraphicsState *gs;
	int i;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return(ERROR);
	}

	if (*pnum < gs->dashNum) {
		/* ԡκǿ­ʤ */
		errorno = PARAMERROR;
		*pnum = gs->dashNum;
		return (ERROR);
	}

	*pnum = gs->dashNum;
	for (i = 0; i < gs->dashNum; i++) {
		pdash[i] = gs->lineDash[i];
	}
	return(OK);
}


/** PDAPI˽ */
int SetLineDashOffset(int printerContext, Fix offset)
{
	GraphicsState *gs;

	DebugPrint("pdapi: %s,%d, offset=%x, %d\n", __FUNCTION__, __LINE__, offset, Fix2i(offset));

	// GraphicsStateμ
	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return(ERROR);
	}

	gs->dashOffset = offset;
	/* åɤΥǤ⥳ޥɤ */
	return WriteLineKind(printerContext);
}


/** PDAPI˽ */
int GetLineDashOffset(int printerContext, Fix *poffset)
{
	GraphicsState *gs;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return(ERROR);
	}

	*poffset = gs->dashOffset;
	return(OK);
}


/** PDAPI˽ */
int SetLineStyle(int printerContext, LineStyle linestyle)
{
	GraphicsState *gs;

	DebugPrint("pdapi: %s,%d printerContext=%d, lineStyle=%d\n", __FUNCTION__, __LINE__, printerContext, linestyle);

	// GraphicsStateμ
	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return(ERROR);
	}

	// GraphicsStateι
	gs->lineStyle = linestyle;

	// ޷򥳥ޥ(lpG)
	// Solid->Dashѹϡnum==0ǤSolidѤΤɬפʤ
	// num!=0ǤСѹɬס
	// Dash->Solidѹϡʤ餺ɬס
	if ((gs->lineStyle == lineStyleDash) && (gs->dashNum == 0)) {
		/* ޥɤ˽λ */
		return(OK);
	}

	return WriteLineKind(printerContext);
}


/** PDAPI˽ */
int GetLineStyle(int printerContext, LineStyle *plinestyle)
{
	GraphicsState *gs;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	// GraphicsStateμ
	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return(ERROR);
	}

	// GraphicsStateǡ
	*plinestyle = gs->lineStyle;

	return(OK);
}


/** PDAPI˽ */
int SetLineCap(int printerContext, LineCap linecap)
{
	GraphicsState *gs;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return(ERROR);
	}

	// GraphicsContextι
	gs->lineCap = linecap;

	// Line°ѹޥɤ
	return WriteLineAttributes(printerContext);
}


/** PDAPI˽ */
int GetLineCap(int printerContext, LineCap *plinecap)
{
	GraphicsState *gs;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return(ERROR);
	}

	*plinecap = gs->lineCap;
	return(OK);
}


/** PDAPI˽ */
int SetLineJoin(int printerContext, LineJoin linejoin)
{
	GraphicsState *gs;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return(ERROR);
	}

	// GraphicsContextι
	gs->lineJoin = linejoin;

	// Line°ѹޥɤ
	return WriteLineAttributes(printerContext);
}


/** PDAPI˽ */
int GetLineJoin(int printerContext, LineJoin *plinejoin)
{
	GraphicsState *gs;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return(ERROR);
	}

	*plinejoin = gs->lineJoin;
	return(OK);
}


/** PDAPI˽ */
int SetMiterLimit(int printerContext, Fix miterlimit)
{
	GraphicsState *gs;

	DebugPrint("pdapi: %s,%d, miterlimit=%x, %d\n", __FUNCTION__, __LINE__, miterlimit, Fix2i(miterlimit));

	// GraphicsStateμ
	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return(ERROR);
	}

	// GraphicsStateι
	gs->miterLimit = miterlimit;

	// MiterLimitͤΥޥϡ
	//   1. join=mitermiterlimitͤѹ줿
	//   2. join=miter򤵤줿
	// 2ĤΥߥ󥰤ɬס

	{
		// Joinμ
		LineJoin join;

		if (GetLineJoin(printerContext, &join) != OK) {
			return(ERROR);
		}
		// join = miterʤСMiterLimitѹޥɤɬ
		if (join == lineJoinMiter) {
			return WriteMiterLimit(printerContext);
		}
	}

	return(OK);
}


/** PDAPI˽ */
int GetMiterLimit(int printerContext, Fix *pmiterlimit)
{
	GraphicsState *gs;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	// GraphicsStateμ
	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return(ERROR);
	}
	*pmiterlimit = gs->miterLimit;

	return(OK);
}


/** PDAPI˽ */
int SetPaintMode(int printerContext, PaintMode paintmode)
{
	GraphicsState *gs;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return(ERROR);
	}

	gs->paintMode = paintmode;

	/* ȥե붦˱ƶΤǡξΥե饰ΩƤ */
	gs->strokePatternChanged = TRUE;
	gs->fillPatternChanged = TRUE;

	return(OK);
}


/** PDAPI˽ */
int GetPaintMode(int printerContext, PaintMode *ppaintmode)
{
	GraphicsState *gs;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	// GraphicsStateμ
	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return(ERROR);
	}

	*ppaintmode = gs->paintMode;
	return(OK);
}


/** PDAPI˽ */
int SetStrokeColor(int printerContext, Brush *brush)
{
	GraphicsState *gs;

	DebugPrint("pdapi: %s,%d printerContext=%d brush.color[1]=%X brush.color[2]=%X brush.color[3]=%X\n", __FUNCTION__, __LINE__, printerContext, brush->color[1], brush->color[2], brush->color[3]);

	/* GraphicsStateι */
	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return(ERROR);
	}

	// BrushƤGraphicsStateǻBrushΰ˥ԡ
	CopyBrush(&(gs->strokeBrush), brush);

	/* ޥɤ */
	{
		char obuf[64];

		/**
		 * Ǥ2Ĥpageޥɤ
		 *  - ѥϣο - ѥONʬοѹ
		 *  - ѥ - ѥȤ˥ѥꤹ
		 */
		/**
		 * ѥON(fpE)
		 * 1:ΰ°, 2:顼ڡ(RGB), 3:ʬ3(RGB3)
		 * 4:1ʬ(0-255), 5:2ʬ(0-255), 6:3ʬ(0-255)
		 * (TBD) castˤ뿧ǡ(2byte->1byte).
		 */
		memset(obuf, 0, 64);
		(void)sprintf(obuf, ESC_GS "1;2;3;%d;%d;%dfpE",
					  (unsigned char)brush->color[2],
					  (unsigned char)brush->color[1],
					  (unsigned char)brush->color[0]);
		if (CheckWrite(printerContext, obuf, strlen(obuf)) != OK) {
			return(ERROR);
		}
	}

	/* ѹե饰ΩƤ */
	gs->strokePatternChanged = TRUE;
	return (OK);
}


/** PDAPI˽ */
int SetFillColor(int printerContext, Brush *brush)
{
	GraphicsState *gs;

	DebugPrint("pdapi: %s,%d printerContext=%d brush.color[1]=%X brush.color[2]=%X brush.color[3]=%X\n", __FUNCTION__, __LINE__, printerContext, brush->color[1], brush->color[2], brush->color[3]);

	/* GraphicsStateι */
	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return(ERROR);
	}

	/* BrushƤGraphicsStateǻBrush˥ԡ */
	CopyBrush(&(gs->fillBrush), brush);

	/* ޥɤ */
	{

		char obuf[64];
		memset(obuf, 0, 64);

		/** ѥϣλ꡿åɥѥ(dfpE) */
		(void)sprintf(obuf, ESC_GS "1;2;3;%d;%d;%dfpE",
					  (unsigned char)brush->color[2],
					  (unsigned char)brush->color[1],
					  (unsigned char)brush->color[0]);
		if (CheckWrite(printerContext, obuf, strlen(obuf)) != OK) {
			return(ERROR);
		}
	}

	/* ѹե饰ΩƤ */
	gs->fillPatternChanged = TRUE;
	return (OK);
}


/** PDAPI˽ */
int SetBgColor(int printerContext, Brush *brush)
{
	GraphicsState *gs;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return(ERROR);
	}

	gs->bgColor[0] = brush->color[0];
	gs->bgColor[1] = brush->color[1];
	gs->bgColor[2] = brush->color[2];
	gs->bgColor[3] = brush->color[3];

	/* ȥե붦˱ƶΤǡξΥե饰ΩƤ */
	gs->strokePatternChanged = TRUE;
	gs->fillPatternChanged = TRUE;

	return(OK);
}

/** ѥŬо */
typedef enum PATTERN_OPERATION_ {
	PATTERN_FILL,		/**< ե */
	PATTERN_STROKE		/**< ȥ */
} PATTERN_OPERATION;

#define         BITS_PER_BYTE   8   /**< 1Byte == 8bits */

/** 1dot == 1bitñ̥ΥХȥХ֤
 * @param x	饹Ĺ
 * @return	ǡΥХȿ
 */
static int RowBytes8(int x)
{
	return  ( (x) + BITS_PER_BYTE - 1 ) / BITS_PER_BYTE ;
}

/** 1bit᡼bitñ̻ǤΥꥢԡ
 * žʬŪʶꥢžλꤷ֤إԡ
 * ˥ꥢǤмͳ˶Ф
 * @param *src		ԡꥢ֥ݥ
 * @param *dst		ԡ襨ꥢ֥ݥ
 * @param sx		ԡꥢϰ̣ءdots
 * @param sy		ԡꥢϰ̣١dots
 * @param dx		ԡ襨ꥢϰ̣ءdots
 * @param dy		ԡ襨ꥢϰ̣١dots
 * @param copydotsH	ԡdots
 * @param copydotsW	ԡ⤵dots
 * @param width		srcdstΥꥢdots
 * @return		ʤOK򡢥顼ʤERROR֤
 * @note ¿롣
 */
static int CopyAreaByBits(char *src,
			  char *dst,
			  int sx,
			  int sy,
			  int dx,
			  int dy,
			  int copydotsH,
			  int copydotsW,
			  int width
			  )
{
	int sx_bytes;   /* ԡꥢءbitñ̡ˤΥХȥХꥵ */
	int dx_bytes;   /* ԡ襨ꥢءbitñ̡ˤΥХȥХꥵ */
	int lineSize_bytes; /* ꥢbitñ̡ˤΥХȥХꥵ     */
	char target;                  /* åbitޤॳԡХ */
	char shiftS;    /* å뤿Υӥåȥե                  */
	char shiftD;    /* å뤿Υӥåȥե                  */
	int offset,i,j;

	lineSize_bytes = RowBytes8(width);

	for(i=0; i<copydotsH; i++){
		for(j=0; j<copydotsW; j++){

			/* ԡ ----------------------------------------*/
			/* ˾ɸˤӥåȡʵڤӤޤХȡˤ */
			sx_bytes = (j+sx)/8;
			shiftS   = 7 - ((j+sx)%BITS_PER_BYTE);
			offset   = sx_bytes + ((i+sy) * lineSize_bytes);
			target   = *(src + offset);
			target   = target >> shiftS;
			target  &= 0x01;

			/* ԡ ----------------------------------------*/
			/* ˾ɸˤӥåȡʵڤӤޤХȡˤ */
			/* ƥԡΥӥåȤ򥻥å */
			dx_bytes = (j+dx)/8;
			offset   = dx_bytes + ((i+dy) * lineSize_bytes);

			shiftD   = 7 - ((j+dx)%BITS_PER_BYTE);
			target   = target << shiftD;

			*(dst+offset) |= target; /* žɸ֤إӥåȤ򥳥ԡ */
		}
	}
	return (OK);
}


/** 1bit depth᡼Υ꡼ưԤ
 * xorg,yorg򺸾帶ʣˤȤƲǡ¤ؤ
 *ʣǢ'''إԡ
 * @verbatim
 *  (0,0) +--------+----------------+
 *        |  '   |    '         |
 *        +--------+(xorg,yorg) ----+--------+
 *        |        |                |        |
 *        |  '   |    'ᣱ     |      |
 *        |        |                |        |
 *        +--------+----------------+--------+
 *                 |              |      |
 *                 +----------------+--------+
 * @endverbatim
 * @param xorg		꡼ 
 * @param yorg		꡼
 * @param width		ɥå
 * @param height	ɥåȹ⤵
 * @param *srcData	ꥢǡ
 * @param *distData	ǥƥ͡󥨥ꥢ
 * @return	ʤOK򡢥顼ʤERROR֤
 * @note ХȥХΥǡΤߥݡȤ롣ʳΥǡ
 *       ݾڤʤ
 */
static int ShiftOrigin(int xorg,
		       int yorg,
		       int width,
		       int height,
		       char *srcData,
		       char *distData
		       )
{
	int sx;         /* ԡءdots=bit */
	int sy;         /* ԡ١dots=bit */
	int dx;         /* ԡ襽ءdots=bit */
	int dy;         /* ԡ襽١dots=bit */
	int copydotsH;  /* ԡɥåȹ⤵dots=bit */
	int copydotsW;  /* ԡɥådots=bit   */

	{
		/* ѥ᡼å ----------------------------------------------*/
		if(width<=0){                               /* å ----------*/
			errorno = PARAMERROR;
			return (ERROR);
		}
		if(height<=0){                              /* ⤵å --------*/
			errorno = PARAMERROR;
			return (ERROR);
		}
		if(xorg>width || xorg<0){                   /* ɸإå ------*/
			errorno = PARAMERROR;
			return (ERROR);
		}
		if(yorg>height || yorg<0){                  /* ɸ٥å ------*/
			errorno = PARAMERROR;
			return (ERROR);
		}
	}

	/* ΰ裱ΰ裱ؤΥԡ --------------------------------------------*/
	/* ̤η */
	sx = xorg;
	sy = yorg;
	dx = 0;
	dy = 0;
	copydotsH = height - yorg;
	copydotsW = width  - xorg;

	CopyAreaByBits(srcData,distData,sx,sy,dx,dy,copydotsH,copydotsW,width);

	/* ΰ裴'ΰ裴ؤΥԡ -------------------------------------------*/
	/* ̤η */
	sx = 0;
	sy = 0;
	dx = width - xorg;
	dy = height- yorg;
	copydotsH = yorg;
	copydotsW = xorg;

	CopyAreaByBits(srcData,distData,sx,sy,dx,dy,copydotsH,copydotsW,width);

	/* ΰ裲'ΰ裲ؤΥԡ -------------------------------------------*/
	/* ̤η */
	sx = 0;
	sy = yorg;
	dx = width - xorg;
	dy = 0;
	copydotsH = height - yorg;
	copydotsW = xorg;

	CopyAreaByBits(srcData,distData,sx,sy,dx,dy,copydotsH,copydotsW,width);


	/* ΰ裳'ΰ裳ؤΥԡ -------------------------------------------*/
	/* ̤η */
	sx = xorg;
	sy = 0;
	dx = 0;
	dy = height - yorg;
	copydotsH = yorg;
	copydotsW = width - xorg;

	CopyAreaByBits(srcData,distData,sx,sy,dx,dy,copydotsH,copydotsW,width);

	return (OK);
}


/** ֥饷Ͽ
 * @param printerContext	OpenPrinter()줿ץ󥿥ƥȤ
 *				ꤹ롣
 * @param brush			֥饷
 * @return ˽λOK ֤롣顼ä-1 
 *         졢errorno ˥顼ɤǼ롣
 */
static int WriteBrush(int printerContext, Brush brush)
{
	DeviceInfo *dev;
	BrushData *pbrush;
	int width_bytes;
	char *obuf, *buf;
	int *dotList;
	int qList[] = { 16, 32, 0 };
	int fList[] = { 16, 32, 64, 0 };
	int i;

	dev = GetDeviceInfo(printerContext);
	if (dev == NULL) {
		return (ERROR);
	}

	pbrush = brush.pbrush;

	/* ǽγǧ */
	if (dev->resolution == RES600) {
		dotList = fList;
	} else if (dev->resolution == RES300) {
		dotList = qList;
	} else {
		errorno = FATALERROR;
		return (ERROR);
	}
	for (i = 0; dotList[i] != 0; i++) {
		if (pbrush->height == dotList[i]
		    && pbrush->height == dotList[i]) {
			break;
		}
	}
	if (dotList[i] == 0) {
		/* ʤΥѥϤ줿 */
		errorno = NOTSUPPORTED;
		return (ERROR);
	}

	width_bytes = ((pbrush->width + 7) >> 3) & 0x1F;
	/* ºݤϿѥǡǼΰ */
	buf = (char *)calloc(width_bytes * pbrush->height, sizeof(char));
	if (buf == NULL) {
		errorno = FATALERROR;
		return (ERROR);
	}

	/**
	   @todo ̤Ƥǡϡѥå󥰤Ƥ
	         ǽ١ǤϤʬΥߤɬפ
	         뤬ʤ̵
	*/

	/* ꡼ꤵƤ */
	if (brush.xorg != 0 || brush.yorg != 0) {
		if (ShiftOrigin(brush.xorg, brush.yorg, pbrush->width, pbrush->height, brush.pbrush->data, buf) != OK) {
			return (ERROR);
		}
	}

	/** ͥѥϿ(sp{E) ޥɤȯԤ */
	(void)sprintf(obuf, ESC_GS "%d;%d;%d;%d;%dsp{E",
		      width_bytes * pbrush->height,
		      0,
		      0,
		      pbrush->width,
		      pbrush->height);

	return CheckWrite(printerContext, obuf, strlen(obuf));
}

/** ѥꥳޥɤ
 * @param printerContext	OpenPrinter()줿ץ󥿥ƥȤ
 *				ꤹ롣
 * @param operation		ޥɤоݤꤹ
 * @return ˽λOK ֤롣顼ä-1 
 *         졢errorno ˥顼ɤǼ롣
 */
static int WritePattern(int printerContext, PATTERN_OPERATION operation)
{
	GraphicsState *gs;
	int pt, id, mode;
	char obuf[128];
	int result;

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return (ERROR);
	}

	/* ޥɥѥ᡼
	   (TBD) Ͽֹϡ0ˤƤ */
	if (operation == PATTERN_FILL) {
		if (gs->fillPatternChanged == FALSE) {
			/* ѥ᡼ѹƤʤΤǡޥ
			   ˽λ */
			return (OK);
		}
		if (gs->fillBrush.pbrush == NULL) {
			pt = 1;
			id = 0;
			mode = 0;
		} else {
			result = WriteBrush(printerContext, gs->fillBrush);
			if (result != OK) {
				return (result);
			}
			pt = 2;
			id = 0;
			if (gs->paintMode == paintModeOpaque) {
				mode = 2;
			} else if (gs->paintMode == paintModeTransparent) {
				mode = 1;
			} else {
				/* ݡȤʤ⡼ */
				errorno = NOTSUPPORTED;
				return (ERROR);
			}
		}


		 /* FILLꤵ줿ˤϡե롢ȥå饹
		  * ȥå顼3Ĥѹ */
		/** 顼ѥϿ(cpE) ޥɤȯԤ */
		(void)sprintf(obuf,
			      ESC_GS "3;2;%d;%d;%dcpE"
			      ESC_GS "4;2;%d;%d;%dcpE"
			      ESC_GS "5;2;%d;%d;%dcpE",
			      pt, id, mode, pt, id, mode, pt, id, mode);

	} else if (operation == PATTERN_STROKE) {
		if (gs->strokePatternChanged == FALSE) {
			/* ѥ᡼ѹƤʤΤǡޥ
			   ˽λ */
			return (OK);
		}
		if (gs->strokeBrush.pbrush == NULL) {
			pt = 1;
			id = 0;
			mode = 0;
		} else {
			result = WriteBrush(printerContext, gs->strokeBrush);
			if (result != OK) {
				return (result);
			}

			pt = 2;
			id = 0;
			if (gs->paintMode == paintModeOpaque) {
				mode = 2;
			} else if (gs->paintMode == paintModeTransparent) {
				mode = 1;
			} else {
				/* ݡȤʤ⡼ */
				errorno = NOTSUPPORTED;
				return (ERROR);
			}
		}
		/** ط°⡼(bgdE) ޥɤȯԤ
		 *  ȥƩ᡿Ʃѹ
		 */
		(void)sprintf(obuf, ESC_GS "2;2;%d;%d;%dcpE" ESC_GS "2;%dbgdE", pt, id, mode, mode);

	} else {
		/* оݤ̤ݡ */
		errorno = NOTSUPPORTED;
		return (ERROR);
	}

	result = CheckWrite(printerContext, obuf, strlen(obuf));
	if (result != OK) {
		return (result);
	}

	/* ǤСѹե饰ݤƤ */
	if (operation == PATTERN_FILL) {
		gs->strokePatternChanged = FALSE;
	} else if (operation == PATTERN_STROKE) {
		gs->fillPatternChanged = FALSE;
	}
	return (OK);
}

/* ----------------------------------------------------------------
 *
 * Path Operations
 *
 * ---------------------------------------------------------------- */

/* 
 * @todo Path Operations
 *  - FlashPathΥޥɥѥ᡼ϡäȴά
 *  - SetLintPathBezierPathϥꥹȤ¿١mallocΥ
 *    Сإåɤ礭⤷ʤ
 *  - PolygonPath, ArcPath̤
 *  - ѥ᡼ͤꤷƤĽ꤬¿
 */


/** ѥϿֹ
    @note ɬפ̵ͭGS-LPDǰʲͤ˻ꤵƤΤǡ
          Τޤ޿򤹤롣
 */
typedef enum EscpagePathId_e {
	ESCPAGE_PATH_ID_PRINT = 0, /**<  */
	ESCPAGE_PATH_ID_CLIP = 1 /**< å */
} EscpagePathId;

/** ѥ */
typedef enum EscpagePathType {
	ESCPAGE_PATH_TYPE_STROKE = 0, /**< ȥ */
	ESCPAGE_PATH_TYPE_FILL = 1, /**< ե */
	ESCPAGE_PATH_TYPE_STROKEFILL = 2, /**< ե륹ȥ */
	ESCPAGE_PATH_TYPE_CLIP_WINDING = 3, /**< å(Non-zero winding number rule) */
	ESCPAGE_PATH_TYPE_CLIP_EVEN_ODD = 4 /**< å(Even-odd rule) */
} EscpagePathType;

static int NewSubPath(int printerContext, Fix x, Fix y);
static int EndSubPath(int printerContext);
static int FlashPath(int printerContext, unsigned char id, EscpagePathType type);
static int CommitRectanglePath(int printerContext, Point point0, Point point1, Fix xellipse, Fix yellipse);


/** ֥ѥ򳫻Ϥ
 * @param printerContext	OpenPrinter()줿ץ󥿥ƥȤ
 *				ꤹ롣
 * @param x	֥ѥϺɸx
 * @param y	֥ѥϺɸy
 * @return ˽λOK ֤롣顼ä-1 
 *         졢errorno ˥顼ɤǼ롣
 */
static int NewSubPath(int printerContext, Fix x, Fix y)
{
	GraphicsState *gs;
	char obuf[64];

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return (ERROR);
	}

	/** ֥ѥ(mvpG) ޥɤȯԤ */
	(void)sprintf(obuf, ESC_GS "%d;%d;%dmvpG",
		      0, /* 0= 1= */
		      (int)ApplyCtmX_Integer(x, y, gs->ctm),
		      (int)ApplyCtmY_Integer(x, y, gs->ctm));
	
	/* ꥹȤκǸ˥ޥɤɲ */
	gs->pathEndPoint = AddList(gs->pathList, obuf, strlen(obuf));
	if (gs->pathEndPoint == NULL) {
		return (ERROR);
	}
	return (OK);
}


/** ֥ѥĤ
 * @param printerContext	OpenPrinter()줿ץ󥿥ƥȤ
 *				ꤹ롣
 * @return ˽λOK ֤롣顼ä-1 
 *         졢errorno ˥顼ɤǼ롣
 */
static int EndSubPath(int printerContext)
{
	GraphicsState *gs;
	char obuf[64];

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return (ERROR);
	}

	/** ֥ѥϿλ(clpG) ޥɤȯԤ */
	(void)sprintf(obuf, ESC_GS "clpG");

	/* ꥹȤκǸ˥ޥɤɲ */
	gs->pathEndPoint = AddList(gs->pathList, obuf, strlen(obuf));
	if (gs->pathEndPoint == NULL) {
		return (ERROR);
	}

	return (OK);
}

/** ȥѥۤ޷褹
 *	ȥѥۤ޷褷ޤGraphicsState
 *	(current pathޤ)ݻޤ
 * @param printerContext	OpenPrinter()줿ץ󥿥ƥȤ
 *				ꤹ롣
 * @param id			ѥϿֹ
 * @note
 * 	ѥǤID=0ΤߤѤƤ뤬ϤID
 * 	ĤɤΤȽǤĤʤ١ǽˤƤtype - 
 * 	
 * @param type			fixμ
 * @return ˽λOK ֤롣顼ä-1 
 *         졢errorno ˥顼ɤǼ롣
 */
static int FlashPath(int printerContext, unsigned char id, EscpagePathType type)
{
	GraphicsState *gs;
	char obuf[64];
	ItemList *node;
	int param;
	int result;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return (ERROR);
	}

	/* ֳǧ */
	if (gs->pathActive == TRUE) {
		/* ѥǤ(EndPathƤФƤʤ) */
		errorno = BADREQUEST;
		return (ERROR);
	} else if (gs->pathList == NULL) {
		/* ѥꥹȤ¸ߤʤΤǡ⤻˽λ */
		return (OK);
	}

	/** ѥϿ(bgpG) ޥɤȯԤ */
	(void)sprintf(obuf, ESC_GS "%dbgpG", id);

	/* ꥹȤƬ˥ޥɤɲ(ư) */
	if (gs->pathList->item != NULL) {
		free(gs->pathList->item);
	}
	gs->pathList->length = strlen(obuf);
	gs->pathList->item = malloc(gs->pathList->length);
	if (gs->pathList->item == NULL) {
		errorno = FATALERROR;
		return (ERROR);
	}

	memcpy(gs->pathList->item, obuf, gs->pathList->length);

	/* ѥ襳ޥ */
	if (type == ESCPAGE_PATH_TYPE_STROKE) {
		param = 0;

	} else if (type == ESCPAGE_PATH_TYPE_FILL) {
		if (gs->fillMode == fillModeEvenOdd) {
			param = 2;

		} else if (gs->fillMode == fillModeWinding) {
			param = 1;

		} else {
			/* ݡȤƤʤpath mode򤵤줿 */
			errorno = NOTSUPPORTED;
			return (ERROR);
		}
	} else if (type == ESCPAGE_PATH_TYPE_STROKEFILL) {
		if (gs->fillMode == fillModeEvenOdd) {
			param = 4;

		} else if (gs->fillMode == fillModeWinding) {
			param = 3;

		} else {
			/* ݡȤƤʤpath mode򤵤줿 */
			errorno = NOTSUPPORTED;
			return (ERROR);
		}

	} else if (type == ESCPAGE_PATH_TYPE_CLIP_EVEN_ODD) {
		param = 2;

	} else if (type == ESCPAGE_PATH_TYPE_CLIP_WINDING) {
		param = 1;

	} else {
		/* ݡȤƤʤpath type򤵤줿 */
		errorno = NOTSUPPORTED;
		return (ERROR);
	}

	if (type == ESCPAGE_PATH_TYPE_CLIP_EVEN_ODD
	    || type == ESCPAGE_PATH_TYPE_CLIP_WINDING) {
		/** ѥåΰ(capG) ޥɤȯԤ */
		(void)sprintf(obuf, ESC_GS "%d;%dcapG", id, param);
	} else {
		/** ѥ(drpG) ޥɤȯԤ */
		(void)sprintf(obuf, ESC_GS "%d;%ddrpG", id, param);
	}

	/* ꥹȤκǸ˥ޥɤɲ */
	/* FillڤStrokeFillξ硢λĤƤʤѥ
	   ۤĤ */
	gs->pathEndPoint = AddList(gs->pathList, obuf, strlen(obuf));
	if (gs->pathEndPoint == NULL) {
		return (ERROR);
	}

	/* ROPͤѹƤСޥɤ */
	result = WriteROP(printerContext);
	if (result != OK) {
		return (result);
	}

	/* ѥѹƤСޥɤ */
	if (type == ESCPAGE_PATH_TYPE_STROKE
	    || type == ESCPAGE_PATH_TYPE_STROKEFILL) {
		WritePattern(printerContext, PATTERN_STROKE);
	}
	if (type == ESCPAGE_PATH_TYPE_FILL
	    || type == ESCPAGE_PATH_TYPE_STROKEFILL) {
		WritePattern(printerContext, PATTERN_FILL);
	}

	/* ѥFDؽ񤭽Ф */
	node = gs->pathList;
	while (node != NULL) {
		result = CheckWrite(printerContext, node->item, node->length);
		if (result != OK) {
			/* 񤭹߼ */
			return (result);
		}
		node = node->next;
	}
	return (OK);
}


/** ѥѥꥹϿ
 *	Cornerxellipseyellipseɽʱ߸̤ˤä³롣
 * @param printerContext	OpenPrinter()줿ץ󥿥ƥȤꤹ롣
 * @param point0		
 * @param point1		г
 * @param xellipse		ʱߤοʿĹ
 * @param yellipse		ʱߤοľĹ
 * @return ˽λOK ֤롣顼ä-1 
 *         졢errorno ˥顼ɤǼ롣
 */
static int CommitRectanglePath(int printerContext, Point point0, Point point1, Fix xellipse, Fix yellipse)
{
	GraphicsState *gs;
	char obuf[128];

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return (ERROR);
	}

	/* (rrpG) ޥɤȯԤ */
	(void)sprintf(obuf, ESC_GS "%d;%d;%d;%d;%d;%d;%drrpG",
		      0, /* 0 */
		      (int)ApplyCtmX_Integer((point0.x), (point0.y), gs->ctm),
		      (int)ApplyCtmY_Integer((point0.x), (point0.y), gs->ctm),
		      (int)ApplyCtmX_Integer((point1.x), (point1.y), gs->ctm),
		      (int)ApplyCtmY_Integer((point1.x), (point1.y), gs->ctm),
		      (int)ApplyCtmX_Integer((xellipse), (yellipse), gs->ctm),
		      (int)ApplyCtmY_Integer((xellipse), (yellipse), gs->ctm));

	/* ѥɲä */
	gs->pathEndPoint = AddList(gs->pathList, obuf, strlen(obuf));
	if (gs->pathEndPoint == NULL) {
		return (ERROR);
	}

	return(OK);
}


/** PDAPI˽ */
int NewPath(int printerContext)
{
	GraphicsState *gs;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return (ERROR);
	}

	/**
	 * ŤѥĤäƤкѥ򳫻Ϥ롣path
	 * ϥޥɤϡѥ񤭽Фľ˥ѥ᡼򥻥åȤɬ
	 * פΤǡǤ϶ˤƤ
	 */
	if (gs->pathList != NULL) {
		DeleteListAll(gs->pathList);
	}
	gs->pathList = NewList("", 0);
	if (gs->pathList == NULL) {
		return (ERROR);
	}

	/* ѥǤ򼨤ե饰򥻥å */
	gs->pathActive = TRUE;

	return (OK);
}


/** PDAPI˽ */
int EndPath(int printerContext)
{
	GraphicsState *gs;
	char obuf[64];

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return (ERROR);
	}

	if (gs->pathActive != TRUE) {
		/* NewPathƤФƤʤΤˡEndPathƤФ줿 */
		errorno = BADREQUEST;
		return (ERROR);
	}

	/* ѥϿλ(enpG) ޥɤȯԤ */
	(void)sprintf(obuf, ESC_GS "enpG");

	/* ꥹȤκǸ˥ޥɤɲ */
	gs->pathEndPoint = AddList(gs->pathList, obuf, strlen(obuf));
	if (gs->pathEndPoint == NULL) {
		return (ERROR);
	}

	/* ѥ꤬λƤ򼨤ե饰򥻥å */
	gs->pathActive = FALSE;
	return (OK);
}


/** PDAPI˽ */
int StrokePath(int printerContext)
{
	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	return FlashPath(printerContext, ESCPAGE_PATH_ID_PRINT, ESCPAGE_PATH_TYPE_STROKE);
}


/** PDAPI˽ */
int FillPath(int printerContext)
{
	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	return FlashPath(printerContext, ESCPAGE_PATH_ID_PRINT, ESCPAGE_PATH_TYPE_FILL);
}


/** PDAPI˽ */
int StrokeFillPath(int printerContext)
{
	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	return FlashPath(printerContext, ESCPAGE_PATH_ID_PRINT, ESCPAGE_PATH_TYPE_STROKEFILL);
}


/** PDAPI˽ */
int SetClipPath(int printerContext, ClipRule clipRule)
{
	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	if (clipRule == clipRuleEvenOdd) {
		return FlashPath(printerContext, ESCPAGE_PATH_ID_CLIP, ESCPAGE_PATH_TYPE_CLIP_EVEN_ODD);

	} else if (clipRule == clipRuleWinding) {
		return FlashPath(printerContext, ESCPAGE_PATH_ID_CLIP, ESCPAGE_PATH_TYPE_CLIP_WINDING);

	}

	/* ̤ݡȤΥ */
	errorno = NOTSUPPORTED;
	return (ERROR);
}

/** PDAPI˽ */
int ResetClipPath( int printerContext)
{
	char obuf[64];

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);
	/** åץ⡼/(cmE) ޥɤȯԤ */
	(void)sprintf(obuf, ESC_GS "0cmE" ESC_GS "2cmE");
	return CheckWrite(printerContext, obuf, strlen(obuf));
}


/** PDAPI˽ */
int SetCurrentPoint(int printerContext, Fix x, Fix y)
{
	GraphicsState *gs;
	char obuf[64];

	DebugPrint("pdapi: %s,%d printerContext=%d, x=%d, y=%d\n", __FUNCTION__, __LINE__, printerContext, Fix2i(x), Fix2i(y));

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return (ERROR);
	}

	if (gs->pathActive == TRUE) {
		/** pathϿξϡ֥ѥ򳫻Ϥ */
		return NewSubPath(printerContext, x, y);
	}

	/** ȥݥȻ(X, Y) ޥɤȯԤ */
	(void)sprintf(obuf, ESC_GS "%dX" ESC_GS "%dY",
		      (int)ApplyCtmX_Integer(x, y, gs->ctm),
		      (int)ApplyCtmY_Integer(x, y, gs->ctm));

	return CheckWrite(printerContext, obuf, strlen(obuf));
}


/** PDAPI˽
 * @note ESC/PageǤƱͤˡȥݥȤϺǸ˰ưΤǡ
 *       ѥɲø˥ȥѥԤɬפ̵
 */
int LinePath(int printerContext, int flag, int npoints, Point *points)
{
	GraphicsState *gs;
	char obuf[64];
	int i;

	DebugPrint("pdapi: %s,%d flag=%d\n", __FUNCTION__, __LINE__, flag);

	if (npoints < 1) {
		/* ĺ */
		errorno = PARAMERROR;
		return (ERROR);
	}

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return (ERROR);
	}

	/** ݥ饤ѥɲ(lnpG) ޥɤȯԤ */
	/* ݥ饤ѥɲåޥ(prefix) */
	(void)sprintf(obuf, ESC_GS "%d;%d",
		      0, 	/* а֥ѥɸ⡼ɤ˸ */
		      npoints);
	/* ꥹȤκǸ˥ޥɤɲ */
	gs->pathEndPoint = AddList(gs->pathList, obuf, strlen(obuf));
	if (gs->pathEndPoint == NULL) {
		return (ERROR);
	}

	/* ݥ饤ѥɲåޥ(ĺɸ) */
	for (i = 0; i < npoints; i++) {
		/* (TBD) CTMŬѤϡѹ */
		(void)sprintf(obuf, ";%d;%d",
			      (int)ApplyCtmX_Integer((points[i].x), (points[i].y), gs->ctm),
			      (int)ApplyCtmY_Integer((points[i].x), (points[i].y), gs->ctm));
		/* ꥹȤκǸ˥ޥɤɲ */
		gs->pathEndPoint = AddList(gs->pathList, obuf, strlen(obuf));
		if (gs->pathEndPoint == NULL) {
			return (ERROR);
		}
	}

	/* ݥ饤ѥɲåޥ(suffix) */
	(void)sprintf(obuf, "lnpG");
	/* ꥹȤκǸ˥ޥɤɲ */
	gs->pathEndPoint = AddList(gs->pathList, obuf, strlen(obuf));
	if (gs->pathEndPoint == NULL) {
		return (ERROR);
	}

	/* ֥ѥ򥯥 */
	if (flag == PathClose) {
		return EndSubPath(printerContext);
	}
	/* ֥ѥ򥯥ʤ */
	return (OK);
}


/** PDAPI˽(̤ݡ) */
int PolygonPath(int printerContext, int npolygons, int *nvertexes, Point *points)
{
	int		result;
	result = 0;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	return(result);
}


/** PDAPI˽ */
int RectanglePath(int printerContext, int nrectangles, Rectangle *rectangles)
{
	Fix fix0;
	int i;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	if (nrectangles <= 0) {
		/*  */
		errorno = PARAMERROR;
		return (ERROR);
	}

	memset(&fix0, 0, sizeof(Fix));
	for (i = 0; i < nrectangles; i++) {
		int result;

		result = CommitRectanglePath(printerContext, rectangles[i].p0, rectangles[i].p1, fix0, fix0);
		if (result != OK) {
			return (result);
		}
	}
	return(OK);
}


/** PDAPI˽ */
int RoundRectanglePath(int printerContext, int nrectangles, RoundRectangle *rectangles)
{
	int i;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	if (nrectangles <= 0) {
		/*  */
		errorno = PARAMERROR;
		return (ERROR);
	}

	for (i = 0; i < nrectangles; i++) {
		int result;

		result = CommitRectanglePath(printerContext, rectangles[i].p0, rectangles[i].p1,
					     rectangles[i].xellipse, rectangles[i].yellipse);
		if (result != OK) {
			return (result);
		}
	}
	return(OK);
}


/** PDAPI˽ */
int BezierPath(int printerContext, int npoints, Point *points)
{
	GraphicsState *gs;
	char obuf[64];
	int offset;
	int j;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	/* Ф(ĤǤ˥ѥ᡼ǧ) */
	if ((npoints % 3) != 0) {
		/* ĺ3ܿǤʤ */
		errorno = PARAMERROR;
		return (ERROR);
	}

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return (ERROR);
	}

	offset = 0;
	/** Bezierɲ(bzpG) ޥɤȯԤ */
	/* Bezierɲåޥ(prefix) */
	(void)sprintf(obuf, ESC_GS "%d;%d",
		      0, /* 0=Хѥ 1=Хѥ */
		      npoints); /* ĺ */
	/* ꥹȤκǸ˥ޥɤɲ */
	gs->pathEndPoint = AddList(gs->pathList, obuf, strlen(obuf));
	if (gs->pathEndPoint == NULL) {
		errorno = FATALERROR;
		return (ERROR);
	}

	for (j = 0; j < npoints; j++) {
		(void)sprintf(obuf, ";%d;%d",
			      (int)ApplyCtmX_Integer((points[offset].x), (points[offset].y), gs->ctm), /* x */
			      (int)ApplyCtmY_Integer((points[offset].x), (points[offset].y), gs->ctm)); /* y */
		offset ++;
		
		/* ꥹȤκǸ˥ޥɤɲ */
		gs->pathEndPoint = AddList(gs->pathList, obuf, strlen(obuf));
		if (gs->pathEndPoint == NULL) {
			return (ERROR);
		}
	}
	/* Bezierɲåޥ(suffix) */
	(void)sprintf(obuf, "bzpG");
	
	/* ꥹȤκǸ˥ޥɤɲ */
	gs->pathEndPoint = AddList(gs->pathList, obuf, strlen(obuf));
	if (gs->pathEndPoint == NULL) {
		return (ERROR);
	}
	return(OK);
}


/** PDAPI˽(̤ݡ) */
int ArchPath(int printerContext, int kind, int dir, Fix bbx0, Fix bby0, Fix bbx1, Fix bby1, Fix x0, Fix y0, Fix x1, Fix y1)
{
	int		result;
	result = 0;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	return(result);
}


/* ----------------------------------------------------------------
 *
 * Raw Data Operations
 *
 * ---------------------------------------------------------------- */

/** ꤵ줿ǡХȥХѴ
 * ⤷ѴѴΥǡγǼ꤬ƱǤСѴϾ񤭤롣
 * @param srcWidthBytes ѴΥǡΥХȿ
 * @param distWidthBytes ѴΥǡΥХȿ
 * @param height ǡι⤵
 * @param srcData ѴΥǡǼؤΥݥ
 * @param distData ѴΥǡǼؤΥݥ
 * @return ɬOK֤
 */
static int ToByteBoundary(int srcWidthBytes, int distWidthBytes, int height, char *srcData, char *distData)
{
	int i;

	if (srcWidthBytes == distWidthBytes) {
		/* Ѵɬפ̵ */
		if (srcData == distData) {
			return (OK);
		} else {
			memcpy(distData, srcData, srcWidthBytes * height);
			return (OK);
		}
	}

	if (srcData == distData) {
		for (i = 0; i < height; i++) {
			memmove((distData + (distWidthBytes * i)),
				(srcData + (srcWidthBytes * i)),
				distWidthBytes);
		}
	} else {
		for (i = 0; i < height; i++) {
			memcpy((distData + (distWidthBytes * i)),
			       (srcData + (srcWidthBytes * i)),
			       distWidthBytes);
		}
	}

	return (OK);
}


/** rawǡ򳫻Ϥ
 * @param printerContext	OpenPrinter()ǼprinterContext
 * @param sourceWidth		β
 * @param sourceHeight		ν
 * @param colorDepth		
 * @param destinationWidth	ǥƥ͡β
 * @param destinationHeight	ǥƥ͡ν
 * @return ˽λOK ֤롣顼ä-1 
 *         졢errorno ˥顼ɤǼ롣
 */
static int StartRaws(int printerContext, int sourceWidth,
		    int sourceHeight, int colorDepth,
		    int destinationWidth, int destinationHeight)
{
	GraphicsState *gs;
	char obuf[64];
	int format;
	int id;
	unsigned long 
	int result;


	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return (ERROR);
	}

	/**
	 *  1 ڤ 24 Τߤ򥵥ݡȤ
	 */
	/* (TBD)ϿֹϤǤ褤 */
	if (colorDepth == 1) {
		format = 201;
		id = 1;
	} else if (colorDepth == 24) {
		format = 102;
		id = 0;
	} else {
		/* ݡȤʤdepthϤ줿 */
		errorno = NOTSUPPORTED;
		return (ERROR);
	}

	/* ROPͤѹƤСޥɤ */
	WriteROP(printerContext);

	/* ѥѹƤСޥɤ */
	WritePattern(printerContext, PATTERN_FILL);

	/** ȥå顼᡼賫(scrI) ޥɤȯԤ */
	(void)sprintf(obuf, ESC_GS "%d;%d;%d;%d;%d;%d;%d;%d;%dscrI",
		      2,	/* ColorSpace = RGB */
		      format,	/* dataη */
		      id,	/* ѥåϿֹ */
		      0,	/* ̵ */
		      sourceWidth, /*  */
		      sourceHeight, /* ι⤵ */
		      destinationWidth,	/* ΰ */
		      destinationHeight, /* ΰι⤵ */
		      0);	/* ž (TBD)ROP? */

	result =  CheckWrite(printerContext, obuf, strlen(obuf));
	if (result != OK) {
		return (result);
	}

	gs->imageWidth = sourceWidth;
	gs->imageHeight = sourceHeight;
	gs->imageDepth = colorDepth;

	return (OK);
}

/** rawǡץ󥿤
 * @param printerContext	OpenPrinter()ǼprinterContext
 * @param ndata			rawǡΥХȿ
 * @param *rawsData		rawǡ
 * @return ˽λOK ֤롣顼ä-1 
 *         졢errorno ˥顼ɤǼ롣
 */
static int WriteRaws(int printerContext, int ndata, void *rawsData)
{
	GraphicsState *gs;
	char obuf[64];
	char *writeData;
	int bytes_pre_pixel;
	unsigned int width_bytes;
	unsigned int prt_width_bytes;
	int result;

	if (ndata == 0) {
		/* ̵ */
		return (OK);
	}

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return (ERROR);
	}

	/* 饹ΥХȿ */
	width_bytes = ndata / gs->imageHeight;

	/* 1pixelդΥХȿ */
	if (gs->imageDepth < 24) {
		bytes_pre_pixel = 1;
	} else {
		bytes_pre_pixel = 3;
	}

	/* Ԥ饹ΥХȿ */
	if (gs->imageDepth == 1) {
		prt_width_bytes = (gs->imageWidth + 7) >> 3;
	} else {
		prt_width_bytes = gs->imageWidth * bytes_pre_pixel;
	}

	/* @note ̤Ƥǡϡѥå󥰤Ƥ
	   ǽ١ǤϤʬΥߤɬפ */
	if (width_bytes != prt_width_bytes) {
		writeData = (char *)malloc(prt_width_bytes * gs->imageHeight);
		if (writeData == NULL) {
			errorno = FATALERROR;
			return (ERROR);
		}
		ToByteBoundary(width_bytes, prt_width_bytes, gs->imageHeight, rawsData, writeData);
	} else {
		writeData = rawsData;
	}

	/* 顼᡼(cu{I) ޥɤȯԤ */
	(void)sprintf(obuf, ESC_GS "%d;%dcu{I",
		      prt_width_bytes * gs->imageHeight, /* ǡΥХȿ */
		      gs->imageHeight);	/* Хɿ */

	result = CheckWrite(printerContext, obuf, strlen(obuf));
	if (result != OK) {
		/* 񤭹߼ */
		if (width_bytes != prt_width_bytes) {
			free(writeData);
		}
		return (result);
	}
	result = CheckWrite(printerContext, writeData, prt_width_bytes * gs->imageHeight);

	if (width_bytes != prt_width_bytes) {
		free(writeData);
	}
	return (result);
}

/** rawǡλ
 * @param printerContext	OpenPrinter()ǼprinterContext
 * @return ˽λOK ֤롣顼ä-1 
 *         졢errorno ˥顼ɤǼ롣
 */
static int EndRaws(int printerContext)
{
	/** 顼᡼轪λ(ecrI) ޥɤȯԤ */
	return CheckWrite(printerContext, ESC_GS "ecrI", strlen(ESC_GS "ecrI"));
}


/* ----------------------------------------------------------------
 *
 * Text Operations
 *
 * ---------------------------------------------------------------- */

/** PDAPI˽
 * @todo ̤ƥ
 */
int DrawBitmapText(int printerContext, int width, int height, int pitch, void *fontdata)
{
	char obuf[64];
	int widthBytes;
	int id;
	int result;

	result = 0;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	/* ȥݥȤʤΤǡ˰ư */
	/** ľ߰л(V) ޥɤȯԤ */
	(void)sprintf(obuf, ESC_GS "%dV",
		      0 - height); /* ľа */

	result =  CheckWrite(printerContext, obuf, strlen(obuf));
	if (result != OK) {
		return (result);
	}

	/* Ͽֹ0Ȥ */
	id = 0;
	widthBytes = pitch * height;
	/** ӥåȥ᡼Ͽ(db{F) ޥɤȯԤ */
	(void)sprintf(obuf, ESC_GS "%d;%d;%d;%d;%ddb{F",
		      widthBytes, /* ǡΥХȿ */
		      id,	/* Ͽֹ */
		      width,	/*  */
		      height,	/* ⤵ */
		      0);	/* ž */

	result =  CheckWrite(printerContext, obuf, strlen(obuf));
	if (result != OK) {
		return (result);
	}


	result =  CheckWrite(printerContext, fontdata, widthBytes);
	if (result != OK) {
		return (result);
	}
	/** ʸӥåȥ᡼Ͽ(dbF) ޥɤȯԤ */
	(void)sprintf(obuf, ESC_GS "%ddbF",
		      id);	/* Ͽֹ */

	result =  CheckWrite(printerContext, obuf, strlen(obuf));
	if (result != OK) {
		return (result);
	}


	return(result);
}



/* ----------------------------------------------------------------
 *
 * Bitmap Image Operations
 *
 * ---------------------------------------------------------------- */


/** PDAPI˽(̤ݡ) */
int DrawImage(int printerContext, int sourceWidth, int sourceHeight, int colorSpace, ImageFormat imageFormat, Rectangle destinationSize, int count, void *imageData)
{
	int		result;
	result = 0;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	return (result);
}


/** PDAPI˽
 * եޥåȤϡdepth=1⤷24iformatRawΤߤ򥵥ݡȤ
 * @todo ᡼ˤơCTMϳȳȽ̤ˤΤ߱ƶ
 *       žˤϱƶʤͤθƤɬס
 */
int StartDrawImage(int printerContext, int sourceWidth, int sourceHeight, int colorDepth, ImageFormat imageFormat, Rectangle destinationSize)
{
	GraphicsState *gs;
	int x, y;

	DebugPrint("pdapi: %s,%d: width=%d, height=%d, imageFormat=%d, imageDepth=%d\n", __FUNCTION__, __LINE__, sourceWidth, sourceHeight, imageFormat, colorDepth);


	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return (ERROR);
	}

	if (gs->imageActive == TRUE) {
		/* imageǤ */
		errorno = BADREQUEST;
		return (ERROR);
	}

	gs->imageFormat = imageFormat;
	gs->imageDepth = colorDepth;

	x = (int)ApplyCtmX_Integer(destinationSize.p0.x, destinationSize.p0.y, gs->ctm)
		- (int)ApplyCtmX_Integer(destinationSize.p1.x, destinationSize.p1.y, gs->ctm);
	x = (int)fabs((double)x);
	y = (int)ApplyCtmY_Integer(destinationSize.p0.x, destinationSize.p0.y, gs->ctm)
		- (int)ApplyCtmY_Integer(destinationSize.p1.x, destinationSize.p1.y, gs->ctm);
	y = (int)fabs((double)y);

	/* ᡼賫 */
	gs->imageActive = TRUE;

	if (imageFormat == iformatRaw) {
		return StartRaws(printerContext, sourceWidth, sourceHeight, colorDepth, x, y);
	}

	/* ̤ݡ */
	errorno = NOTSUPPORTED;
	return (ERROR);
}


/** PDAPI˽
 * @todo
 * 	饹ʬ䤷ƽʤᡢStartDrawImage
 * 	EndDrawImageδ֤TransferDrawImageϰ٤Ƥ֤Ȥ
 * 	Х̤Τ˻Ф1饹˽Ǥ褦
 * 	ɬפ롣
 */
int TransferDrawImage(int printerContext, int count, void *imageData)
{
	GraphicsState *gs;

	DebugPrint("pdapi: %s,%d; count=%d\n", __FUNCTION__, __LINE__, count);

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return (ERROR);
	}

	if (gs->imageActive != TRUE) {
		/* imageǤʤ */
		errorno = BADREQUEST;
		return (ERROR);
	}

	if (gs->imageFormat == iformatRaw) {
		return WriteRaws(printerContext, count, imageData);
	}

	/* ̤ݡ */
	errorno = NOTSUPPORTED;
	return (ERROR);
}


/** PDAPI˽ */
int EndDrawImage(int printerContext)
{
	GraphicsState *gs;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return (ERROR);
	}

	if (gs->imageActive != TRUE) {
		/* imageǤʤ */
		errorno = BADREQUEST;
		return (ERROR);
	}

	/* ᡼轪λ */
	gs->imageActive = FALSE;

	return EndRaws(printerContext);
}



/* ----------------------------------------------------------------
 *
 * Scan Line Operations
 *
 * ---------------------------------------------------------------- */


/** PDAPI˽(̤ݡ) */
int StartScanline(int printerContext, int yposition)
{
	int		result;
	result = 0;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	return(result);
}


/** PDAPI˽(̤ݡ) */
int Scanline(int printerContext, int nscanpairs, int *scanpairs)
{
	int		result;
	result = 0;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	return(result);
}


/** PDAPI˽(̤ݡ) */
int EndScanline(int printerContext)
{
	int		result;
	result = 0;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	return(result);
}



/* ----------------------------------------------------------------
 *
 * Raster Image Operations
 *
 * ---------------------------------------------------------------- */

/** PDAPI˽
 * 顼ڡϡcspaceStandardRGBΤߤ򥵥ݡȤ
 */
int StartRaster(int printerContext, int rasterWidth){
	GraphicsState *gs;

	DebugPrint("pdapi: %s,%d: width=%d\n", __FUNCTION__, __LINE__, rasterWidth);

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return (ERROR);
	}

	if (gs->imageActive == TRUE) {
		/* imageǤ */
		errorno = BADREQUEST;
		return (ERROR);
	}

	/* depthformat꤬ʤᡢGstateϿƤΤѤ */
	/* ϿƤΤǽϰǤФ褤ݡȳä  */
	/* åԤ */
	if(gs->colorSpace != cspaceStandardRGB){
		errorno = BADREQUEST;
		return (ERROR);
	}

	/* ROPͤѹƤСޥɤ */
	WriteROP(printerContext);

	/* ѥѹƤСޥɤ */
	WritePattern(printerContext, PATTERN_FILL);

	/* ᡼(ꤵ줿rasterWidth),᡼ι⤵(1)Ͽ */
	gs->imageWidth  = rasterWidth;
	gs->imageHeight = 1;

	/* ᡼賫 */
	gs->imageActive = TRUE;

	return (OK);
}


/** PDAPI˽ */
int TransferRasterData(int printerContext, int count, unsigned char *data){
	GraphicsState *gs;
	char obuf[64];
	int dataFormatType;
	int pltNo;
	int result;
	result = 0;
	dataFormatType=0;
	pltNo=0;

	DebugPrint("pdapi: %s,%d: count=%d\n", __FUNCTION__, __LINE__, count);

	/** ˡᣱ饹⤵Ԥ
	 * StartRasterǤ賫ϥޥɤʤᡢǥ
	 * ʸޤܥϤ줿ǡ
	 * Ǹ˥ɥޥɤäƽꡣȤϱ䡹
	 * Ȥ򷫤֤
	 */

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return (ERROR);
	}
	/* startRaster̤ξ֤Ǥϼ¹Բ */
	if (gs->imageActive == FALSE) {
		errorno = BADREQUEST;
		return (ERROR);
	}

	/* ǡϥ顼ڡ */
	/* ۤηϤɤޤ礦? */
	if(gs->colorSpace == cspaceStandardRGB){
		dataFormatType = 102;
		pltNo = 0;
	}

	/* ȥå顼᡼賫(scrI) ޥɤȯԤ */
	(void)sprintf(obuf, ESC_GS "%d;%d;%d;%d;%d;%d;%d;%dscrI",
	      2,		/* ColorSpace = RGB */
	      dataFormatType,	/* dataη */
	      pltNo,		/* ѥåϿֹ */
	      0,		/* ̵ */
	      gs->imageWidth, 	/*  */
	      gs->imageHeight, 	/* ι⤵ */
		      /** @note
		       * destinationΰꤹ뤬饹ʤΤ
		       * CTMϱƶͿʤϤʤĤޤ꥽ΰȰ
		       * פʤФʤʤ
		       */
	      gs->imageWidth,	/* ΰ */
	      gs->imageHeight 	/* ΰι⤵ */
	);

	result =  CheckWrite(printerContext, obuf, strlen(obuf));
	if (result != OK) {
		return (result);
	}

#if FALSE
	/* ǡΤ */
	result = WriteRaws(printerContext, count, data);
	if (result != OK) {
		return (result);
	}
#endif

	(void)sprintf(obuf, ESC_GS "%d;%dcu{I", count, gs->imageHeight);
	result =  CheckWrite(printerContext, obuf, strlen(obuf));
	if (result != OK) {
		return (result);
	}

	result = CheckWrite(printerContext, data, count);
	if (result != OK) {
		return (result);
	}

	/* λޥ */
	result = EndRaws(printerContext);
	if (result != OK) {
		return (result);
	}

	/* Y position coutup */
	/** ľ߰л(V) ޥɤȯԤ */
	(void)sprintf(obuf, ESC_GS "+1V");

	result =  CheckWrite(printerContext, obuf, strlen(obuf));
	return (result);
}


/** PDAPI˽(̤ݡ) */
int SkipRaster(int printerContext, int count)
{
	int		result;
	result = 0;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	return(result);
}


/** PDAPI˽ */
int EndRaster(int printerContext){
	GraphicsState *gs;
	char obuf[64];
	int result;
	result = 0;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	gs = GetGraphicsState(printerContext);
	if (gs == NULL) {
		return (ERROR);
	}

	if (gs->imageActive != TRUE) {
		/* imageǤʤ */
		errorno = BADREQUEST;
		return (ERROR);
	}

	/* ᡼轪λ */
	gs->imageActive = FALSE;

	/* Y position */
	/** ľ߰л(V) ޥɤȯԤ */
	(void)sprintf(obuf, ESC_GS "-1V");

	result =  CheckWrite(printerContext, obuf, strlen(obuf));
	if (result != OK) {
		return (result);
	}
	return (OK);
}


/* ----------------------------------------------------------------
 *
 * Stream Data Operations
 *
 * ---------------------------------------------------------------- */


/** PDAPI˽(̤ݡ) */
int StartStream(int printerContext)
{
	int		result;
	result = 0;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	return(result);
}


/** PDAPI˽(̤ݡ) */
int TransferStreamData(int printerContext, int count, void *data)
{
	int		result;
	result = 0;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	return(result);
}


/** PDAPI˽(̤ݡ) */
int EndStream(int printerContext)
{
	int		result;
	result = 0;

	DebugPrint("pdapi: %s,%d \n", __FUNCTION__, __LINE__);

	return(result);
}

/* end of pdapi.c */
