#!/bin/sh
#
# Copyright (c) 2003 by 
# Roland Mainz <roland.mainz@informatik.med.uni-giessen.de>
# All rights reserved.
#
#ident  "@(#)startxprint.sh   0.1    2003/02/08 gisburn"
#
# (Hack-like) wrapper script to use /etc/init.d/xprint directly from a
# tarball without the need to be "root" nor to do any special installation
# steps
#
############################################################################

# Force use of a POSIX conformant sh 
# (Solaris /sbin/sh is plain Bourne shell)
[ "$1" != "posix_sh_forced" -a -x /bin/ksh  ]   && exec /bin/ksh  "$0" posix_sh_forced "$@"
[ "$1" != "posix_sh_forced" -a -x /bin/ksh93  ] && exec /bin/ksh93  "$0" posix_sh_forced "$@"
[ "$1" != "posix_sh_forced" -a -x /bin/bash ]   && exec /bin/bash "$0" posix_sh_forced "$@"
if [ "$1" != "posix_sh_forced" ] ; then
    echo "${0}: No POSIX-shell found." 
    exit 1
fi

shift # Drop "posix_sh_forced"

PATH=/usr/sbin:/usr/bin:/sbin:/bin
export PATH

fatal_error()
{
    echo "${this_script}: ## FATAL ERROR: ${1}" 1>&2
    exit 1
}

relpath2abspath()
{
    (
      dirpart="`dirname "${1}"`"
      basepart="`basename "${1}"`"
    
      echo "`cd "${dirpart}" ; pwd`/${basepart}"
    )
    return 0
}

##### main
this_script="$(relpath2abspath "${0}")"
unpackdir="$(dirname "${this_script}")/install"

## prechecks
[ ! -d "${unpackdir}/etc"               ] && fatal_error "${unpackdir}/etc not found"
[ ! -d "${unpackdir}/usr"               ] && fatal_error "${unpackdir}/usr not found"
[ ! -r "${unpackdir}/etc/init.d/xprint" ] && fatal_error "${unpackdir}/etc/init.d/xprint not found"

xprt_path="$(find "${unpackdir}" -name Xprt -print)"
xplsprinters_path="$(find "${unpackdir}" -name xplsprinters -print)"
psdefault_path="$(find "${unpackdir}" -name PSdefault -print)"

[ "${psdefault_path}"         = "" ] && fatal_error "no PSdefault model found in ${unpackdir}"
[ "${xprt_path}"              = "" ] && fatal_error "no Xprt binary found in ${unpackdir}"
[ "${xplsprinters_path}"      = "" ] && fatal_error "no xplsprinters binary found in ${unpackdir}"
[ ! -r "${psdefault_path}"         ] && fatal_error "PSdefault model \"${psdefault_path}\" found in ${unpackdir} not accessible"
[ ! -x "${xprt_path}"              ] && fatal_error "Xprt binary \"${xprt_path}\" found in ${unpackdir} not accessible/executable"

xpconfig_dir="${psdefault_path%*/C/print/models/PSdefault}"

# run the "/etc/init.d/xprint start" using the binaries/config files/scripts in the tarball
ETC_INITD_XPRINT_XPRT_PATH="${xprt_path}"
ETC_INITD_XPRINT_XPLSPRINTERS_PATH="${xplsprinters_path}"
ETC_INITD_XPRINT_XPCONFIGDIR="${xpconfig_dir}"
export ETC_INITD_XPRINT_XPRT_PATH ETC_INITD_XPRINT_XPCONFIGDIR ETC_INITD_XPRINT_XPLSPRINTERS_PATH
/bin/sh "${unpackdir}/etc/init.d/xprint" "$@"

# EOF.
